import os
import sys
import xlwings as xl
from colorhash import ColorHash
from typing import Any

# Entry point to this file starts with if __name__ == '__main__'
# line at the bottom


# Formatting, alignment, etc

def align_assemblies(sheet: xl.Sheet) -> None:
    "Aligns text in the first three columns to the right"

    # (setq range (xl-ExpandRange (xl-GetRange sheet "A1:C1") "down"))
    # (foreach row range
    #  (foreach cell row
    #   (xl-put-HorizontalAlignment cell acHAlignRight)))

    sheet.range("A1:C1").expand(
        "down"
    ).api.HorizontalAlignment = xl.constants.HAlign.xlHAlignRight


def color_assemblies(sheet: xl.Sheet) -> None:
    "Sets the background color for the first three columns"

    # (setq range (xl-ExpandRange (xl-GetRange sheet "A1:C1") "down"))
    # (foreach row range
    #  (foreach cell row
    #   (xl-put-Color cell "grey")))

    sheet.range("A1:C1").expand("down").color = (240, 240, 240)


def freeze_pane(book: xl.Book) -> None:
    "Freeze the first three columns and first row"
    a = book.app.api.ActiveWindow
    a.FreezePanes = False
    a.SplitColumn = 3
    a.SplitRow = 1
    a.FreezePanes = True


def format_header(sheet: xl.Sheet) -> None:
    """Formats text in the header row

    Formatting: bold, wrap text, heigh 100px, vertical orientation"""
    r = sheet.cells.rows[0]
    r.font.bold = True
    r.wrap_text = True
    r.row_height = 100
    r.api.Orientation = xl.constants.Orientation.xlUpward
    r.rows.autofit()


def autofit(sheet: xl.Sheet) -> None:
    "Autofits column widths from the second column on"
    sheet.range("B1").expand().columns.autofit()


# Coloring columns
# uses ColorHash to assign a different color to each unique value

colors = {}  # global colors cache


def color_columns(sheet: xl.Sheet) -> None:
    "Sets the font color of all cells to their hashed colors"
    for col in sheet.range("D2:EZ2").expand("down").columns:
        if col[0].value is None:
            return
        _color_column(col)


def _color_column(col: xl.Range) -> None:
    """Sets the font color of each cell in col to their hashed colors

    If all the cells in col are the same or empty, skip the whole column"""

    # skip homogeneous columns
    if len(set(filter(lambda x: x is not None, col.value))) == 1:
        return
    for cell in col:
        _color_cell(cell)


def _color_cell(cell: xl.Range) -> None:
    "Sets the font color of cell to its hashed color"
    cell.font.color = _color_hash(cell.value)


def _color_hash(thing: Any) -> None:
    "Returns the cached color of thing, creating if necessary"
    global colors
    if thing not in colors:
        colors[thing] = ColorHash(thing, lightness=(0.3, 0.4, 0.5)).rgb
    return colors[thing]


if __name__ == '__main__':
    # approximate AutoLISP equivalents for each line in comments

    # pretend there is a system that provides an object API similar to
    # vla- functions, that uses the prefix xl- (Excel)

    # first command line argument (no real AutoLISP equiv)
    # sets path variable to the file name path (provided by AutoLISP 'startapp')
    path = sys.argv[1]

    # (setq book (xl-CreateBook path))
    book = xl.Book(path)

    # (setq sheet (nth 0 (xl-get-Sheets book)))
    sheet = book.sheets[0]

    # (xl-Activate (xl-get-App book) t)
    book.app.activate(steal_focus=True)

    # (align_assemblies sheet), etc
    align_assemblies(sheet)
    color_assemblies(sheet)
    freeze_pane(book)
    format_header(sheet)
    autofit(sheet)
    color_columns(sheet)

    # (xl-Save book (string-subst "xls" "csv" path))
    book.save(path.replace('csv', 'xls'))

    # (vl-file-delete path)
    os.remove(path)
