(set-file-docstring
 (strcat "Dialog utilities\n\n" (read-file-to-string "dialog/README.txt")))

(defun-q dialog-init (file name source
                           / dcl-id defaults-delimiter popup-keys-strings string-keys
                           edit-box-keys all-keys data sd sd-keys restart-p
                           dialog-dumpout-func dialog-dumpout-args popup-tile-lists lst)
  "Runs a dialog and returns a data list of keys and values

FILE is the path to a DCL file. NAME is the name of the dialog in the file to run. SOURCE
is either nil or a data list of keys and values to initialize the dialog with. SOURCE
being nil means to use defaults-file."
  (declare (vars (file str)
                 (name str)))
  (setq dcl-id (dialog-dynamic-load
                (list
                 (strcat psc-src-dir "/dialog/common.dcl")
                 (strcat psc-src-dir file)))
        defaults-file (if defaults-file defaults-file "")
        defaults-delimiter ":"
        restart-p T)

  (while restart-p
    (dialog-set-key-vars)
    (new_dialog name dcl-id)

    ;; popup key lists must be populated before defaults load
    (foreach key popup-keys-strings
             (setq lst (dialog-popup-list-get-def key))
             (add-to-list 'popup-tile-lists (cons key lst))
             (dialog-popup-list-set key lst))
             
    (dialog-load-defaults source)

    ;; actions must be added after defaults load
    (foreach tie dialog-ties
             (dialog-tie-via-toggle (car tie)
                                    (cadr tie)))
    (foreach key edit-box-keys
             (dialog-add-action key "(dialog-edit-box-action $key (dialog-tile-val $key))"))
    (foreach key all-keys
             (dialog-add-default-action key))

    ;; hook up sub-dialogs
    (foreach sd sub-dialogs
             (setq sd-keys (if (= (type sd) 'STR) (list sd) (cdr sd))
                   sd (if (= (type sd) 'STR) sd (car sd)))
             ;; add action to buttons
             (foreach sd-key sd-keys
                      (dialog-add-action sd-key
                                         (strcat "(" sd "-sub-init $key)"))))

    (attempt-apply dialog-before-start nil)
    (if (member (start_dialog) '(0 1))
        (setq restart-p nil)
      (progn
        (attempt-apply dialog-dumpout-func dialog-dumpout-args)
        (setq dialog-dumpout-func nil
              dialog-dumpout-args nil))))
  ;; data populated by dialog-save-action or left as nil
  (unload_dialog dcl-id)
  data)
(defun-r 'dialog-init)

(defun-q dialog-set-key-vars ()
  "Sets key variables in %dialog-init"
  (setq popup-keys-strings (mapcar 'car popup-keys)
        edit-box-keys (append dist-keys
                              int-keys
                              plain-string-keys)
        string-keys (append edit-box-keys
                            toggle-keys)
        all-keys (append popup-keys-strings string-keys)
        tile-actions nil))
(defun-r 'dialog-set-key-vars)

(defun-q dialog-tile-val (key / val)
  "Returns the correctly-typed value of an arbitrary dialog tile"
  (declare (vars (key str)))
  (cond ((member key edit-box-keys)
         (edit-box-val key))
        ((member key toggle-keys)
         (get_tile key))
        ((member key popup-keys-strings)
         (dialog-popup-val key))))
(defun-r 'dialog-tile-val)

(defun-q dialog-popup-set (key value / idx)
  "Sets the value of a popup list tile"
  (declare (vars (key str (member key popup-keys-strings))))
  (if (not (setq idx (dialog-popup-idx key value)))
      (setq idx 0))
  (set_tile key (to-string idx)))
(defun-r 'dialog-popup-set)

(defun-q dialog-popup-val (key)
  "Returns the value of popup list tile KEY"
  (declare (vars (key str)))
  (?s (safe-nth (atoi (get_tile key)) (dialog-popup-list-get key))))
(defun-r 'dialog-popup-val)

(defun-q dialog-popup-idx (key value / )
  "Returns the index of VALUE in popup list for KEY"
  (declare (vars (key str)
                 (value str)))
  (vl-position value (dialog-popup-list-get key)))
(defun-r 'dialog-popup-idx)

(defun-q dialog-popup-list-set (key lst / old new)
  "Assigns the list property of popup list tile KEY to LST"
  (declare (vars (key str)))
  (if (not lst) (setq lst '("")))
  (start_list key)
  (mapcar 'add_list lst)
  (end_list)
  (setq new (cons key lst)
        old (assoc key popup-tile-lists))
  (if old
      (subst! new old 'popup-tile-lists)
      (add-to-list 'popup-tile-lists new)))
(defun-r 'dialog-popup-list-set)

(defun-q dialog-popup-list-get-def (key / entry tmp all-strings error-out check-subr)
  "Returns the list associated with popup tile KEY"
  (declare (vars (key str)))
  (setq entry (cdr (safe-nth (vl-position key popup-keys-strings) popup-keys)))

  (defun all-strings (list)
    (vl-every '(lambda (x) (= (type x) 'STR)) list))
  (defun error-out (key)
    (error (strcat "dialog.lsp: Popup list error: " key)))
  (defun check-subr (f / tmp)
    (setq tmp (apply f '()))
    (if (all-strings tmp) tmp (error-out key)))

  (cond
   ;; symbol passed
   ((= (type entry) 'SYM)
    (setq tmp (vl-symbol-value entry))
    (cond
     ;; symbol points to list of strings
     ((and (= (type tmp) 'LIST)
           (all-strings tmp))
      tmp)
     ;; symbol points to function which returns a list of strings
     ((function-p tmp)
      (tmp))
     (T (error-out key))))
   ;; list passed
   ((= (type entry) 'LIST)
    (cond
     ;; all strings
     ((all-strings entry)
      entry)
     ;; lambda which returns all strings
     ((= (car entry) 'lambda)
      (check-subr entry))
     (T (error-out key))))))
(defun-r 'dialog-popup-list-get-def)

(defun-q dialog-popup-list-get (key)
  "Retusnr the list currently associated with popup tile KEY"
  (cdr (assoc key popup-tile-lists)))
(defun-r 'dialog-popup-list-get)

(defun-q dialog-popup-list-update (key)
  "Updates the list associated with popup tile KEY"
  (declare (vars (key str)))
  (dialog-popup-list-set key (dialog-popup-list-get-def key))
  (set_tile key "0"))
(defun-r 'dialog-popup-list-update)

(defun-q edit-box-val (key)
  "Returns the typed value of an arbitrary edit box tile"
  (declare (vars (key str)))
  ((cond ((member key dist-keys) distof)
         ((member key int-keys) safe-atoi)
         ((member key edit-box-keys)
          (lambda (str)
            (if (= str "")
                nil
              str))))
   (get_tile key)))
(defun-r 'edit-box-val)

(defun-q dialog-edit-box-action (key string / value tmp)
  "Default action for edit box tiles

Sets box value to ERR if left blank"
  (declare (vars (key str)))
  (if (setq value (edit-box-val key))
      (set_tile key (to-string value))
      ;; allow comment and sequence to be nil
      (if (member key '("comment" "sequence"))
          (set_tile key "")
          (set_tile key "ERR")))
  value)
(defun-r 'dialog-edit-box-action)

(defun-q dialog-add-action (dac-key action)
  "Adds an action to tile DAC-KEY

Keeps track of current actions and appends rather than clobbering existing actions."
  (declare (vars (dac-key str)
                 (action (list str))))
  (if (= (type action) 'LIST)
      (setq action (to-string (macro-expand action))))
  (add-to-alist 'tile-actions dac-key action t)
  (action_tile dac-key (cdr (assoc dac-key tile-actions))))
(defun-r 'dialog-add-action)

(defun-q dialog-add-default-action (key / name action)
  "Automatically add the default action for KEY"
  (declare (vars (key str)))
  (setq name (strcat key "-action")
        action (vl-symbol-value (read name)))
  (cond ((member (type action) '(SUBR USUBR))
         (dialog-add-action key (strcat "(" name " (dialog-tile-val $key))"))))
  (cond ((member key (mapcar 'car default-actions))
         (setq name (cdr (assoc key default-actions))
               action (vl-symbol-value name))
         (dialog-add-action key (strcat "(" (to-string name) " $key (dialog-tile-val $key))")))))
(defun-r 'dialog-add-default-action)

(defun-q dialog-toggle-enable (value key)
  "Enables or disables tile KEY based on toggle VALUE"
  (declare (vars (value str (member value '("0" "1")))
                 (key str)))
  (if (= value "0")
      (mode_tile key 1)
      (mode_tile key 0)))
(defun-r 'dialog-toggle-enable)

(defun-q dialog-copy-value-if-disabled (value key toggle)
  "Set KEY to VALUE if TOGGLE is checked"
  (declare (vars (value str)
                 (key str)
                 (toggle str)))
  (if (= (get_tile toggle) "0")
      (set_tile key value)))
(defun-r 'dialog-copy-value-if-disabled)

(defun-q dialog-tie-via-toggle (alist toggle / primary secondary begin-tied)
  "Connect the values of tiles when TOGGLE is unchecked"
  (declare (vars (alist list)
                 (toggle str)))
  (setq begin-tied T)
  (foreach x alist
     (setq primary (car x)
           secondary (cdr x))
     (dialog-add-action primary '(dialog-copy-value-if-disabled $value #secondary #toggle))
     (dialog-add-action toggle '(dialog-toggle-enable $value #secondary))
     (dialog-add-action toggle '(if (= "0" $value) (set_tile #secondary (get_tile #primary))))
     (if (not (= (get_tile primary)
                 (get_tile secondary)))
         (progn
           (setq begin-tied nil))))
  (if begin-tied
      (foreach x alist
               (mode_tile (cdr x) 1))
      (set_tile toggle "1")))
(defun-r 'dialog-tie-via-toggle)

(defun-q dialog-get-data (typed / data key val lst)
  "Returns a list of keys and values from the current dialog

If TYPED is non-nil, return them as typed values. Otherwise, strings."
  (foreach key popup-keys-strings
           (add-to-list 'data
                        (list key (dialog-popup-val key))))
  (foreach key string-keys
           (setq val (get_tile key))
           (if (= val "ERR")
               (error "dialog.lsp: Input errors")
               (add-to-list 'data
                            (list key
                                  (if typed
                                      (dialog-tile-val key)
                                      val)))))
  (foreach buttons sub-dialogs
           (if (= (type buttons) 'LIST)
               (setq buttons (cdr buttons))
               (setq buttons (list buttons)))
           (foreach button buttons
                    (mapcar '(lambda (b)
                              (add-to-list 'data
                               (list (car b)
                                (if typed
                                    (cadr b)
                                    (to-string (cadr b))))))
                            (vl-symbol-value (read (strcat button "-val"))))))
  (remove! 'data '(member (cadr x) (list "" nil)))
  (vl-sort data sort-data-by-key))
(defun-r 'dialog-get-data)

(defun-q dialog-update-defaults-file (dialog-name / files local-file curver copy)
  "Replaces defaults file if it is older than the current version"
  (setq files (reverse (vl-directory-files psc-data-dir
                                           (strcat dialog-name "-defaults.dat")))
        local-file (strcat psc-data-dir dialog-name "-defaults.dat")
        main-file (strcat psc-src-dir dialog-name "/defaults.dat")
        copy nil)

  ;; list to string
  (defun lts (lst / str)
    (setq str "")
    ;; remove day of week
    (foreach elt (remove-nth 2 lst)
             (setq str (strcat str
                               (if (< elt 10)
                                   (strcat "0" (vl-princ-to-string elt))
                                 (vl-princ-to-string elt)))))
    str)

  (if files
      (if (< (lts (vl-file-systime (strcat psc-data-dir (car files))))
             (lts (vl-file-systime main-file)))
          (progn
            (alert "Defaults file updated. Default settings will be reset.")
            (foreach file files
                     (vl-file-delete (strcat psc-data-dir file)))
            (setq copy T))
        (setq local-file (strcat psc-data-dir (car files))))
    (setq copy T))
  (if copy
      (vl-file-copy main-file local-file))
  local-file)
(defun-r 'dialog-update-defaults-file)

(defun-q dialog-save-defaults (file / filename data key val )
  "Writes current dialog state to the defaults file"
  (if (not file)
      (setq file defaults-file))
  (setq filename file)

  (if (vl-catch-all-error-p (setq data (vl-catch-all-apply 'dialog-get-data '(nil))))
      (alert "Input errors!\nDefaults not saved")
    (if (setq file (open file "w"))
        (if (has-duplicate-keys-p data)
            (progn
              (close file)
              (vl-file-delete filename)
              (error "dialog.lsp: Attempting to save defaults with duplicate keys present"))
          (progn
            (foreach item data
                     (setq key (car item)
                           val (cadr item))
                     (write-line (strcat key defaults-delimiter val) file))
            (close file)))
        (error "dialog.lsp: Defaults file not found"))))
(defun-r 'dialog-save-defaults)

(defun-q dialog-load-defaults (source / file line key keytype value
                                      process-entry sd sd-keys
                                      sd-dist-keys key-p dist-key-p
                                      valsym prefix)
  "Populates the dialog with values from SOURCE, or the defaults file"
  (setq extra-keys nil)
  (defun process-entry (key value)
    (cond ((member key popup-keys-strings)
           (dialog-popup-list-update key)
           (dialog-popup-set key value))
          ((member key string-keys)
           (set_tile key value))
          (T
           (add-to-list 'extra-keys (list key value)))))
  (if (and (not source)
           (setq file (open defaults-file "r")))
      (progn
        (while (setq line (read-line file))
          (setq line (strsplit line defaults-delimiter))
          (process-entry (car line) (cadr line)))
        (close file))
    (foreach x source
             (process-entry (car x) (to-string (cadr x)))))
  ;; sub-dialog keys
  (foreach sd sub-dialogs
           (if (= (type sd) 'LIST)
               ;; multi sub dialog
               (setq sd-buttons (cdr sd)
                     sd (car sd)
                     key-p (lambda (x / prefix)
                             (setq prefix (substr b 1 (vl-string-search "_" b)))
                             (and (begins-with (car x) prefix)
                                  (member (car x) sd-keys))))
             ;; single sub dialog
             (setq sd-buttons (list sd)
                   key-p (lambda (x) (member (car x) sd-keys))))
           
           (setq sd-keys (vl-symbol-value (read (strcat sd "-keys")))
                 sd-dist-keys (vl-symbol-value (read (strcat sd "-dist-keys")))
                 dist-key-p (lambda (x)
                              (member (car x) sd-dist-keys)))

           (foreach b sd-buttons
              (setq valsym (read (strcat b "-val")))
              (split-list extra-keys '(key-p x) valsym 'extra-keys)
              ;; translate dist-keys in place
              (foreach k (vl-symbol-value valsym)
                 (if (dist-key-p k)
                     (subst-key! (car k) (distof (cadr k)) valsym)))))

  ;; hook optionally defined by user
  (attempt-apply dialog-after-load nil))
(defun-r 'dialog-load-defaults)

(defun-q dialog-load-defaults-partial ( / )
  "Loads keys from load-defaults-keys into dialog"
  (setq source (dialog-get-data nil)
        defs (read-key-val-file defaults-file))
  (foreach key load-defaults-keys
           (subst-key! key (cadr (assoc key defs)) 'source))
  (dialog-load-defaults source))
(defun-r 'dialog-load-defaults-partial)

(defun-q dialog-save-action ( / )
  "Dialog default save action"
  (attempt-apply dialog-before-save nil)
  (setq data (dialog-get-data T))
  (attempt-apply dialog-before-done nil)
  (done_dialog 0))
(defun-r 'dialog-save-action)

(defun-q dialog-dynamic-load (files / tmpfile f dcl-id)
  "Dynamically loads a number of DCL files together into a single dcl-id"
  (setq tmpfile (vl-filename-mktemp "dialog" tmpdir ".dcl")
        files (if (= (type files) 'STR) (list files) files))
  (foreach file files
           (vl-file-copy file tmpfile T))
  (setq dcl-id (load_dialog tmpfile))
  (vl-file-delete tmpfile)
  (if dcl-id dcl-id (error "dialog.lsp: Load failure")))
(defun-r 'dialog-dynamic-load)

(defun-q dialog-start ( / return)
  "Replacement for start_dialog and unload_dialog

Calls both and returns the value of start_dialog"
  (setq return (start_dialog))
  (unload_dialog dcl-id)
  return)
(defun-r 'dialog-start)

(defun-q dialog-sub-update-from-def (button def /
                                          valsym
                                          curval
                                          unisubs
                                          multisubs
                                          sd
                                          dist-keys
                                          plain-string-keys
                                          popup-keys
                                          toggle-keys
                                          string-keys
                                            process-key)
  "Updates the -val variable for a sub dialog BUTTON based on DEF"
  (setq valsym (read (strcat button "-val"))
        curval (vl-symbol-value valsym))
  ;; ascertain whether this is a single-button sub (unisub) or
  ;; multi-button sub (multisub)
  (split-list sub-dialogs '(= (type x) 'STR) 'unisubs 'multisubs)
  (setq sd (if (member button unisubs)
               button
             (caar (vl-member-if '(lambda(x)
                                    (vl-some '(lambda(y) (= y button)) x))
                                 multisubs))))
  ;; load key lists for sub dialog in question
  ;; in order to process the definition into concrete values
  ;; as popup-keys may refer to variables for their current defaults
  (foreach x '(dist-keys plain-string-keys popup-keys toggle-keys)
           (set x (vl-symbol-value (read (strcat sd "-" (to-string x))))))
  (setq string-keys (append dist-keys plain-string-keys toggle-keys))
  (defun process-key (x / key value fix-popup-keys)
    (setq key (car x))
    (cond ((member key string-keys) x)
          ((member key popup-keys)
           (defun fix-popup-keys (key value) (subst! (cons key value) key 'popup-keys))
           (list key
                 (if (= (type (cdr x)) 'LIST)
                     ;; plain string list
                     (progn
                       (setq value (cdr x))
                       (fix-popup-keys key value)
                       (car value))
                   ;; symbol points to list val
                   (progn
                     (setq value (vl-symbol-value (cdr x)))
                     (fix-popup-keys key value)
                     (car value)))))))
  (mapcar! 'process-key 'def)
  (remove! 'popup-keys '(= (type x) 'STR))
  ;; remove all keys which are not part of def from -val
  (setq def-strings (mapcar 'car def))
  (remove! 'curval '(not (member (car x) def-strings)))
  ;; add ones that are not present
  (setq curval-strings (mapcar 'car curval))
  (foreach key def-strings
     (if (not (member key curval-strings))
         (add-to-list 'curval (assoc key def))
       ;; reset popup-keys whose backend lists have changed
       (if (and (member key (mapcar 'car popup-keys))
                (not (member (cadr (assoc key curval))
                             (cdr (assoc key popup-keys)))))
           (subst-key! key (cadr (assoc key def)) 'curval))))
  (set valsym curval))
(defun-r 'dialog-sub-update-from-def)
