import json
from jinja2 import Environment, FileSystemLoader
import os
import re
import sys
from typing import Optional, NoReturn


def read_json() -> dict:
    "Returns the JSON file files.json as a dictionary"
    with open('files.json') as f:
        return json.load(f)


def filename_as_html(filename: str) -> str:
    "Converts a lsp filename to HTML"
    return filename.replace('lsp', 'html').replace('/', '-')


def filelink(value: str, filename: Optional[str] = None) -> str:
    "Custom Jinja2 filter for linking to a lsp file"
    if filename is None:
        filename = value
    html_filename = filename_as_html(filename)
    return f'<a href="./{html_filename}">{value}</a>'


def funclink(value: str, funcname: Optional[str] = None) -> str:
    "Custom Jinja2 filter for linking to a function"
    if funcname is None:
        funcname = value
    filename = filename_as_html(functions[funcname.lower()]['file'])
    return f'<a href="{filename}#{funcname}">{value}</a>'


def _first_extra_paren(s: str) -> int:
    "Returns the index of the first extra right paren in string s"
    depth = 0
    idx = 0
    for c in s:
        if c == '(':
            depth += 1
        if c == ')':
            depth -= 1
        if depth < 0:
            return idx
        idx += 1


def prettify_quotes(string: str) -> str:
    "Custom Jinja2 filter to turn (QUOTE FORM) into 'FORM"
    while (start := string.find('(QUOTE ')) > 0:
        s7 = start + 7
        end = _first_extra_paren(string[s7:]) + s7
        string = string[0:start] + "'" + string[s7:end] + string[end+1:]
    return string


def parse_funclinks(string: str) -> str:
    "Custom Jinja2 filter for embedding function links"
    while match := re.search(r'%[!:&a-zA-Z0-9_<>\*\-]+', string):
        match = match[0]
        string = re.sub(match, funclink(match[1:]), string, 1)
    return string


def write_file_template(env: Environment, file_dict: dict) -> None:
    """Creates a doc page for all the functions in a file dictionary

Uses file.jinja template"""
    html_filename = filename_as_html(file_dict['name'])
    file_dict['priv_functions'] = []
    file_dict['pub_functions'] = []
    for func in file_dict['functions']:
        file_dict[f"{func['access']}_functions"].append(func)
    file_dict['functions'] = file_dict['pub_functions'] + file_dict['priv_functions']
    with open(html_filename, 'w') as f:
        f.write(env.get_template('file.jinja').render(file_dict))
    print(f'Wrote {html_filename}')


def error_pause(e: Exception) -> NoReturn:
    "Displays an Exception's message, waits for input, and exits"
    print(f"[{type(e).__name__}] {e}")
    input("Press ENTER to continue...")
    sys.exit()


if __name__ == '__main__':
    global files
    global functions

    try:
        os.chdir(sys.argv[1])
        files = read_json()
    except Exception as e:
        error_pause(e)

    env = Environment(
        loader=FileSystemLoader('templates'),
    )
    env.filters['filelink'] = filelink
    env.filters['funclink'] = funclink
    env.filters['parsefunclinks'] = parse_funclinks
    env.filters['prettifyquotes'] = prettify_quotes

    functions = {func['name']: func for f in files for func in f['functions']}

    try:
        # write main index file using index.jinja template
        with open('index.html', 'w') as f:
            f.write(env.get_template('index.jinja').render(
                {'files': files,
                 'functions': sorted(
                     list(functions.values()),
                     key=lambda d: d['name'],
                 )}
            ))
            print("Wrote index.html")
        for f in files:
            write_file_template(env, f)
    except Exception as e:
        error_pause(e)
