(set-file-docstring "Support functions for FABDATA

For scoping, every function here expects to be called by %fabdata-process-dwgs")

(psc-include '("fab/util.lsp"))

(defun-q fabdata-process-dwgs (files
                               / jobno matlst material hdwlst hardware assemblies stairno)
  "Processes all FILES using %fabdata--process-dwg and writes result to a .csv"
  (declare (vars (files list (vl-every 'stringp files))))
  (setq jobno (xd-value "number" (get-job-info))
        matlst (read-csv "fab/material.csv")
        hdwlst (read-csv "fab/hardware.csv"))
  (foreach file files (fabdata--process-dwg file))
  (if hardware
      (setq hardware (mapcar '(lambda(hdw)
                               (list
                                jobno
                                ""
                                stairno
                                "" "" "" ""
                                (xd-value "hdw_code" hdw)
                                (xd-value "desc" hdw)
                                "" "" "" "" "" "" ""
                                (xd-value "qty" hdw)))
                             (combine-like-data hardware '("desc" "hdw_code") nil))))
  (setq ret (append assemblies material hardware))
  (debug-print-vars '(assemblies material hardware))
  (if ret
      (write-csv ret (strcat (vla-get-Path acadDoc) "/" jobno " stair " stairno " fabdata.csv") T)))
(defun-r 'fabdata-process-dwgs)

(defun-q fabdata--process-dwg (filename / inner ret)
  "Opens a fab drawing and processes all the tables"
  (declare (vars (filename str)))
  (defun inner (dwgno / finish)
    (setq finish (get-finish))
    (foreach table (get-all-tables)
      ((lambda(table / title)
         (setq title (table-get-title table))
         (cond ((= title "Material") (fabdata--process-assembly-material table))
               ((= title "Misc Material") (fabdata--process-misc-material table))
               ((= title "Bracket Hardware") (fabdata--process-bracket-hardware table))
               ((wcmatch title "*Hardware") (fabdata--process-hardware table))
               ((or (= title "Handrail Returns")
                    (= title "Wall Rails")
                    (wcmatch title "*Stair Rails")
                    (wcmatch title "Assembly*"))
                (fabdata--process-assemblies table))
               ))
       table)))
  (open-and (strcat (vla-get-Path acadDoc) "/" filename)
            T
            'inner
            (list (string-subst-all "" ".dwg" filename))))
(defun-r 'fabdata--process-dwg)

(defun-q fabdata--mat-table-total-weight (table / n title)
  "Returns the total weight

Calculates the total based on row weight values; does not rely on summary row."
  (declare (vars (table vla-object (material-table-p table))))
  (setq parsed-table (parse-table table)
        n (length parsed-table))
  (if (> 0)
      (rtos (/ (apply '+
                      (mapcar '(lambda(row / weight)
                                (setq weight (xd-value "weight" row))
                                (if weight
                                    (?f (read weight))
                                    0.0))
                              parsed-table))
               n)
            2 2)
      ""))
(defun-r 'fabdata--mat-table-total-weight)

(defun-q fabdata--process-assemblies (table / parsed-table weight cols desc)
  "Processes an Assembly* table for assemblies

Attempts to set the piece weight using the Material table."
  (declare (vars (table vla-object (table-p table))))
  ((lambda(/ mat-table)
     (setq mat-table (get-table-by-title "Material")
           weight (if mat-table (fabdata--mat-table-total-weight mat-table) ""))))
  (if (wcmatch title "Assembly:*")
      (setq desc (vl-string-left-trim " " (substr title 10))
            cols '("pm" "stair" "lvl" "seq"))
      (setq desc (get-custom-dwgprop "description")
            cols '("mark" "stair_number" "level" "sequence")))
  (foreach assembly (parse-table table)
    (setq stairno (xd-value (nth 1 cols) assembly))
    (add-to-list 'assemblies
      (list
       jobno
       (xd-value (nth 0 cols) assembly)
       (xd-value (nth 1 cols) assembly)
       (xd-value (nth 2 cols) assembly)
       (xd-value (nth 3 cols) assembly)
       "" "" ""
       desc
       "" ""
       weight
       ""
       finish
       dwgno
       ""
       "1"))))
(defun-r 'fabdata--process-assemblies)

(defun-q fabdata--get-material-code (desc / mat)
  "Returns the short code for a material description, if it exists, or an empty string"
  (declare (vars (desc str)))
  (setq mat (assoc desc matlst))
  (if mat (nth 5 mat) ""))
(defun-r 'fabdata--get-material-code)

(defun-q fabdata--process-assembly-material (table / assemblies n)
  "Processes a Material table (paired with an Assembly* table) for ship loose items"
  (declare (vars (table vla-object (table-p table))))
  (setq assemblies (get-table-by-title "Assembly*"))
  (if (not assemblies) (error "No Assembly* table found while processing Material table"))
  (setq assemblies (parse-table assemblies)
        n (length assemblies))
  (foreach mat (parse-table table)
    (if (wcmatch (xd-value "note" mat) "*SL*")
        (foreach assembly assemblies
          (add-to-list 'material
            (append
             (with-data assembly
               '((list jobno pm stair lvl seq)))
             (with-data mat
               '((list
                  (fabdata--get-material-code desc)
                  mark
                  ""
                  desc
                  width
                  length
                  weight
                  note
                  finish
                  dwgno
                  ""
                  (/ (atoi qty) n))))))))))
(defun-r 'fabdata--process-assembly-material)

(defun-q fabdata--process-misc-material (table)
  "Processes a Misc Material table for ship loose items"
  (declare (vars (table vla-object (table-p table))))
  (foreach mat (parse-table table)
    (if (wcmatch (xd-value "note" mat) "*SL*")
        (add-to-list 'material
          (with-data mat
            '((setq stairno stair_number)
              (list
               jobno piece_type stair_number level sequence
               (fabdata--get-material-code desc)
               mark
               ""
               desc width length weight note
               finish dwgno
               ""
               qty)))))))
(defun-r 'fabdata--process-misc-material)

(defun-q fabdata--get-hardware-code (desc / hdw)
  "Returns the short code for a hardware description, if it exists, or an empty string"
  (declare (vars (desc str)))
  (setq hdw (assoc desc hdwlst))
  (if hdw (nth 1 hdw) ""))
(defun-r 'fabdata--get-hardware-code)

(defun-q fabdata--format-hardware (hdw)
  "Adds hardware code and makes qty an integer"
  (declare (vars (hdw list (data-list-p hdw))))
  (cons (list "hdw_code" (fabdata--get-hardware-code (xd-value "desc" hdw)))
        (subst-key "qty" (atoi (xd-value "qty" hdw)) hdw)))
(defun-r 'fabdata--format-hardware)

(defun-q fabdata--process-hardware (table)
  "Processes a *Hardware table for ship loose entries"
  (declare (vars (table vla-object (table-p table))))
  (foreach hdw (parse-table table)
    (if (wcmatch (xd-value "note" hdw) "*SL*")
        (add-to-list 'hardware (fabdata--format-hardware hdw)))))
(defun-r 'fabdata--process-hardware)

(defun-q fabdata--process-bracket-hardware (table / parsed-table)
  "Processes a legacy Bracket Hardware table

This can be removed after people finish the jobs they started during this rewrite and this table style disappears"
  (setq parsed-table (parse-table table))
  (mapcar '(lambda(row / dn ts)
            (setq dn (xd-value "#_drive_nails" row)
                  dn (if dn dn (xd-value "1/4_x_2_drive_nails" row))
                  ts (xd-value "#_tek_screws" row)
                  ts (if ts ts (xd-value "#12_tek_screws" row)))
             (if (> (safe-atoi dn) 0)
                 (add-to-list 'hardware
                   (fabdata--format-hardware (#! '(("qty" #dn)
                                                   ("desc" "DRIVE NAIL 1/4 x 2")
                                                   ("note" "SL"))))))
             (if (> (safe-atoi ts) 0)
                 (add-to-list 'hardware
                    (fabdata--format-hardware (#! '(("qty" #ts)
                                                    ("desc" "TEK SCREW #12 x 3")
                                                    ("note" "SL")))))))
          (take (1- (length parsed-table)) parsed-table)))
(defun-r 'fabdata--process-bracket-hardware)
