(set-file-docstring "Functions for handrail return fab drawings")

(psc-include '("rail/return.lsp"))

(defun-q make-handrail-return-fabs (stairs / finish)
  "Creates all the handrail return fab drawings for this group of STAIRS"
  (declare (vars (stairs list (vl-every 'data-list-p stairs))))
  (foreach returns (calc-handrail-return-fabs stairs)
           (setq finish (pop! 'returns))
           (while returns (make-handrail-return-fab finish (take! 48 'returns)))))
(defun-r 'make-handrail-return-fabs)

(defun-q make-handrail-return-fab (finish returns)
  "Uses %make-rail-fab-common to create a handrail return spreadsheet"
  (declare (vars (finish str)
                 (returns list (vl-every 'data-list-p returns))))
  (debug-print '("Making handrail return fab"))
  (make-rail-fab-common "handrail-return" '(handrail-return-write-table finish returns)))
(defun-r 'make-handrail-return-fab)

(defun-q handrail-return-write-table (finish returns / table)
  "Function used by %make-rail-fab-common to write the main table"
  (declare (vars (finish str)
                 (returns list (vl-every 'data-list-p returns))))
  (set-finish finish)
  (table-add-data (get-table-by-title "Handrail Returns") returns))
(defun-r 'handrail-return-write-table)

(defun-q calc-handrail-return-fabs (stairs)
  "Returns the handrail returns for this group of STAIRS grouped by finish"
  (declare (vars (stairs list (vl-every 'data-list-p stairs))))
  (combine-handrail-return-fabs
   (apply-by-twos 'calc-handrail-return-fab stairs)))
(defun-r 'calc-handrail-return-fabs)

(defun-q combine-handrail-return-fabs (returns / key return-map)
  "Combines RETURNS with equal finishes into a map"
  (declare (vars (returns list)))
  (foreach return (strip-nil returns)
           (setq key (pop! 'return))
           (add-to-alist 'return-map key (list return) T))
  return-map)
(defun-r 'combine-handrail-return-fabs)

(defun-q calc-handrail-return-fab (s1 s2
                                      / return-type data s2-ext s2-r-ext s2-line s1-ext
                                      s1-r-ext s1-line )
  "Returns a list of strings representing the return between stairs S1 and S2, or nil

Returns nil when %rail-return-type returns nil"
  (declare (vars (s1 list (data-list-p s1))
                 (s2 list (data-list-p s2))))
  (defun tk2 (explst) (translate-keys s2 explst))
  (setq return-type (rail-return-type s1 s2))
  (if return-type
      (progn
        (rail-return-common)
        (setq top-centerline (+ (xd-value "dist_btwn" s1)
                                4.66
                                (?f (if (= (xd-value "rail_topmnt" s1) "1") 1.5))
                                (?f (if (= (xd-value "rail_topmnt" s2) "1") 1.5))))
        (cons (xd-value "finish" s1)
              (append
               (#! '(("mark" "HR")
                     ("stair_number" #(xd-value "number" s1))
                     ("level" #(xd-value "level" s1))
                     ("sequence" #(xd-value "sequence" s1))
                     ("ascend" #(xd-value "ascend" s1))
                     ("detail" #(strcat
                                 (substr return-type 1 1) "-" (substr return-type 2 1)))
                     ("top_rail_centerline" #(rtos top-centerline 2 2))))
               (mapcar 'list
                 '("line_1" "angle_1" "line_2" "angle_2" "line_3" "angle_3" "line_4" "angle_4"
                   "cut_length" "rotation")
                 (apply (read (strcat "calc-" return-type "-return"))
                        (list s1 s2))))))))
(defun-r 'calc-handrail-return-fab)

(defun-q pad-return-lines (lines / len)
  "Ensures that each return is 8 cells"
  (declare (vars (lines list)))
  (setq lines (reverse lines)
        len (rtos (ceil (distof (pop! 'lines))) 2 0))
  (while (< (length lines) 8)
    (setq lines (cons "" lines)))
  (reverse (cons len lines)))
(defun-r 'pad-return-lines)

(defun-q calc-return-fab-common (far-pts rotation-def)
  "Returns the lines and rotation portion of the return table entry"
  (setq z (append (lst* top-centerline far-pts)
                  (lst* 0 (- (length pts) far-pts)))
        pts (mapcar 'rcons pts z)
        segs (calc-segs pts)                   
        angles (calc-angles pts)
        lines (pad-return-lines (pipe-bend-lines (zip segs angles)))
        rotation (rtos (eval rotation-def) 2 2))
  (rcons lines rotation))
(defun-r 'calc-return-fab-common)

(defun-q calc-AA-or-AH-return (s1 s2 AA-p
                                  / angles segs pts z lines rotation)
  "Common function for calculating returns with a rotation

Used by %calc-AA-return and %calc-AH-return"
  (setq pts ((if AA-p rail-return-AA-pts rail-return-AH-pts) s1 s2)
        rotation-def '(rtd (dihedral (apply 'normal-vec (take 3 pts))
                                     (apply 'normal-vec (take 3 (cdr pts))))))
  (calc-return-fab-common 2 rotation-def))
(defun-r 'calc-AA-or-AH-return)
  
(defun-q calc-AA-return (s1 s2)
  "Calculates an AA return for fab"
  (calc-AA-or-AH-return s1 s2 T))
(defun-r 'calc-AA-return)

(defun-q calc-AH-return (s1 s2)
  "Calculates an AH return for fab"
  (calc-AA-or-AH-return s1 s2 nil))
(defun-r 'calc-AH-return)

(defun-q calc-EA-return (s1 s2 / pts)
  "Calculates an EA return for fab"
  (setq pts (apply 'append (rail-return-EA-lines s1 s2)))
  (calc-return-fab-common 2 (+ (xd-value "pitch" s1) (xd-value "pitch" s2))))
(defun-r 'calc-EA-return)

(defun-q calc-EH-return (s1 s2 / pts)
  "Calculates an EH return for fab"
  (setq pts (apply 'append (rail-return-EH-lines s1 s2)))
  (calc-return-fab-common 2 (xd-value "pitch" s2)))
(defun-r 'calc-EH-return)

(defun-q calc-HA-return (s1 s2)
  "Calculates an HA return for fab"
  (setq pts (apply 'append (rail-return-HA-lines s1 s2)))
  (calc-return-fab-common 3 (xd-value "pitch" s1)))
(defun-r 'calc-HA-return)

(defun-q calc-HH-return (s1 s2)
  "Calculates an HH return for fab"
  (setq pts (apply 'append (rail-return-HH-lines s1 s2)))
  (calc-return-fab-common 3 0))
(defun-r 'calc-HH-return)
