(set-file-docstring "Common functions for rail calculations")

(defun-q calc-stair-rail-opcl (side)
  "Returns the outer post centerline length of a stair rail"
  (declare with-data)
  (+ (- run 6.25)
     (- bot_cdist 1.25)
     (- top_cdist 1.25)
     (if (wcmatch bot_conn "*haunch*")
         (min tread_depth bot_ext)
         bot_ext)
     (sidekey-rail "top_ext")
     (sidekey-rail "bot_ext")))
(defun-r 'calc-stair-rail-opcl)

(defun-q rail-calc-n-mid-posts (len / n)
  "Returns the number of mid posts based on outside post centerline

Maximum post centerline is 48 inches"
  (declare (vars (len nil (numberp len)))
           (tests (= (rail-calc-n-mid-posts 12) 0)
                  (= (rail-calc-n-mid-posts 48) 0)
                  (= (rail-calc-n-mid-posts 48.01) 1)))
  (setq n 0)
  (while (> len 48.0)
    (setq len (- len 48.0)
          n (1+ n)))
  n)
(defun-r 'rail-calc-n-mid-posts)

(defun-q rail-calc-picket-spacing (len / n spc)
  "Returns the space between pickets based on panel length

Max space between is 3 7/8 inches"
  (declare (vars (len nil (numberp len))))
  (setq n 1
        spc 4)
  (while (>= spc 3.875)
    (setq spc (/ (- len (* n 0.5))
                 (1+ n))
          n (1+ n)))
  spc)
(defun-r 'rail-calc-picket-spacing)

(defun-q z-offset-stringer (side / ret)
  "Returns the 3D Z-offset to the face of stringer at SIDE"
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  ((if (or (and (= z_pos "Near") (= side "o"))
           (and (= z_pos "Far") (= side "i")))
       + -)
   (/ width 2.0)))
(defun-r 'z-offset-stringer)

(defun-q z-offset-stair-rail (side / ret)
  "Returns the 3D Z-offset to the face of the stair rail at SIDE

If the rail is not top-mounted, returns the same value as %z-offset-stringer"
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  (setq ret (z-offset-stringer side))
  (if (= (sidekey-rail "topmnt") "1")
      (- ret (* (unit-dir ret) 1.5))
      ret))
(defun-r 'z-offset-stair-rail)

(defun-q z-offset-stair-rail-handgrab (side / off z-dir)
  "Returns the 3D Z-offset to centerline of stair rail handgrab at SIDE"
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  (setq off (z-offset-stair-rail side)
        z-dir (unit-dir off))
  (- off (* z-dir 2.33)))
(defun-r 'z-offset-stair-rail-handgrab)

(defun-q z-offset-wall-rail (side / ret)
  "Returns the 3D Z-offset to centerline of wall rail at SIDE"
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  (setq ret (z-offset-stringer side))
  (- ret (* (unit-dir ret) 1.33)))
(defun-r 'z-offset-wall-rail)

(defun-q rail-sequence (side / otherside sr-p)
  "Returns the sequence for the stair or wall rail at SIDE

Handles giving rails -L and -R sequence when rail is the same type both sides."
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  (setq sr-p (wcmatch (sidekey-rail "type") "*Stair*")
        otherside (if (= side "i") "o" "i"))
  (strcat (to-string sequence)
          (if (= (wcmatch (symcat '(#otherside "-rail_type"))
                          "*Stair*")
                 sr-p)
              (if (or (and (= ascend "Left")
                           (= side "i"))
                      (and (= ascend "Right")
                           (= side "o")))
                  "-L"
                  "-R")
              "")))
(defun-r 'rail-sequence)

(defun-q stair-rail-AB (side / topx botx)
  "Returns the AB length of the handgrab at SIDE"
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  (setq botx (if (= (sidekey-rail "bot_dloop") "0")
                 (- 3 (min bot_ext tread_depth) (sidekey-rail "bot_ext"))
                 (if (= rail_lvl-off "0")
                     (min (- tread_depth)
                          ;; account for minimum 6" horizontal leg @ 3-bend lower
                          (- 3.125 bot_ext (sidekey-rail "bot_ext") 6))
                     (- tread_depth)))
        topx (+ run
                (if (= (sidekey-rail "top_dloop") "0")
                    (- (sidekey-rail "top_ext") 3)
                    0)))
  (/ (- topx botx) (cos (dtr pitch))))
(defun-r 'stair-rail-AB)

(defun-q rail-ascend (side)
  "Returns the ascend direction of the stair or wall rail at SIDE"
  (declare (vars (side str (member side '("i" "o"))))
           with-data)
  (if (= side "i")
      ascend
      (if (= ascend "Left") "Right" "Left")))
(defun-r 'rail-ascend)

(defun-q wall-rail-brackets (ab / n b spc ret)
  "Returns the list of wall rail bracket locations for a given AB length

Max centerline distance is 60 inches"
  (declare (vars (ab nil (numberp ab)))
           (tests (equal (wall-rail-brackets 24) '(6 18))
                  (equal (wall-rail-brackets 72) '(6 66))
                  (equal (wall-rail-brackets 73) '(6 36 67))))
  (setq ab (- ab 12)
        n 2
        b 0)
  (while (> (setq spc (/ (float ab) (1- n))) 60)
    (inc! 'n))
  (repeat n
          (add-to-list 'ret b)
          (setq b (+ b spc)))
  (mapcar '(lambda(x) (fix (+ 6 x))) ret))
(defun-r 'wall-rail-brackets)

(defun-q make-rail-fab-common (template eval-form 
                                        / oldacaddoc path jobno dwgno login prefix style)
  "Common function to make rail fab drawings

Opens TEMPLATE, sets up title block and variables, and evals EVAL-FORM. Then it saves and
closes the drawing."
  (cond ((begins-with template "stair-rail")
         (setq prefix "SR"
               style "STAIR RAILS"))
        ((begins-with template "wall-rail")
         (setq prefix "WR"
               style "WALL RAILS"))
        ((begins-with template "guard-rail")
         (setq prefix "GR"
               style "GUARD RAILS"))
        ((begins-with template "handrail-returns")
         (setq prefix "HR"
               style "HANDRAIL RETURNS")))

  (setq oldAcadDoc acadDoc
        path (vla-get-Path acadDoc)
        jobno (xd-value "number" job-info)
        dwgno (calc-dwg-no jobno stairno prefix)
        login (getvar 'loginname)
        acadDoc (vla-Add (vla-get-Documents acadObj)
                         (strcat psc-src-dir "fab/" template ".dwt"))
        modelspace (vla-get-modelspace acaddoc))

  (vla-SaveAs acadDoc (strcat path "/" dwgno ".dwg"))

  (write-title-block job-info stairno login style)
  (eval eval-form)

  (vla-Regen acadDoc acAllViewports)
  (vla-Save acadDoc)
  (if (not *dev-mode*) (vla-Close acadDoc))
  (reset-doc))
(defun-r 'make-rail-fab-common)

  (defun-q g-post (cfg lng / cen qty obj con)
    "adds posts to rails created by topviewrails and dd-guard"
    (psc-include '("shapes/shape.lsp"))
    (setq
      dia (zerop (read (?? guard_sty rail_style)))
      lng (- lng (if dia 0.0 1.5))
      qty (1+ (rail-calc-n-mid-posts lng))
    ) ;_ setq
    (cond
      ( (= cfg 'dcl)
        (if (= 11 (guard-con-read lpost_con))
          (setq cen (list (+ 0.75 (/ lng 2.0))))
          (progn
            (setq cen (list (if dia 0.0 0.75)))
            (repeat qty
              (add-to-list 'cen (+ (last cen) (/ lng qty)))
            ) ;_ repeat
          ) ;_ progn
        ) ;_ if
      ) ;_ condif
      ( (= cfg 'prp)
        (setq cen (list -1.5625))
        (repeat qty
          (add-to-list 'cen (- (last cen) (/ lng qty)))
        ) ;_ repeat
      ) ;_ condif
      ( (= cfg 'par)
        (setq cen (list (+ (?? x 3.625) 0.75)))
        (if (or (= break "no") (= break tb))
          (repeat qty
            (add-to-list 'cen (+ (last cen) (/ lng qty)))
          ) ;_ repeat
        ) ;_if
      ) ;_ condif
    ) ;_ cond
    (set-active-layer "Rails")
    (foreach c cen
      (cond
        ( (not (= cfg 'dcl)) (list))
        ( (= c (car cen)) ; lpost
          (add-to-list 'obj
            (cond
              ( (= 30 (guard-con-read lpost_con))
                (vla-put-layer-r
                  (vla-addcircle blk (3dpt c 0.83) 2.0625)
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (member (guard-con-read lpost_con) '(41 42 48))
                (vla-put-layer-r
                  (if (= (guard-con-read lpost_con) 48)
                    (add-rectangle blk 3.0 20.0 (list (- c 0.75) -9.25))
                    (add-rectangle blk 3.0 6.0 (list (- c 0.75) -2.25))
                  ) ;_ if
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (member (guard-con-read lpost_con) '(7 47 51))
                (vla-put-layer-r
                  (add-rectangle blk 0.25 (- lpost_tab) (list (- c 0.125) 0))
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (= T) (list))
            ) ;_ cond
          ) ;_ add-to-list
        ) ;_ condif
        ( (= c (last cen)) ; rpost
          (add-to-list 'obj
            (cond
              ( (= 30 (guard-con-read rpost_con))
                (vla-put-layer-r
                  (vla-addcircle blk (3dpt c 0.83) 2.0625)
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (member (guard-con-read rpost_con) '(41 42 48))
                (vla-put-layer-r
                  (if (= (guard-con-read rpost_con) 48)
                    (add-rectangle blk 3.0 20.0 (list (- c 2.25) -9.25))
                    (add-rectangle blk 3.0 6.0 (list (- c 2.25) -2.25))
                  ) ;_ if
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (member (guard-con-read rpost_con) '(7 47 51))
                (vla-put-layer-r
                  (add-rectangle blk 0.25 (- rpost_tab) (list (- c 0.125) 0))
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (= T) (list))
            ) ;_ cond
          ) ;_ add-to-list
        ) ;_ condif
        ( (= T) ; mposts
          (add-to-list 'obj
            (cond
              ( (= 30 (guard-con-read mpost_con))
                (vla-put-layer-r
                  (vla-addcircle blk (3dpt c 0.83) 2.0625)
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (member (guard-con-read mpost_con) '(41 42 48))
                (vla-put-layer-r
                  (if (= (guard-con-read mpost_con) 48)
                    (add-rectangle blk 3.0 20.0 (list (- c 1.5) -9.25))
                    (add-rectangle blk 3.0 6.0 (list (- c 1.5) -2.25))
                  ) ;_ if
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (member (guard-con-read mpost_con) '(7 47 51))
                (vla-put-layer-r
                  (add-rectangle blk 0.25 (- mpost_tab) (list (- c 0.125) 0))
                  "No Plot"
                ) ;_ vla-put-layer-r
              ) ;_ condif
              ( (= T) (list))
            ) ;_ cond
          ) ;_ add-to-list
        ) ;_ condif
      ) ;_ cond
      ( (lambda (shp / col)
          (if (not (listp shp)) (setq shp (list shp)))
          (setq col
            (cond ((= 1 (length cen)) (if fab-p 3 256)) (fab-p 3) (T 253))
          ) ;_ setq
          (foreach tmp shp (add-to-list 'obj (vla-put-color-r tmp col)))
        ) ;_ lambda
        (cond
          ( (= cfg 'dcl)
            (cond
              (dia (vla-addcircle blk (3dpt c 0.83) 0.83))
              (fab-p (add-shape blk (list c 1.5) "HSS1.5x1.5x3/16"))
              (T (add-rectangle blk 1.5 1.5 (list (- c 0.75) 0)))
            ) ;_ cond
          ) ;_ condif
          ( (= cfg 'prp)
            (add-rectangle blockObj -1.5 -1.5 (list -3.625 c))
          ) ;_ condif
          ( (= cfg 'par)
            (add-rectangle blockObj 1.5 -1.5 (list (- c 0.75) 0))
          ) ;_ condif
        ) ;_ cond
      ) ;_ lambda-func
    ) ;_ foreach
    (if (and (= cfg 'par) (not tb) (not (= break "no")))
      (add-to-list 'cen (+ (?? x 3.625) wid))
    ) ;_ if
    (set-active-layer (cond (fab-p "Solid") (T "Rails")))
    (if (not dia) (g-fill cen))
    (setq obj (vl-remove-if 'null obj))
    (cond
      ( (not (= cfg 'dcl)) (mapcar 'mirror obj))
      ( (= (length cen) 1) (list))
      ( (not dia)
        (add-to-list 'obj (add-rectangle blk guard_lng 1.5 '(0 0)))
      ) ;_ condif
      ( (= T)
        (add-to-list 'obj
          (add-polyline blk '(0.0 0.0)
            (#! '((0 0 0) (#lng 0 1) (#lng 1.66 0) (0 1.66 1)))
            T
          ) ;_ add-polyline
        ) ;_ add-to-list
      ) ;_ condif
    ) ;_ cond
    obj
  ) ;_ defun g-post
  (defun-r 'g-post)

  (defun-q g-fill (lst / off ins)
    "adds appropriate fill to rails created by topviewrails and dd-guard"
    (setq off
      (if (and (= cfg 'par) (not (or (= break "no") (= break tb))))
        0.75
        1.5
      ) ;_ if
    ) ;_ setq
    (if (and (= cfg 'dcl) (= guard_kick "1"))
      (add-to-list 'obj
        (vla-put-layer-r
          (add-rectangle blk guard_lng -0.25 '(0 0))
          "No Plot"
        ) ;_ vla-put-layer-r
      ) ;_ add-to-list
    ) ;_ if
    (cond
      ( (= (length lst) 1)
        (mapcar
         '(lambda (ins / tmp)
            (setq tmp (add-rectangle blk (/ lng 2.0) 0.25 ins))
            (if fab-p (vla-rotate tmp (3dpt 0 0) pi))
          ) ;_ lambda
          (list '(0 0.625) (list (+ (car cen) 0.75) 0.625))
        ) ;_ mapcar
      ) ;_ condif
      ( (= (read (?? guard_sty rail_style)) 300)
        (apply-by-twos
         '(lambda (p1 p2 / lng tmp)
            (setq tmp
              (apply 'add-rectangle
                (cons
                  (?? blk blockObj)
                  (cond
                    ( (= cfg 'dcl)
                      (setq lng (- p2 p1 off))
                      (list lng 0.75 (list (+ p1 0.75) 0.375))
                    ) ;_ condif
                    ( (= cfg 'prp)
                      (setq lng (- p2 p1 -1.5))
                      (list -0.75 lng (list -4.0 (- p1 1.5)))
                    ) ;_ condif
                    ( (= cfg 'par)
                      (setq lng (- p2 p1 1.5))
                      (list lng 0.75 (list (+ p1 0.75) -1.125))
                    ) ;_ condif
                  ) ;_ cond
                ) ;_ cons
              ) ;_ add-rectangle
            ) ;_ setq
            (vla-put-layer tmp "HIDDEN")
            (vla-put-color tmp 253)
            (add-to-list 'obj tmp)
          ) ;_ lambda
          lst
        ) ;_ apply-by-twos
      ) ;_ condif
      ( (= (read (?? guard_sty rail_style)) 400)
        (apply-by-twos
         '(lambda (p1 p2 / x sp)
            (setq
              x (+ p1 0.75)
              sp (rail-calc-picket-spacing (- p2 p1 off))
            ) ;_ setq
            (while (< (+ x sp 0.5) (- p2 0.75))
              (add-to-list 'obj
                (vla-put-color-r
                  (add-rectangle (?? blk blockObj) 0.5 0.5
                    (cond
                      ( (= cfg 'dcl) (list (+ x sp) 0.5))
                      ( (= cfg 'prp) (list -4.625 (- (+ x sp 1.25))))
                      ( (= cfg 'par) (list (+ x sp) -1.0))
                    ) ;_ cond
                  ) ;_ add-rectangle
                  253
                ) ;_ vla-put-color-r
              ) ;_ add-to-list
              (setq x (+ x sp 0.5))
            ) ;_ while
          ) ;_ lambda
          (if (= cfg 'prp) (mapcar 'abs lst) lst)
        ) ;_ apply-by-twos
      ) ;_ condif
      ( (= T)
        ( (lambda (def / tmp)
            (setq tmp (add-polyline (?? blk blockObj) (car def) (cdr def) T))
            (if (= guard_hand "1") (vla-put-color tmp 253))
            (add-to-list 'obj tmp)
          ) ;_ lambda
          (cond
            ( (= cfg 'dcl)
              (list
                (list 0.25 0.0)
               '(0 0 0)
                (list (- guard_lng 0.5) 0 -0.301205)
                (list (- guard_lng 0.5) -1.66 0)
               '(0 -1.66 -0.301205)
              ) ;_ list
            ) ;_ condif
            ( (= cfg 'prp)
              (list
                (list -3.625 -1.8125)
               '(0 0 0)
                (list 0 (- (- wid 0.5)) 0.301205)
                (list 1.66 (- (- wid 0.5)) 0)
               '(1.66 0 0.301205)
              ) ;_ list
            ) ;_ condif
            ( (= cfg 'par)
              (apply
               '(lambda (l b)
                  (list
                    (list (+ (?? x 3.625) 0.25) 0.0)
                   '(0 0 0)
                    (list (- wid l) 0 b)
                    (list (- wid l) 1.66 0)
                    (list 0 1.66 0.301205)
                  ) ;_ list
                ) ;_ lambda
                (if (or (= break "no") (= break tb))
                 '(0.5 0.301205)
                 '(0.25 0.0)
                ) ;_ if
              ) ;_ apply
            ) ;_ condif
          ) ;_ cond
        ) ;_ lambda-func
      ) ;_ condif
    ) ;_ cond
  ) ;_ defun g-fill
  (defun-r 'g-fill)

  (defun-q guard-con-read (str)
    "Returns intiger from guard connection string by stripping X"
    (setq str (to-string (read str)))
    (cond
      ( (numberp (read str)) (read str))
      ( (begins-with str "X") (read (substr str 2 2)))
      ( (ends-with str "X") (read (substr str 1 2)))
      ( (= T) "")
      ( (= T) (error "Invalid guard connection value"))
    ) ;_ cond
  ) ;_ defun guard-con-read
  (defun-r 'guard-con-read)

  (defun-q all-X12 ( / lst)
    "returns T if all guard post connections are X12 BTWN. WIDE"
    (setq lst (list lpost_con mpost_con rpost_con))
    (setq lst (cons 'X12 (mapcar 'read lst)))
    (apply '= lst)
  ) ;_ defun-q all-X12
  (defun-r 'all-X12)
