  (set-file-docstring "utilities for AISC shapes")

  (psc-include '("shapes/line.lsp"))

  (defun-q shape-def (name tagged-p / filename file line headers def)
    "return the defining properties of a shape, taken from the AISC shapes database
if tagged-p is T, include data headers"
  (setq name (uppercase name))
  ;; HSS and PL do not need the database
  (cond ((begins-with name "HSS")
         (setq def (mapcar 'distof (strsplit (substr name 4) "X"))
               headers '("dpt" "wid" "thick")))
        ((begins-with name "PL")
         (setq def (mapcar 'distof (strsplit (substr name 3) "X"))
               headers '("wid" "dpt")))
        (T
         (setq filename (strcat psc-src-dir
                                "shapes/"
                                (cond ((begins-with name "W") "wf")
                                      ((begins-with name "L") "angles")
                                      ((or (begins-with name "C")
                                           (begins-with name "MC")) "channels")
                                      (T (error (strcat "shape.lsp: Shape not found: " name))))
                                ".csv")
               file (open filename "r")
               headers (cdr (strsplit (read-line file) ",")))
         (while (setq line (read-line file))
           (if (begins-with line name)
               (setq def (mapcar 'distof (cdr (strsplit line ",")))
                     name (car (strsplit line ",")))))
         (close file)))
  (if def
      (if tagged-p
          (append
           (list (list "name" name))
           (mapcar 'list headers def))
        def)
    (error (strcat "shape.lsp: Shape not found: " name)))
  ) ;_ defun-q shape-def
  (defun-r 'shape-def)

  (defun-q add-shape (container ins name / def f)
    "add a shape polyline to a container"
    (setq
      name (uppercase name)
      def (shape-def name nil)
      f (cond
          ( (begins-with name "W") 'wf-pts)
          ( (begins-with name "L") 'angle-pts)
          ( (begins-with name "C") 'c-channel-pts)
          ( (begins-with name "M") 'mc-channel-pts)
          ( (begins-with name "PL") 'pl-pts)
          ( (begins-with name "H") 'hss-pts)
        ) ;_ cond
    ) ;_ setq
    (if (begins-with name "H")
      (mapcar
       '(lambda (lst) (add-polyline container ins lst T))
        (apply f def)
      ) ;_ mapcar
      (add-polyline container ins (apply f def) T)
    ) ;_ if
  ) ;_ defun-q add-shape
  (defun-r 'add-shape)

  (defun-q angle-pts (long short thick k / crad trad b90)
    "return list of 2D pline vertices for an angle based on leg lengths, thickness, and distance to fillet"
    (setq
      ;; corner radius
      crad (- k thick)
      ;; toe radius
      trad (/ thick 2.0)
      ;; 90 degree bulge
      b90 (- (calc-bulge (dtr 90)))
    ) ;_ setq
    (list
      (list 0 0 0)
      (list long 0 0)
      (list long (- trad) b90)
      (list (- long trad) (- thick) 0)
      (list k (- thick) (- b90))
      (list thick (- (+ thick crad)) 0)
      (list thick (- (- short trad)) b90)
      (list trad (- short) 0)
      (list 0 (- short) 0)
    ) ;_ list
  ) ;_ defun-q angle-pts
  (defun-r 'angle-pts)

  (defun-q c-channel-pts (depth bf tw tf / slope)
    "set correct slope for C channels"
    (setq slope (/ 2.0 12.0))
    (channel-pts depth bf tw tf)
  ) ;_ defun-q c-channel-pts

  (defun-q mc-channel-pts (depth bf tw tf / slope)
    "set correct slope for MC channels"
    (setq slope (/ 1.0 16.0))
    (channel-pts depth bf tw tf)
  ) ;_ defun-q mc-channel-pts
  (defun-r 'mc-channel-pts)

  (defun-q channel-pts (depth bf tw tf / ang bulge wrad trad topflange botflange instfill inswfill)
    "return list of 2D pline vertices for a channel based on depth, width, web thickness, and flange (avg) thickness"
    (setq
      ;; web radius (approximation)
      wrad (/ depth 25.0)
      ;; toe radius (approximation)
      trad (/ wrad 2.0)
      ;; flange slope
      ang (atan slope)
      ;; bulge for fillets
      bulge (- (calc-bulge (- (dtr 90) ang)))
      ;; vertical line between toe edge and inside of web
      mid (+ tw (/ (- bf tw) 2.0))
      ;; top sloping line (slope-int format)
      topflange (pt-slope slope mid (- tf))
      ;; bottom sloping line
      botflange (pt-slope (- slope) mid (- (- depth tf)))
      ;; inside toe fillet (x value)
      instfill (+ (- bf trad) (* trad (sin ang)))
      ;; inside web fillet (x value)
      inswfill (- (+ tw wrad) (* wrad (sin ang)))
    ) ;_ setq
    (list
      ;; top
      (list 0 0 0)
      (list bf 0 0)
      (list bf (line-val (line-offset topflange trad) instfill) bulge)
      (list instfill (line-val topflange instfill) 0)
      (list inswfill (line-val topflange inswfill) (- bulge))
      (list tw (line-val (line-offset topflange (- wrad)) (+ tw wrad)) 0)
      ;; bottom
      (list tw (line-val (line-offset botflange wrad) inswfill) (- bulge))
      (list inswfill (line-val botflange inswfill) 0)
      (list instfill (line-val botflange instfill) bulge)
      (list bf (line-val (line-offset botflange (- trad)) bf) 0)
      (list bf (- depth) 0)
      (list 0 (- depth) 0)
    ) ;_ list
  ) ;_ defun-q channel-pts
  (defun-r 'channel-pts)

  (defun-q hss-pts (d w thick / )
    "return a 2-item list whose elements are each lists of of 2D pline vertices for HSS based on depth, width, and thickness

NOTE this returns differently from the rest because it contains 2 plines"
    (defun rounded-rect (d w rad)
      (setq
        ;; horizontal edges
        redge (/ w 2.0)
        ledge (- redge)
        ;; horizontal k
        rk (- redge rad)
        lk (- rk)
        ;; vertical k
        topk (- rad)
        botk (- (- d rad))
        bulge (- (calc-bulge (dtr 90)))
        d (- d)
      ) ;_ setq
      (list
       (list lk 0 0)
       (list rk 0 bulge)
       (list redge topk 0)
       (list redge botk bulge)
       (list rk d 0)
       (list lk d bulge)
       (list ledge botk 0)
       (list ledge topk bulge)
      ) ;_ list
    ) ;_ defun rounded-rect
    ( (lambda(/ thick2 inner-rad)
        (setq
          thick2 (* 2 thick)
          inner-rad (max 0.125 thick)
        ) ;_ setq
        (list
          (rounded-rect d w (+ inner-rad thick))
          (mapcar
           '(lambda (x) (list (car x) (- (cadr x) thick) (caddr x)))
            (rounded-rect (- d thick2) (- w thick2) inner-rad)
          ) ;_ mapcar
        ) ;_ list
    ) ) ;_ lambda
  ) ;_ defun-q hss-pts
  (defun-r 'hss-pts)

  (defun-q hss-pts-ctr (d w thick)
    "like hss-pts but cenetered"
    (mapcar
     '(lambda (x)
        (mapcar
         '(lambda (x) (list (car x) (+ (/ d 2.0) (cadr x)) (caddr x)))
          x
        ) ;_ mapcar
      ) ;_ lambda
      (hss-pts d w thick)
    ) ;_ mapcar
  ) ;_ defun-q hss-pts-ctr
  (defun-r 'hss-pts-ctr)

  (defun-q wf-pts (d bf tw tf k / rad bulge rk1 lk1 redge ledge rweb lweb tflange bflange topk botk)
    "return list of 2D pline vertices for a wide-flange beam based on depth, width, web thickness, flange thickness, and distance to fillet"
    (setq
      ;; radius
      rad (- k tf)
      bulge (calc-bulge (dtr 90))
      ;; horizontal k
      rk1 (+ rad (/ tw 2.0))
      lk1 (- rk1)
      ;; horizontal edges
      redge (/ bf 2.0)
      ledge (- redge)
      ;; web faces
      rweb (/ tw 2.0)
      lweb (- rweb)
      ;; insides of flanges
      tflange (- tf)
      bflange (- (- d tf))
      ;; top/bottom of webs
      topk (- k)
      botk (- (- d k))
    ) ;_ setq
    (list
      ;; right
      (list ledge 0 0)
      (list redge 0 0)
      (list redge tflange 0)
      (list rk1 tflange bulge)
      (list rweb topk 0)
      (list rweb botk bulge)
      (list rk1 bflange 0)
      (list redge bflange 0)
      (list redge (- d) 0)
      ;; left
      (list ledge (- d) 0)
      (list ledge bflange 0)
      (list lk1 bflange bulge)
      (list lweb botk 0)
      (list lweb topk bulge)
      (list lk1 tflange 0)
      (list ledge tflange 0)
    ) ;_ list
  ) ;_ defun-q wf-pts
  (defun-r 'wf-pts)

  (defun-q wf-pts-ctr (d bf tw tf k)
    "like wf-pts, but centered"
    (mapcar
     '(lambda (x) (list (car x) (+ (/ d 2.0) (cadr x)) (caddr x)))
      (wf-pts d bf tw tf k)
    ) ;_ mapcar
  ) ;_ defun-q wf-pts-ctr
  (defun-r 'wf-pts-ctr)

  (defun-q pl-pts (h w)
    "similar to add-rectangle here for consistent batch processing of connections @ 3d"
    (mapcar
     '(lambda (l) (mapcar 'float l))
      (list (list 0 (- w) 0) (list h (- w) 0)  (list h 0 0) '(0 0 0))
    ) ;_ mapcar
  ) ;_ defun-q pl-pts
  (defun-r 'pl-pts)

  (defun-q shape-pts (shp)
    "returns points list from shape def allows interaction with raw points list"
    (setq shp (uppercase shp))
    (apply
      (cond
        ( (begins-with shp "W") 'wf-pts)
        ( (begins-with shp "L") 'angle-pts)
        ( (begins-with shp "C") 'c-channel-pts)
        ( (begins-with shp "M") 'mc-channel-pts)
        ( (begins-with shp "PL") 'pl-pts)
        ( (begins-with shp "H") 'hss-pts)
      ) ;_ cond
      (shape-def shp nil)
    ) ;_ apply
  ) ;_ defun-q shape-pts
  (defun-r 'shape-pts)

(defun-q drawshape--draw-section ()
  "draw a section block in model space"
  (expect-with-data "drawshape--draw-section")
  (setq profile (uppercase (substr name 1 1))
        def (mapcar 'cadr (cdr data))
        blockname (string-subst-all "_" "/" name))
  (cond
   ;; angles
   ((= profile "L")
    (setq plines (apply 'angle-pts def)))
   ;; channels
   ((= profile "C")
    (setq plines (apply 'c-channel-pts def)))
   ((= profile "M")
    (setq plines (apply 'mc-channel-pts def)))
   ;; wide-flange & HSS
   ((member profile '("W" "H"))
    (setq ctr (getdkword '(0 "Top Center") '*drawshape-insert-on* "Insert on")
          ctr (= ctr "Center")
          blockname (strcat blockname (if ctr "_CTR" "_TOP"))
          plines
          (apply (if (= profile "W")
                     (if ctr
                         'wf-pts-ctr
                       'wf-pts)
                   (if ctr
                       'hss-pts-ctr
                     'hss-pts))
                 def))))
  (define-block-maybe blockname (if (= profile "H") plines (list plines)))
  ;; insert block
  (setvar 'cmdecho 0)
  (command "-insert" blockname pause "" ""
           ;; rotate angles and centered things
           (if (or (= profile "L")
                   ctr)
               pause
             ""))
  (assign-xdata (entlast) "Shape" (format-xdata data))
  (cond
   ;; optionally mirror unequal angles
   ((and (= profile "L")
         (not (= leg1 leg2))
         (not (initget "Y N"))
         (member (getkword "\nMirror? [Y/N] ") '("y" "Y")))
    (setq ent (entget (entlast))
          ins (cdr (assoc 10 ent))
          ang (cdr (assoc 50 ent)))
    (osnap-off)
    (command "mirror"
             (entlast)
             ""
             ins
             (polar ins ang 1)
             "Y")
    (osnap-on))
   ;; optionally mirror channels
   ((and (member profile '("C" "M"))
         (not (initget "Y N"))
         (member (getkword "\nMirror? [Y/N] ") '("y" "Y")))
    (setq ent (entget (entlast))
          ins (cdr (assoc 10 ent)))
    (osnap-off)
    (command "mirror"
             (entlast)
             ""
             ins
             (list (car ins)
                   (+ 1 (cadr ins))
                   0)
             "Y")
    (osnap-on)))
  (setvar 'cmdecho 1)
  (princ)
) ;_ defun-q drawshape--draw-section
(defun-r 'drawshape--draw-section)

(defun-q drawshape--draw-plan (/ pt1 endpts profile symmetrical-p)
  "draw a shapes plan view in model space"
  (expect-with-data "drawshape--draw-plan")
  (setq profile (substr name 1 1))
  (if (= profile "L")
      (if (= leg1 leg2)
          (setq wid leg1)
          (if (= (getdkword '(0 "H V") '*drawshape-l-orient* "Long leg...") "H")
              (setq wid leg1)
              (setq wid leg2))))
  (setq pt1 (getpoint "\nPoint 1: ")
        endpts (list pt1 (getpoint pt1 "\nPoint 2: "))
        symmetrical-p (member profile '("W" "H")))
  (if symmetrical-p
      (setq wid (/ wid 2.0)))
  (setq thick (cond ((= profile "W") (/ web 2.0))
                    ((member profile '("C" "M")) web)
                    ((= profile "H") (- wid thick))
                    (T thick)))
  ((lambda(/ pl)
     (setq pl (add-polyline modelSpace '(0 0) endpts nil))
     (if symmetrical-p
         (progn
           (vla-put-Layer (offset&close pl thick T nil) "HIDDEN")
           (dualoffset&delete pl wid nil))
         (progn
           (vla-put-Layer (move-to-bottom (offset-polyline pl thick) modelSpace) "HIDDEN")
           (offset&close pl wid nil nil)))))
) ;_ defun-q drawshape--draw-plan
(defun-r 'drawshape--draw-plan)

(defun-q drawshape--draw-elevation (/ profile pt1)
  "draw a shapes elevation view in model space"
  (expect-with-data "drawshape--draw-elevation")
  (setq profile (substr name 1 1))
  (if (= profile "L")
      (if (= leg1 leg2)
          (setq wid leg1)
          (if (= (getdkword '(0 "H V") '*drawshape-l-orient* "Long leg...") "H")
              (setq dpt leg2)
              (setq dpt leg1))))
  (if (member profile '("W" "C" "M")) (setq thick flg))
  (setq pt1 (getpoint "\nPoint 1: ")
        endpts (vl-sort (list pt1 (getpoint pt1 "\nPoint 2: ")) (sort-by 'car '<)))
  ((lambda(/ pl)
     (setq pl (add-polyline modelSpace '(0 0) endpts nil))
     (offset-polyline pl thick)
     (if (/= profile "L") (offset-polyline pl (- dpt thick)))
     (offset&close pl dpt nil nil)))
) ;_ defun-q drawshape--draw-elevation
(defun-r 'drawshape--draw-elevation)

(defun-q shapelist-init (profile
                       /
                       shapes
                       file
                       line
                       save-action
                       ;; variables defined here for dialog-init
                       defaults-file
                       popup-keys
                       dist-keys
                       int-keys
                       toggle-keys
                       plain-string-keys
                       default-actions
                       dialog-ties
                       sub-dialogs
                       ;; hook
                       dialog-after-load
                       ;; variables included here for scope,
                       ;; but not defined here directly
                       extra-keys
                       tile-actions
                       )

  "DCL initialization for drawshape function"
  (defun save-action ( / )
    (setq data (nth (atoi (get_tile "box")) shapes))
    (done_dialog 0))

  (defun dialog-after-load ( / )
    (dialog-popup-list-set "box" shapes)
    (action_tile "ok" "(save-action)"))

  (setq file (strcat psc-src-dir
                     "shapes/"
                     (cond ((= profile "W") "wf.csv")
                           ((= profile "C") "channels.csv")
                           ((= profile "L") "angles.csv")))
        file (open file "r"))
  (read-line file)
  (while (setq line (read-line file))
    (add-to-list 'shapes (car (strsplit line ","))))
  (close file)
  (dialog-init "shapes/shapelist.dcl" "shapelist" (mapcar '(lambda(x) (list x x)) shapes))
) ;_ defun-q shapelist-init
(defun-r 'shapelist-init)
