(set-file-docstring
 "Functions defining stair connection details for submittal and fabrication stages

Detail definitions are comprised of the following functions:

detail-side-X        adds side view connection material to blockObj
detail-top-X         adds top view connection material to blockObj
material-X           returns a list of material and hardware

Functions in this file are helpers for detail definitions to
use. Detail definitions themselves are in the other files in this
directory, arranged by series.")

(psc-include (append '("shapes/shape.lsp")
                     (mapcar '(lambda(x) (strcat "stair/details/" x))
                             (vl-directory-files (strcat psc-src-dir "stair/details") "*.lsp" 1))))

(defun-q detail-func (num top-or-side)
  "If a detail function for detail NUM exists, calls it and returns its value"
  (declare (vars (num str)
                 (top-or-side str)))
  (attempt-apply (read (strcat "detail-" top-or-side "-" num))))
(defun-r 'detail-func)

(defun-q material-func (num)
  "If a material function for detail NUM exists, calls it and returns its value"
  (declare (vars (num str)))
  (attempt-apply (read (strcat "material-" num))))
(defun-r 'material-func)

(defun-q angle-desc (size)
  "Translates an AISC angle size to a PSC material description"
  (declare (vars (size str)))
  (strcat "ANGLE " (strjoin (strsplit (substr size 2) "x") " x ")))
(defun-r 'angle-desc)

(defun-q add-line-btwn (x)
  "Adds a line between stringers in top view at x-value X"
  (declare (vars (x nil (numberp x))))
  (add-polyline blockObj
                (list x 0 0)
                (list '(0 0 0)
                      (list 0 btwn_stringers 0))
                nil))
(defun-r 'add-line-btwn)

(defun-q detail-side-add-dim-right (pt2 space rev / first-space x1 y1 pt1)
  "Helper for adding vertical dims to top connection details

PT2 is the second dim point. The first point is handled automatically. SPACE is the number
of dimspace jumps to make to the right. REV being non-nil means to draw the dimension in
the reverse direction."
  (declare (vars (pt2 variant)
                 (space nil (numberp space))
                 (dimspace nil (numberp dimspace))
                 (side str)
                 (xtra-tex nil (numberp xtra-tex)))
           with-data)
  (setq x1 (if (= (sidekey "nosing_top") 0)
               (sidekey "string_xtra_tex")
               (?f top_pan_ext))
        y1 (if (and (member (substr tread_style 1 1) '("1" "4"))
                    (/= stn_style "None"))
               (- stn_thick)
               0)
        pt1 (3dpt x1 y1)
        first-space (+ xtra-tex
                       (* (cond ((< xtra-tex dimspace) 3)
                                ((< xtra-tex (* 2 dimspace)) 2)
                                (T 1))
                          dimspace)))
  (if rev (swap! 'pt1 'pt2))
  (vla-AddDimRotated blockObj
                     pt1
                     pt2
                     (3dpt (+ first-space (* space dimspace)) 0)
                     (dtr -90)))
(defun-r 'detail-side-add-dim-right)

(defun-q detail-side-add-dim-left (pt2 space rev / x1 pt1)
  "Same as %detail-side-add-dim-right but for bottom connections

See same for arguments"
  (declare (vars (pt2 variant)
                 (space nil (numberp space))
                 (dimspace nil (numberp dimspace))
                 (left-dim-line nil (numberp left-dim-line))
                 (side str))
           with-data)
  (setq x1 (- (if (= starter_p "1")
                  (sidekey "string_xtra_bex")
                  (?f bot_pan_ext)))
        pt1 (3dpt x1 0))
  (if rev (swap! 'pt1 'pt2))
  (vla-AddDimRotated blockObj
                     pt1
                     pt2
                     (3dpt (- left-dim-line (* space dimspace)) 0)
                     (dtr -90)))
(defun-r 'detail-side-add-dim-left)

(defun-q bolts-top-right (bolt-qty / bot-bolt top-bolt right-dim-line)
  "Adds bolt centerlines and dimensions on right side of top view"
  (declare with-data
           (vars (bolt-qty int)
                 (dimspace nil (numberp dimspace))))
  (setq bot-bolt (- top_hbolt r-string_wid)
        top-bolt (- (+ btwn_stringers l-string_wid) top_hbolt)
        right-dim-line (+ (max r-string_xtra_tex l-string_xtra_tex) (* 2 dimspace)))

  ;; centerlines
  (add-bolt-centerline (list 0 bot-bolt 0) 'h)
  (add-bolt-centerline (list 0 top-bolt 0) 'h)
  (if (= bolt-qty 4)
      (progn
        (add-bolt-centerline (list 0 (+ bot-bolt top_hbolt2) 0) 'h)
        (add-bolt-centerline (list 0 (- top-bolt top_hbolt2) 0) 'h)))
  ;; dims
  (set-active-layer "DIM")
  (vla-AddDimRotated blockObj
                     (3dpt 1 bot-bolt)
                     (3dpt r-string_xtra_tex (- r-string_wid))
                     (3dpt right-dim-line 0)
                     (dtr 90))
  (vla-AddDimRotated blockObj
                     (3dpt 1 top-bolt)
                     (3dpt l-string_xtra_tex (+ btwn_stringers l-string_wid))
                     (3dpt right-dim-line 0)
                     (dtr 90))
  (if (= bolt-qty 2)
      (vla-AddDimRotated blockObj
                         (3dpt 1 bot-bolt)
                         (3dpt 1 top-bolt)
                         (3dpt right-dim-line 0)
                         (dtr 90))
    (progn
      (setq right-dim-line (+ right-dim-line dimspace))
      (vla-AddDimRotated blockObj
                         (3dpt 1 (+ bot-bolt top_hbolt2))
                         (3dpt 1 (- top-bolt top_hbolt2))
                         (3dpt right-dim-line 0)
                         (dtr 90))
      (vla-AddDimRotated blockObj
                         (3dpt 1 (+ bot-bolt top_hbolt2))
                         (3dpt 1 bot-bolt)
                         (3dpt right-dim-line 0)
                         (dtr 90))
      (vla-AddDimRotated blockObj
                         (3dpt 1 (- top-bolt top_hbolt2))
                         (3dpt 1 top-bolt)
                         (3dpt right-dim-line 0)
                         (dtr 90)))))
(defun-r 'bolts-top-right)

(defun-q bolts-top-left (bolt-qty / bot-bolt top-bolt left-dim-line)
  "Adds bolt centerlines and dimensions on left side of top view"
  (declare with-data
           (vars (bolt-qty int)
                 (dimspace nil (numberp dimspace))))
  (setq bot-bolt (- bot_hbolt r-string_wid)
        top-bolt (- width r-string_wid bot_hbolt)
        left-dim-line (* (if (= bolt-qty 2) -3 -2) dimspace))

    ;; centerlines
  (add-bolt-centerline (list 0 bot-bolt 0) 'h)
  (add-bolt-centerline (list 0 top-bolt 0) 'h)
  (if (= bolt-qty 4)
      (progn
        (add-bolt-centerline (list 0 (+ bot-bolt bot_hbolt2) 0) 'h)
        (add-bolt-centerline (list 0 (- top-bolt bot_hbolt2) 0) 'h)))

  ;; dims
  (set-active-layer "DIM")
  (vla-AddDimRotated blockObj
                     (3dpt -1 bot-bolt)
                     (3dpt (- r-string_xtra_bex) (- r-string_wid))
                     (3dpt left-dim-line 0)
                     (dtr 90))
  (vla-AddDimRotated blockObj
                     (3dpt -1 top-bolt)
                     (3dpt (- l-string_xtra_bex) (- width r-string_wid))
                     (3dpt left-dim-line 0)
                     (dtr 90))
  (if (= bolt-qty 2)
      (vla-AddDimRotated blockObj
                         (3dpt -1 bot-bolt)
                         (3dpt -1 top-bolt)
                         (3dpt left-dim-line 0)
                         (dtr 90))
      (progn
        (vla-AddDimRotated blockObj
                         (3dpt -1 bot-bolt)
                         (3dpt -1 (+ bot-bolt bot_hbolt2))
                         (3dpt left-dim-line 0)
                         (dtr 90))
        (vla-AddDimRotated blockObj
                         (3dpt -1 top-bolt)
                         (3dpt -1 (- top-bolt bot_hbolt2))
                         (3dpt left-dim-line 0)
                         (dtr 90))
        (setq left-dim-line (- left-dim-line dimspace))
        (vla-AddDimRotated blockObj
                         (3dpt -1 (+ bot-bolt bot_hbolt2))
                         (3dpt -1 (- top-bolt bot_hbolt2))
                         (3dpt left-dim-line 0)
                         (dtr 90)))))
(defun-r 'bolts-top-left)

(defun-q bot-string-line-from-top (side)
  "Returns the line along the bottom of the stringer w/ top connection point as origin"
  (declare with-data
           (vars (side str)))
  (line-offset
   (pt-slope (/ rise_hgt tread_depth)
             (- 0 top_cdist top_ext)
      0)
   (- (- (sidekey "string_dpt")
         (sidekey "nosing_mid")))))
(defun-r 'bot-string-line-from-top)

(defun-q bot-string-line-from-bot (side)
  "Returns the line along the bottom of the stringer w/ bottom connection point as origin"
  (declare with-data
           (vars (side str)))
  (line-offset
   (pt-slope (/ rise_hgt tread_depth)
             (+ bot_cdist bot_ext)
             rise_hgt)
   (- (- (sidekey "string_dpt")
         (sidekey "nosing_mid")))))
(defun-r 'bot-string-line-from-bot)

(defun-q bot-triangle-gusset-p (ang-y ang-hlen side add-p / string-line ret)
  "Returns nil if there is no bottom triangle gusset required

If there is, and ADD-P is non-nil, adds it to blockObj"
  (declare with-data
           (vars (ang-y nil (numberp ang-y))
                 (ang-hlen nil (numberp ang-hlen))
                 (side str)))
  (setq string-line (bot-string-line-from-bot side))
  (if (and (setq ret (inters (list 0 ang-y)
                             (list ang-hlen ang-y)
                             (list 0 (line-val string-line 0))
                             (list ang-hlen (line-val string-line ang-hlen))))
           add-p)
      (add-polyline blockObj
                    (list 0 (line-val string-line 0) 0)
                    (list '(0 0 0)
                          '(0 -4.125 0)
                          (line-int (pt-slope (/ rise_hgt tread_depth) 0 0)
                                    (pt-slope (sqrt 3) 0 -4.125)
                                    T))
                    T))
  ret)
(defun-r 'bot-triangle-gusset-p)

(defun-q add-top-triangle-gusset (side / string-line hgt ret)
  "Inserts a top triangle gusset"
  (declare with-data
           (vars (side str)))
  (setq hgt (- (* 6 (tan (dtr pitch))))
        string-line (bot-string-line-from-top side))
  (add-polyline blockObj
                (list 0 (line-val string-line 0) 0)
                (list '(0 0 0)
                      (list 0 hgt 0)
                      (line-int (pt-slope (/ rise_hgt tread_depth) 0 0)
                                (list 0 hgt)
                                T))
                T)
  ret)
(defun-r 'add-top-triangle-gusset)

(defun-q get-bot-kick-dist ()
  "Returns the bottom kick distance"
  (declare with-data)
  (cond ((begins-with tread_style "1") (+ 0.75 pan_thick))
        ((begins-with tread_style "2") (+ 0.5 (* (if (= stn_style "None") 2 1) pan_thick)))
        ((begins-with tread_style "4") (+ 0.625 pan_thick))
        ((begins-with tread_style "6") 1)))
(defun-r 'get-bot-kick-dist)

(defun-q get-end-braced-len (tb / series)
  "Returns the extent of the connection angle supporting the pan at stringer end"
  (declare with-data
           (vars (tb str)))
  (setq series (tbkey "conn_ser"))
  (cond ((member series '("2" "5"))
         (if (begins-with tread_style "2") 4 2))
        ((= series "3") 2.5)
        ((member series '("4" "6"))
         (distof (cadr (strsplit (x4x-inner-angle-size tb) "x"))))
        ((= series "X")
         (if (begins-with tread_style "1") 1.0 0.0))))
(defun-r 'get-end-braced-len)

(defun-q top-chkr-nosing-braced-len ()
  "Returns the braced length at nosing of top extensions"
  (declare with-data)
  (cond ((= stn_style "None") 0.8125)
        ((= stn_style "STNS") 2.5625)
        ((= stn_style "STNRS") 3.0625)))
(defun-r 'top-chkr-nosing-braced-len)

(defun-q add-qt (x y wid)
  "Adds a single piece of Quiet Tread"
  (declare with-data
           (vars (x nil (numberp x))
                 (y nil (numberp y))
                 (wid nil (numberp wid))))
  (setq wid (/ wid 2.0))
  (vla-put-Layer-r
   (add-polyline
    blockObj
    (list x y 0)
    (list
     (list (- wid) 0 0)
     (list (- wid) -0.125 0)
     (list wid -0.125 0)
     (list wid 0 0))
    T)
   "Stair Quiet Tread"))
(defun-r 'add-qt)

(defun-q add-bot-ext-qt (/ end-braced-len unbraced-len)
  "Adds any required Quiet Tread at bottom extensions"
  (declare with-data)
  (setq end-braced-len (cond ((member bot_conn_ser '("2" "5")) 4)
                             ((= bot_conn_ser "3") 2.5)
                             ((member bot_conn_ser '("4" "6"))
                              (distof (cadr (strsplit (x4x-inner-angle-size "bot") "x")))))
        unbraced-len (- (+ bot_cdist bot_ext (get-bot-kick-dist)) end-braced-len))
  (if (>= unbraced-len 3)
      (add-qt (+ end-braced-len (/ unbraced-len 2.0))
              (- pan_thick) 
              (cond ((>= unbraced-len 8) 7.5)
                    ((>= unbraced-len 5) 4)
                    (T 2)))))
(defun-r 'add-bot-ext-qt)

(defun-q add-top-ext-qt (/ end-braced-len unbraced-len)
  "Adds any required Quiet Tread at top extensions"
  (declare with-data)
  (setq end-braced-len (cond ((member top_conn_ser '("2" "5")) 4)
                             ((= top_conn_ser "3") 2.5)
                             ((member top_conn_ser '("4" "6"))
                              (distof (cadr (strsplit ang-size "x")))))
        unbraced-len (- (+ top_cdist top_ext)
                        end-braced-len
                        (top-chkr-nosing-braced-len)))
  (if (>= unbraced-len 3)
      (add-qt (- (+ end-braced-len (/ unbraced-len 2.0)))
              (- pan_thick)
              (cond ((>= unbraced-len 8) 7.5)
                    ((>= unbraced-len 5) 4)
                    (T 2)))))
(defun-r 'add-top-ext-qt)

(defun-q haunch-stringer-callout (tb / dir pts)
  "Adds stringer callout to haunch side views"
  (declare with-data
           (vars (tb str)))
  (setq dir (if (= tb "bot") -1 1)
        pts (#! '((0 0 0)
                  (0 4 0)
                  (#(* dir 2) 4 0))))
  (add-mleader blockObj
               (list (- dir) (sidekey (strcat "nosing_" tb)) 0)
               pts
               (xd-value (strcat side "-string-" tb) marks)))
(defun-r 'haunch-stringer-callout)

;;;;;;;;;;;;;;;;;;
;; conc haunches
;;;;;;;;;;;;;;;;;;

(defun-q top-conc-haunch-material (/)
  "Returns the material list for a top concrete haunch"
  (declare with-data)
  (#! '(
        (("id" "top-haunch-l-btwn")
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("qty" 1)
         ("length" #(frac (- btwn_stringers 0.25))))
        (("id" "top-haunch-l-sides")
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("qty" 2)
         ("length" #(frac (- top_ext 4))))
        (("id" "top-haunch-decking")
         ("desc" "BR DECKING 1 1/2 x 36")
         ("qty" #(ceil (/ (- top_ext 4) 36.0)))
         ("length" #(frac (- btwn_stringers 0.25))))
        )))
(defun-r 'top-conc-haunch-material)

(defun-q top-conc-haunch-side (fab-p / ang-len ang-x-offset clayer)
  "Draws a top concrete haunch side view"
  (declare with-data)
  (setq ang-x-offset (- (- (get-end-braced-len "top")) 0.125)
        ang-len (- (- (+ top_ext top_cdist) (+ (- ang-x-offset) 2.625))))
  (add-rectangle blockObj
                 ang-len
                 -2
                 (list ang-x-offset -3.5))
  (add-polyline blockObj
                (list ang-x-offset -3.75 0)
                (list '(0 0 0)
                      (list ang-len 0 0))
                nil)
  (add-decking blockObj
               (list (+ ang-x-offset 0.375) -3.5 0)
               (list (+ ang-x-offset ang-len -0.375) -3.5 0))
  (mirror&delete (add-shape blockObj (list (+ ang-len ang-x-offset -0.125) -3.5) "L2x2x1/4")
                 (list (+ ang-len ang-x-offset -0.125) 0)
                 'y)
  (if fab-p
      (progn
        (setq clayer (vla-getvar 'clayer))
        (set-active-layer "DIM")

        (haunch-stringer-callout "top")

        (add-mleader blockObj
                     (list (1- ang-x-offset) -3.5 0)
                     '((0 0 0)
                       (0 -14 0)
                       (-1 -14 0))
                     (strcat (xd-value "top-haunch-l-sides" marks) "\n"
                             (xd-value "top-haunch-decking" marks)))

        (add-mleader blockObj
                     (list (- 0 top_ext -1.25) -5.5 0)
                     '((0 0 0)
                       (0 -8 0)
                       (1 -8 0))
                     (xd-value "top-haunch-l-btwn" marks))

        (set-active-layer clayer))))
(defun-r 'top-conc-haunch-side)

(defun-q top-conc-haunch-top (/ clayer ang-wid l-btwn-ins l-side-ins-x l-side-len)
  "Draws a top concrete haunch top view"
  (declare with-data)
  (setq clayer (vla-getvar 'clayer)
        ang-wid (get-end-braced-len "top")
        l-btwn-ins-x (- 0 top_ext -1.25)
        l-side-ins-x (- 0 ang-wid 0.125)
        l-side-len (- (- top_ext 1.5 ang-wid)))
  (set-active-layer "HIDDEN")

  ;; back
  (add-rectangle blockObj -2 btwn_stringers (list l-btwn-ins-x 0 0))
  (add-line-btwn (- l-btwn-ins-x 0.25))
  ;; right side
  (add-rectangle blockObj l-side-len 2 (list l-side-ins-x 0 0))
  (add-polyline blockObj
                (list l-side-ins-x 0.25 0)
                (list '(0 0 0) (list l-side-len 0 0))
                nil)
  ;; left side
  (add-rectangle blockObj l-side-len -2 (list l-side-ins-x btwn_stringers 0))
  (add-polyline blockObj
                (list l-side-ins-x (- btwn_stringers 0.25) 0)
                (list '(0 0 0) (list l-side-len 0 0))
                nil)

  (set-active-layer "DIM")

  ;; back
  (add-mleader blockObj
               (list (1- l-btwn-ins-x) 0 0)
               '((0 0 0)
                 (0 -6 0)
                 (-1 -6 0))
               (xd-value "top-haunch-l-btwn" marks))
  ;; right side
  (add-mleader blockObj
               (list (- 0 ang-wid 1.125) 0 0)
               '((0 0 0)
                 (0 -6 0)
                 (-1 -6 0))
               (xd-value "top-haunch-l-sides" marks))
  ;; left side
  (add-mleader blockObj
               (list (- 0 ang-wid 1.125) btwn_stringers 0)
               '((0 0 0)
                 (0 6 0)
                 (-1 6 0))
               (xd-value "top-haunch-l-sides" marks))

  (set-active-layer clayer))
(defun-r 'top-conc-haunch-top)

(defun-q bot-conc-haunch-material (/)
  "Returns the material list for a bottom concrete haunch"
  (declare with-data)
  (#! '(
        (("id" "bot-haunch-l-btwn-1")
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("qty" 1)
         ("length" #(frac (- btwn_stringers 0.25))))
        (("id" "bot-haunch-l-btwn-2")
         ("desc" "ANGLE 3 x 2 x 1/4")
         ("qty" 1)
         ("length" #(frac (- btwn_stringers 0.25))))
        (("id" "bot-haunch-l-sides")
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("qty" 2)
         ("length" #(frac (- bot_ext 6.75))))
        (("id" "bot-haunch-decking")
         ("desc" "BR DECKING 1 1/2 x 36")
         ("qty" #(ceil (/ (- bot_ext 2.75) 36.0)))
         ("length" #(frac (- btwn_stringers 0.25))))
        )))
(defun-r 'bot-conc-haunch-material)

(defun-q bot-conc-haunch-side (fab-p / ang-len ang-x-offset clayer)
  "Draws a bottom concrete haunch side view"
  (declare with-data)
  (setq ang-x-offset (+ (get-end-braced-len "bot") 0.125)
        ang-len (- (+ bot_ext bot_cdist) (+ ang-x-offset 5.375)))
  (add-rectangle blockObj
                 ang-len
                 -2
                 (list ang-x-offset -3.5))
  (add-polyline blockObj
                (list ang-x-offset -3.75 0)
                (list '(0 0 0)
                      (list ang-len 0 0))
                nil)
  (add-decking blockObj
               (list (- ang-x-offset 0.375) -3.5 0)
               (list (+ ang-x-offset ang-len 0.375) -3.5 0))
  (vla-Rotate (add-shape blockObj (list (+ ang-x-offset ang-len 2.125) -3.75) "L2x2x1/4")
              (3dpt (+ ang-x-offset ang-len 2.125) -3.75)
              (dtr 180))
  (add-shape blockObj
             (list (+ ang-x-offset ang-len 2.125)
                   (- (+ tread_thick pan_thick)))
             "L3x2x1/4")

  (if fab-p
      (progn
        (setq clayer (vla-getvar 'clayer))
        (set-active-layer "DIM")

        (haunch-stringer-callout "bot")

        (add-mleader blockObj
                     (list (1+ ang-x-offset) -3.5 0)
                     '((0 0 0)
                       (1 -10 0)
                       (2 -10 0))
                     (strcat (xd-value "bot-haunch-l-sides" marks) "\n"
                             (xd-value "bot-haunch-decking" marks)))

        (add-mleader blockObj
                     (list (- bot_ext 2) -3.5 0)
                     '((0 0 0)
                       (1 -10 0)
                       (2 -10 0))
                     (strcat (xd-value "bot-haunch-l-btwn-1" marks) "\n"
                             (xd-value "bot-haunch-l-btwn-2" marks) " (LLH)"))

        (set-active-layer clayer))))
(defun-r 'bot-conc-haunch-side)

(defun-q bot-conc-haunch-top (/ ang-wid clayer l-btwn-ins l-side-ins-x l-side-len)
  "Draws a bottom concrete haunch top view"
  (declare with-data)
  (setq clayer (vla-getvar 'clayer)
        ang-wid (get-end-braced-len "bot")
        l-btwn-ins-x (- bot_ext 2)
        l-side-ins-x (+ ang-wid 0.125)
        l-side-len (- bot_ext 4.25 ang-wid))
  (set-active-layer "HIDDEN")

  ;; 2x2 between
  (add-rectangle blockObj -2 btwn_stringers (list l-btwn-ins-x 0 0))
  (add-line-btwn (- l-btwn-ins-x 0.25))
  ;; 3x2 between
  (add-rectangle blockObj 3 btwn_stringers (list l-btwn-ins-x 0 0))
  (add-line-btwn (+ l-btwn-ins-x 0.25))
  ;; right side
  (add-rectangle blockObj l-side-len 2 (list l-side-ins-x 0 0))
  (add-polyline blockObj
                (list l-side-ins-x 0.25 0)
                (list '(0 0 0)
                      (list l-side-len 0 0))
                nil)
  ;; left side
  (add-rectangle blockObj l-side-len -2 (list l-side-ins-x btwn_stringers 0))
  (add-polyline blockObj
                (list l-side-ins-x (- btwn_stringers 0.25) 0)
                (list '(0 0 0)
                      (list l-side-len 0 0))
                nil)

  (set-active-layer "DIM")

  ;; between
  (add-mleader blockObj
               (list l-btwn-ins-x 0 0)
               '((0 0 0)
                 (0 -6 0)
                 (1 -6 0))
               (strcat (xd-value "bot-haunch-l-btwn-1" marks) "\n"
                       (xd-value "bot-haunch-l-btwn-2" marks) " (LLH)"))
  ;; right side
  (add-mleader blockObj
               (list (1+ ang-wid) 0 0)
               '((0 0 0)
                 (0 -6 0)
                 (1 -6 0))
               (xd-value "bot-haunch-l-sides" marks))
  ;; left side
  (add-mleader blockObj
               (list (1+ ang-wid) btwn_stringers 0)
               '((0 0 0)
                 (0 6 0)
                 (1 6 0))
               (xd-value "bot-haunch-l-sides" marks))
  (set-active-layer clayer))
(defun-r 'bot-conc-haunch-top)

;;;;;;;;;;;;;;;;;;;;;
;; checker haunches
;;;;;;;;;;;;;;;;;;;;;

(defun-q stair-calc-can-spacing (end-braced-len unbraced-len / ncans gap min-gap ret)
  "Returns can spacing as a list of X offsets from end of stringer"
  (declare with-data)
  (setq ncans (ceil (/ (- unbraced-len 6.0) 12.0))
        gap (/ (- unbraced-len (* ncans 6)) (1+ ncans))
        min-gap (if (= quiet_tread "1") 5 3))
  (while (and (> ncans 0)
              (< gap min-gap))
      (setq ncans (1- ncans)
            gap (/ (- unbraced-len (* ncans 6)) (1+ ncans))))
  (if (> ncans 0)
      (progn
        (setq gap (rnd-dn gap 0.125)
              ncans (1- ncans)
              ret (list (+ end-braced-len gap 3))
              i 0)
        (while (< i ncans)
          (inc! 'i)
          (add-to-list 'ret (+ (car ret) (* i (+ 6 gap)))))
        ret)
      (if (>= unbraced-len 8.0)
          (list (+ end-braced-len (/ unbraced-len 2.0))))))
(defun-r 'stair-calc-can-spacing)

(defun-q chkr-haunch-common (end-braced-len unbraced-len top-p
                             / ang-len ang-x ang-y can-pts qts oc )
  "Adds cans and Quiet Tread to checker plate haunches"
  (declare with-data)
  (if (setq cans (stair-calc-can-spacing end-braced-len unbraced-len))
      (progn
        (setq ang-len ((if top-p - +) (if (> (length cans) 1) (+ (- (last cans) (car cans)) 6) 6))
              cans (mapcar (if top-p '- '+) cans)
              ang-x ((if top-p + -) (car cans) 3)
              ang-y (- (+ pan_thick 2.5))
              can-pts '((3.0 2.5 0.0)
                        (3.0 0.25 -0.414214)
                        (2.75 0.0 0.0)
                        (-2.75 0.0 -0.414214)
                        (-3.0 0.25 0.0)
                        (-3.0 2.5 0.0)
                        (-2.9253 2.5 0.0)
                        (-2.9253 0.25 0.414214)
                        (-2.75 0.0747 0.0)
                        (2.75 0.0747 0.414214)
                        (2.9253 0.25 0.0)
                        (2.9253 2.5 0.0)))
        (add-rectangle blockObj ang-len -2 (list ang-x ang-y 0))
        (add-polyline blockObj (list ang-x (- ang-y 0.25) 0) (list '(0 0 0) (list ang-len 0 0)) nil)
        (if (= quiet_tread "1")
          (foreach qt-cen (qt-cen-lst (if top-p "top" "bot"))
            (add-qt (car qt-cen) (- pan_thick) (cadr qt-cen))))
        (foreach can cans
           (if (= quiet_tread "1")
               (add-qt can ang-y 4))
           (add-polyline blockObj (list can ang-y 0) can-pts nil)))
      (if (and (= quiet_tread "1")
               (> unbraced-len 5.0))
          (add-qt ((if top-p - +) (+ end-braced-len (/ unbraced-len 2.0))) (- pan_thick) 4))))
(defun-r 'chkr-haunch-common)

(defun-q qt-cen-lst (t/b / ext b-y b-n cen spc off ret)
  "Returns a list of X offsets for high Quiet Tread"
  (declare with-data
           (vars (t/b str)))
  (setq ext (if (= t/b "top") (+ top_ext top_cdist) (- bot_ext 1))
        b-y (get-end-braced-len t/b)
        b-n (- ext b-y (if (= t/b "top") (top-chkr-nosing-braced-len) 0))
        cen (mapcar (if (= t/b "top") '- '+) (stair-calc-can-spacing b-y b-n)))
  (cond ((> (length cen) 1)
         (setq spc (/ (- (cadr cen) (car cen)) 2.0)
               ret (mapcar '(lambda(x) (- x spc)) cen)
               ret (append ret (list (+ (last ret) (* 2 spc))))
               ret (mapcar '(lambda (x) (list x 4)) ret)))
        ((>= b-n 12)
         (setq spc (/ (+ b-n 6) 4.0)
               off (mapcar '(lambda (f) (f spc)) (list - +))
               ret (mapcar '(lambda (off) (+ (car cen) off)) off)
               ret (mapcar '(lambda (x) (list x (if (>= b-n 16) 4 2))) ret)))))
(defun-r 'qt-cen-lst)

(defun-q top-chkr-haunch-side (fab-p / end-braced-len unbraced-len cans clayer)
  "Draws a top checker plate haunch side view"
  (declare with-data)
  (setq end-braced-len (get-end-braced-len "top")
        unbraced-len (- (+ top_ext top_cdist)
                        end-braced-len
                        (top-chkr-nosing-braced-len)))
  (chkr-haunch-common end-braced-len unbraced-len T)

  (if fab-p
      (progn
        (setq clayer (vla-getvar 'clayer))
        (set-active-layer "DIM")
        (haunch-stringer-callout "top")
        (add-mleader blockObj
                     (list (car cans) (- 0 pan_thick 2.5) 0)
                     '((0 0 0)
                       (0 -12 0)
                       (-1 -12 0))
                     (strcat (xd-value "top-haunch-l-sides" marks) "\n"
                             (xd-value "top-haunch-can" marks)
                             (if (> (length cans) 1) " TYP" "")))
        (set-active-layer clayer))))
(defun-r 'top-chkr-haunch-side)

(defun-q top-chkr-haunch-material (/ end-braced-len unbraced-len cans)
  "Returns the material list for a top checker plate haunch"
  (declare with-data)
  (setq end-braced-len (get-end-braced-len "top")
        unbraced-len (- (+ top_ext top_cdist)
                        end-braced-len
                        (top-chkr-nosing-braced-len))
        cans (stair-calc-can-spacing end-braced-len unbraced-len))
  (#! '(
        (("id" "top-haunch-l-sides")
         ("qty" 2)
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("length" #(frac (- (last cans) (car cans) -6))))
        (("id" "top-haunch-can")
         ("qty" #(length cans))
         ("desc" "PLATE 14GA")
         ("width" "10.64")
         ("length" #(frac (- btwn_stringers 0.25)))
         ("detail" (can)))
        )))
(defun-r 'top-chkr-haunch-material)

(defun-q top-chkr-haunch-top (/ end-braced-len unbraced-len cans l-sides-len cdimstyle)
  "Draws a top checker plate haunch top view"
  (declare with-data)
  (setq end-braced-len (get-end-braced-len "top")
        unbraced-len (- (+ top_ext top_cdist)
                        end-braced-len
                        (top-chkr-nosing-braced-len))
        cans (stair-calc-can-spacing end-braced-len unbraced-len)
        l-sides-len (- (- (last cans) (car cans) -6))
        cans (mapcar '(lambda(x) (- (- x 3))) cans)
        cdimstyle (vla-getvar 'dimstyle))

  (set-active-layer "HIDDEN")
  ;; right side
  (add-rectangle blockObj l-sides-len 2 (list (car cans) 0 0))
  (add-polyline blockObj
                (list (car cans) 0.25 0)
                (list '(0 0 0)
                      (list l-sides-len 0 0))
                nil)
  ;; left side
  (add-rectangle blockObj l-sides-len -2 (list (car cans) btwn_stringers 0))
  (add-polyline blockObj
                (list (car cans) (- btwn_stringers 0.25) 0)
                (list '(0 0 0)
                      (list l-sides-len 0 0))
                nil)
  ;; cans
  (foreach c cans
           (vla-put-Layer
            (add-polyline blockObj
                          (list (- c 3) 0 0)
                          (list (list 0 (- 0 r-string_wid 2) 0)
                                (list 0 (+ btwn_stringers l-string_wid 2) 0))
                          nil)
            "CENTER")
           (add-rectangle blockObj -6 btwn_stringers (list c 0 0)))

  ;; dims
  (mapcar! '(lambda(x) (- x 3)) 'cans)
  (set-active-layer "DIM")
  (set-active-dimstyle "1xCOMPACT")
  (vla-AddDimRotated blockObj
                     (3dpt 0 (+ btwn_stringers l-string_wid))
                     (3dpt (car cans) (+ btwn_stringers l-string_wid 2))
                     (3dpt 0 (+ btwn_stringers l-string_wid (* 2 dimspace)))
                     0)
  (if (> (length cans) 1)
      (progn
        (set-active-dimstyle "1xCOMPACT_TYP")
        (vla-AddDimRotated blockObj
                           (3dpt (car cans) (+ btwn_stringers l-string_wid 2))
                           (3dpt (cadr cans) (+ btwn_stringers l-string_wid 2))
                           (3dpt 0 (+ btwn_stringers l-string_wid (* 2 dimspace)))
                           0)))
  (set-active-dimstyle cdimstyle)

  ;; callouts
  (add-mleader blockObj
               (list (car cans) 0 0)
               '((0 0 0)
                 (0 -6 0)
                 (-1 -6 0))
               (strcat (xd-value "top-haunch-l-sides" marks) "\n"
                       (xd-value "top-haunch-can" marks)
                       (if (> (length cans) 1) " TYP" ""))))
(defun-r 'top-chkr-haunch-top)

(defun-q bot-chkr-haunch-side (fab-p
                               /
                               end-braced-len
                               unbraced-len
                               cans
                               ang-len
                               ang-y
                               ang-x
                               can-pts
                               clayer
                               )
  "Draws a bottom checker plate haunch side view"
  (declare with-data)
  (setq end-braced-len (get-end-braced-len "bot")
        unbraced-len (- bot_ext end-braced-len 1))
  (chkr-haunch-common end-braced-len unbraced-len nil)
  (add-shape blockObj (list (- bot_ext 1) (- pan_thick)) "L2x2x1/4")

  (if fab-p
      (progn
        (setq clayer (vla-getvar 'clayer))
        (set-active-layer "DIM")
        (haunch-stringer-callout "bot")
        (vla-AddDimRotated blockObj
                           (3dpt 0 (max 0 (sidekey "nosing_bot")))
                           (3dpt (1- bot_ext) (- pan_thick))
                           (3dpt 0 (* 4 dimspace))
                           0)
        (add-mleader blockObj
                     (list (car cans) (- -2.5 pan_thick) 0)
                     '((0 0 0)
                       (0 -12 0)
                       (1 -12 0))
                     (strcat (xd-value "bot-haunch-l-sides" marks) "\n"
                             (xd-value "bot-haunch-can" marks)
                             (if (> (length cans) 1) " TYP" "")))
        (set-active-layer clayer))))
(defun-r 'bot-chkr-haunch-side)

(defun-q bot-chkr-haunch-material (/ end-braced-len unbraced-len cans)
  "Returns the material list for a bottom checker plate haunch"
  (declare with-data)
  (setq end-braced-len (get-end-braced-len "bot")
        unbraced-len (- bot_ext end-braced-len 1)
        cans (stair-calc-can-spacing end-braced-len unbraced-len))
  (#! '(
        (("id" "bot-haunch-l-btwn")
         ("qty" 1)
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("length" #(frac (- btwn_stringers 0.25))))
        (("id" "bot-haunch-l-sides")
         ("qty" 2)
         ("desc" "ANGLE 2 x 2 x 1/4")
         ("length" #(frac (- (last cans) (car cans) -6))))
        (("id" "bot-haunch-can")
         ("qty" #(length cans))
         ("desc" "PLATE 14GA")
         ("width" "10.64")
         ("length" #(frac (- btwn_stringers 0.25)))
         ("detail" (can)))
        )))
(defun-r 'bot-chkr-haunch-material)

(defun-q bot-chkr-haunch-top (/ end-braced-len unbraced-len cans l-sides-len l-btwn-ins-x cdimstyle)
  "Draws a bottom checker plate haunch top view"
  (declare with-data)
  (setq end-braced-len (get-end-braced-len "bot")
        unbraced-len (- bot_ext end-braced-len 1)
        l-btwn-ins-x (- bot_ext 1)
        cans (stair-calc-can-spacing end-braced-len unbraced-len)
        l-sides-len (- (last cans) (car cans) -6)
        cans (mapcar '(lambda(x) (- x 3)) cans)
        cdimstyle (vla-getvar 'dimstyle))

  (set-active-layer "HIDDEN")
  ;; between
  (add-rectangle blockObj 2 btwn_stringers (list l-btwn-ins-x 0 0))
  (add-line-btwn (+ l-btwn-ins-x 0.25))
  ;; right side
  (add-rectangle blockObj l-sides-len 2 (list (car cans) 0 0))
  (add-polyline blockObj
                (list (car cans) 0.25 0)
                (list '(0 0 0)
                      (list l-sides-len 0 0))
                nil)
  ;; left side
  (add-rectangle blockObj l-sides-len -2 (list (car cans) btwn_stringers 0))
  (add-polyline blockObj
                (list (car cans) (- btwn_stringers 0.25) 0)
                (list '(0 0 0)
                      (list l-sides-len 0 0))
                nil)
  ;; cans
  (foreach c cans
           (vla-put-Layer
            (add-polyline blockObj
                          (list (+ c 3) 0 0)
                          (list (list 0 (- 0 r-string_wid 2) 0)
                                (list 0 (+ btwn_stringers l-string_wid 2) 0))
                          nil)
            "CENTER")
           (add-rectangle blockObj 6 btwn_stringers (list c 0 0)))

  ;; dims
  (mapcar! '(lambda(x) (+ x 3)) 'cans)
  (set-active-layer "DIM")
  (set-active-dimstyle "1xCOMPACT")
  (vla-AddDimRotated blockObj
                     (3dpt 0 (+ btwn_stringers l-string_wid))
                     (3dpt (car cans) (+ btwn_stringers l-string_wid 2))
                     (3dpt 0 (+ btwn_stringers l-string_wid (* 2 dimspace)))
                     0)
  (if (> (length cans) 1)
      (progn
        (set-active-dimstyle "1xCOMPACT_TYP")
        (vla-AddDimRotated blockObj
                           (3dpt (car cans) (+ btwn_stringers l-string_wid 2))
                           (3dpt (cadr cans) (+ btwn_stringers l-string_wid 2))
                           (3dpt 0 (+ btwn_stringers l-string_wid (* 2 dimspace)))
                           0)))
  (set-active-dimstyle cdimstyle)

  ;; callouts
  (add-mleader blockObj
               (list (car cans) 0 0)
               '((0 0 0)
                 (0 -6 0)
                 (1 -6 0))
               (strcat (xd-value "bot-haunch-l-sides" marks) "\n"
                       (xd-value "bot-haunch-can" marks)
                       (if (> (length cans) 1) " TYP" "")))
  (add-mleader blockObj
               (list l-btwn-ins-x (+ btwn_stringers l-string_wid) 0)
               '((0 0 0)
                 (0 6 0)
                 (1 6 0))
               (xd-value "bot-haunch-l-btwn" marks)))
(defun-r 'bot-chkr-haunch-top)

(defun-q tc-bolt-desc (desc)
  "Returns the proper TC bolt description, taking into account finish"
  (declare (vars (desc str)))
  (strcat (if (= finish "Galvanized") "GALVANIZED " "")
          "T.C. BOLT (A325) "
          desc))
(defun-r 'tc-bolt-desc)
