(set-file-docstring "Utility functions for stairs")

(psc-include '("rail/util.lsp"
               "shapes/line.lsp"
               "stair/tread.lsp"))

(defun-q stair-dir (ascend z-position)
  "Return the stair direction (either +1 or -1) based on ASCEND and Z-POSITION."
  (declare (vars (ascend str
                   (member ascend '("Left" "Right")))
                 (z-position str
                             (member z-position '("Near" "Far"))))
           (tests (= (stair-dir "Left" "Near") 1)
                  (= (stair-dir "Right" "Far") 1)
                  (= (stair-dir "Left" "Far") -1)
                  (= (stair-dir "Right" "Near") -1)))
  (apply '* (mapcar '(lambda (x) (if (member x '("Near" "Left")) +1 -1))
                     (list ascend z-position))))
(defun-r 'stair-dir)

(defun-q stair-slope ()
  "Returns the nosing line in slope-intercept form."
  (declare with-data)
  (apply 'pt-slope
         (cons (/ rise_hgt (* dir tread_depth))
               (car (calc-nose-pts)))))
(defun-r 'stair-slope)

(defun-q stair-hr-assembly ()
  "Returns the assembly name of the HG return for this stair."
  (declare with-data
           (vars (assembly str (wcmatch assembly "*S*"))))
  (vl-string-subst "HR" "S" assembly))
(defun-r 'stair-hr-assembly)

(defun-q stair-rail-assembly (side)
  "Returns the assembly name of the stair rail on SIDE of this stair."
  (declare with-data
           (vars (side str (member side '("i" "o" "l" "r")))
                 (number str)
                 (level str)))
  (strcat (if (wcmatch (sidekey-rail "type") "*Stair*") "S" "W")
          "R" number "-" level (rail-sequence side)))
(defun-r 'stair-rail-assembly)

(defun-q stair-guard-assembly (location)
  "Returns the assembly name of the guard rail at LOCATION.

LOCATION should be one of the following strings:

- \"ibot\" (inside bottom)
- \"obot\" (outside bottom)
- \"itop\" (inside top)
- \"otop\" (outside top)
- \"btwn\" (between stairs)"
  (declare with-data
           (vars (location str (member location '("ibot" "obot" "itop" "otop" "btwn")))
                 (number str)
                 (level str)))
  (strcat "GR"
          number
          "-"
          level
          (?s sequence)
          ((lambda(/ num)
           (if (and (setq num (stair-guard--num location))
                    (> num 0))
               (strcat "-" (to-string num))
               "")))))
(defun-r 'stair-guard-assembly)

(defun-q stair-guard--num (loc / hamming setp codex)
  "Returns the number of a guard rail

This function calculates the -N at the end of a stair guard rail's sequence number based
on location LOC. See %stair-guard-assembly for format of LOC.

If there is no guard rail at LOC, returns nil. If there is only one guard rail on the
flight, returns 0, indicating numbers are unnecessary."
  (declare with-data
           (vars (loc str (member loc '("ibot" "obot" "itop" "otop" "btwn")))))
  ;; hamming weight of a bit string
  (defun hamming (n / ret)
    (setq ret 0)
    (while (> n 0)
      (setq n (logand n (1- n))
            ret (1+ ret)))
    ret)
  ;; T if bit at loc is set
  (defun setp (loc) (/= (logand (cdr (assoc loc codex)) guard_rails) 0))
  (setq codex '(("ibot" . 1)
                ("obot" . 2)
                ("itop" . 4)
                ("otop" . 8)
                ("btwn" . 16)))
  (if (> (hamming guard_rails) 1)
      (if (= loc "ibot")
          (if (setp "ibot") 1)
          (if (setp loc)
              (1+ (hamming (logand guard_rails (1- (cdr (assoc loc codex))))))))
      (if (setp loc) 0)))
(defun-r 'stair-guard--num)

(defun-q add-stair-stringer (inc-list container side str-lay flg-lay)
  "Add a stair stringer outline to CONTAINER

INC-LIST is the stringer properties list. SIDE is a string starting with l, r, i, or
o. STR-LAY and FLG-LAY are the layers on which to draw the stringer outline and flange
lines, respectively."
  (declare (vars (inc-list list)
                 (container vla-object)
                 (side str)
                 (str-lay str)
                 (flg-lay str))
           with-data)
  (mapcar
   '(lambda (lst)
     (if (cdr lst)
         (vla-put-Layer-r (add-polyline container
                                        '(0 0 0)
                                        (mapcar '(lambda (pt)
                                                  (append (mapcar '* pt (list dir 1)) '(0.0)))
                                                (cdr lst))
                                        (> (length (cdr lst)) 2))
                          (if (> (length (cdr lst)) 2) str-lay flg-lay))))
   (vl-remove-if-not '(lambda (lst)
                       (apply 'or
                        (mapcar '(lambda (wcm) (wcmatch (vl-symbol-name (car lst)) wcm))
                         (mapcar '(lambda (wcl) (strcat "*" (vl-symbol-name wcl) "*"))
                          inc-list))))
                     (stair-stringer-calc side))))
(defun-r 'add-stair-stringer)

(defun-q stair-dim-format (h/v dimObj)
  "Formats a stair dimension

DIMOBJ is the dimension object. H/V is either the symbol 'h or 'v. 'h means to format as a
run dimension and 'v means to format it as a rise dimension."
  (declare (vars (h/v sym (member h/v '(h v)))
                 (dimobj vla-object)))
  (vla-put-textoverride dimObj txt)
  (vla-put-verticaltextposition dimObj 0)
  (vla-put-textinsidealign dimObj (vla-logic (= h/v 'v)))
  (vla-put-textoutsidealign dimObj (vla-logic (= h/v 'v))))
(defun-r 'stair-dim-format)

(defun-q stair-block-qt-takeoff (blockdef / ret)
  "Returns a list of QT quantities in BLOCKDEF"
  (declare (vars (blockdef vla-object)))
  (vlax-for obj blockdef
            (if (= (vla-get-ObjectName obj) "AcDbBlockReference")
                (append! 'ret (stair-block-qt-takeoff (blockref2def obj)))
                (if (= (vla-get-Layer obj) "Stair Quiet Tread")
                      (add-to-list 'ret
                                   (list (/ (vla-get-Area obj) 0.125)
                                         (if (< (cadar (get-vlist obj)) -1)
                                             (- btwn_stringers 6)
                                             (- btwn_stringers 2)))))))
  (vl-sort ret '(lambda(a b) (> (car a) (car b)))))
(defun-r 'stair-block-qt-takeoff)

(defun-q stair-qt-takeoff (ename / bd data inner)
  "Returns a Quiet Tread takeoff for the stair object referred to by ENAME"
  (setq bd (blockdef ename)
        data (draw-xdata-migrate 'ename "Stair"))
  (defun inner(bd / takeoff)
    (setq takeoff (stair-block-qt-takeoff bd)
          takeoff (count-items
                   (mapcar 'process-generic-def
                           takeoff
                           (lst* (strcat (find-job-dir) "/qt.csv") (length takeoff))))))
  (with-data data '((inner bd))))
(defun-r 'stair-qt-takeoff)

(defun-q stair-stn-p ()
  "Returns T when this stair has STN"
  (declare with-data)
  (not (member stn_style '(nil "None"))))
(defun-r 'stair-stn-p)

(defun-q get-stn-desc-by-sku (sku / ret)
  "Returns the material list entry for a STN with SKU"
  (foreach x (read-csv (findfile "fab/material.csv"))
           (if (= (last x) sku)
               (setq ret (car x))))
  ret)
(defun-r 'get-stn-desc-by-sku)

;;;;;;;;;;;
;; tests ;;
;;;;;;;;;;;

(if *test-build-p*
  (progn
    (defun-q test-stair-hr-assembly (/ from-with-data assembly)
      "stair-hr-assembly"
      (setq from-with-data t
            assembly "S1-1A")
      (assert-all '((= (stair-hr-assembly) "HR1-1A"))))
    (defun-t 'test-stair-hr-assembly)))
