(set-file-docstring
 "Functions related to basic math")

(defun-q square (n)
  "Returns the square of N"
  (declare (vars (n nil (numberp n)))
           (tests (= (square 2) 4)
                  (= (square -3) 9)))
  (* n n))
(defun-r 'square)

(defun-q is-even (n)
  "Returns nil if N is not divisible by 2"
  (declare (vars (n int))
           (tests (not (is-even 1))
                  (is-even 2)))
  (zerop (rem n 2.0)))
(defun-r 'is-even)

;;;;;;;;;;;;;;;;;;;;;;;;;
;; increment/decrement ;;
;;;;;;;;;;;;;;;;;;;;;;;;;

(defun-q inc! (int-sym)
  "Increments INT-SYM destructively"
  (set int-sym (1+ (vl-symbol-value int-sym))))
(declare-late 'inc!
 '((vars (int-sym sym (numberp (vl-symbol-value int-sym))))))
(defun-r 'inc!)

(defun-q post-inc! (int-sym / val)
  "Returns the value of INT-SYM then increment it destructively"
  (declare (vars (int-sym sym (numberp (vl-symbol-value int-sym)))))
  (1- (inc! int-sym)))
(defun-r 'post-inc!)

(defun-q dec! (int-sym)
  "Decrements INT-SYM destructively"
  (declare (vars (int-sym sym (numberp (vl-symbol-value int-sym)))))
  (set int-sym (1- (vl-symbol-value int-sym))))
(defun-r 'dec!)

(defun-q post-dec! (int-sym / val)
  "Returns the value of INT-SYM then decrement it destructively"
  (declare (vars (int-sym sym (numberp (vl-symbol-value int-sym)))))
  (1- (inc! int-sym)))
(defun-r 'post-dec!)

;;;;;;;;;;;;;;;;;;;;
;; trig functions ;;
;;;;;;;;;;;;;;;;;;;;

(defun-q tan (ang)
  "Returns the tangent of ANG (radians)"
  (declare (vars (ang nil (numberp ang)))
           (tests (= (tan 0) 0.0)))
  (/ (sin ang) (cos ang)))
(defun-r 'tan)

(defun-q acos (ang)
  "Returns the arccosine of ANG (radians)"
  (declare (vars (ang nil (numberp ang))))
  (atan (/ (sqrt (- 1 (* ang ang))) ang)))
(defun-r 'acos)

;;;;;;;;;;;;;;
;; rounding ;;
;;;;;;;;;;;;;;

(defun-q ceil (n / fn)
  "Returns the smallest integer greater than or equal to N"
  (declare (vars (n nil (numberp n)))
           (tests (= (ceil 0.1) 1)
                  (= (ceil 1) 1)
                  (= (ceil -0.9) 0)))
  (setq fn (fix n))
  (if (> n fn) (1+ fn) fn))
(defun-r 'ceil)

(defun-q rnd-dn (num prc)
  "Rounds NUM down to the next multiple of precision PRC"
  (declare (vars (num nil (numberp num))
                 (prc nil (numberp prc)))
           (tests (= (rnd-dn 1.3 0.25) 1.25)
                  (= (rnd-dn 1.9 0.5) 1.5)))
  (- num (rem num (float prc))))
(defun-r 'rnd-dn)

(defun-q rnd-up (num prc)
  "Rounds NUM up to the next multiple of precision PRC"
  (declare (vars (num nil (numberp num))
                 (prc nil (numberp prc)))
           (tests (= (rnd-up 1.1 0.25) 1.25)
                  (= (rnd-up 1.1 0.5) 1.5)))
  (- (+ num prc) (rem num prc)))
(defun-r 'rnd-up)
