(set-file-docstring
 "Functions related to comparisons")

(defun-q identity (x)
  "Returns X."
  (declare (tests (= (identity 'x) 'x)))
  x)
(defun-r 'identity)

(defun-q sort-by (func cmp)
  "Returns a lambda for processed sorting

The result of this function is a lambda suitable for passing to vl-sort. CMP should be a
comparison function of 2 arguments that would normally be passed to vl-sort. FUNC should
be a function of 1 argument that will be used to process both arguments to CMP.

Returns the equivalent of '(lambda (a b) (CMP (FUNC a) (FUNC b)))."
  (declare (vars (func (sym list subr usubr))
                 (cmp (sym list subr usubr)))
           (tests (equal (sort-by 'car '<)
                         '(lambda (a b) (< (car a) (car b))))))
  (append '(lambda (a b))
          (#! '((#cmp (#func a) (#func b))))))
(defun-r 'sort-by)

(defun-q sort-by-key (key cmp)
  "Returns a lambda for sorting a data list by the value of KEY

The result of this function is a lambda suitable for passing to vl-sort. CMP should be a
comparison function of 2 arguments that would normally be passed to vl-sort. KEY should be
a key from the data list as a string."
  (declare (vars (key str)
                 (cmp (sym list subr usubr))))
  (append '(lambda (a b))
          (#! '((#cmp (xd-value #key a) (xd-value #key b))))))
(defun-r 'sort-by-key)

(defun-q sort-strings-asc (a b)
  "A comparison function that accounts for length when sorting strings

Pass as comparison argument to vl-sort."
  (if (= (strlen a) (strlen b))
      (< a b)
      (< (strlen a) (strlen b))))
(defun-r 'sort-strings-asc)

;; for use with vl-sort
(setq sort-data-by-key '(lambda(a b) (< (car a) (car b))))

;; float comparison operators with fuzz
(defun-q >f (a b fuzz)
  "Returns nil unless A is greater than B by FUZZ"
  (> (- a b) fuzz))
(defun-r '>f)

(defun-q <f (a b fuzz)
  "Returns nil unless A is less than B by FUZZ"
  (> (- b a) fuzz))
(defun-r '<f)

(defun-q =f (a b fuzz)
  "Return nil unless A and B are equal within FUZZ"
  (< (abs (- a b)) fuzz))
(defun-r '=f)

(defun-q sort-stairland (a b / ay az by bz)
  "Sorts a part of stair/landing enames in ascending order.

This function is designed to be passed to vl-sort. The resulting sort order will sort by
elevation and put landings in front of stairs at the same level. Multiple landings at the
same level are not considered."
  (setq ay (caddr (assoc 10 (entget a)))
        az (cadddr (assoc 10 (entget a)))
        by (caddr (assoc 10 (entget b)))
        bz (cadddr (assoc 10 (entget b))))
  (if (wcmatch (caadr (assoc -3 (entget a '("*")))) "*3D*")
      (if (< (abs (- az bz)) 0.001)
          (= (caadr (assoc -3 (entget b '("*")))) "3D_Stair")
          (< az bz))
      (if (< (abs (- ay by)) 0.001)
          (= (caadr (assoc -3 (entget b '("*")))) "Stair")
          (< ay by))))
(defun-r 'sort-stairland)

(defun-q or* (branches)
  "Returns the first non-nil element of BRANCHES.

or* doesn't eval the branches, so doesn't do short-circuiting."
  (declare (vars (branches nil (listp branches)))
           (tests (= (or* '(nil 1 2)) 1)
                  (= (or* '(nil nil)) nil)))
  (while (and branches (not (car branches))) (pop! 'branches))
  (car branches))
(defun-r 'or*)
