(set-file-docstring
 "Functions related to conversions")

(defun-q to-string (x)
  "Returns X as a string

Essentially a wrapper for vl-prin1-to-string that returns the empty string for nil rather
than \"nil\"."
  (if x (if (stringp x)
          x
        (vl-prin1-to-string x))
      ""))
(defun-r 'to-string)

(defun-q safe-atoi (str / )
  "Wrapper for atoi that returns nil if there was no number

(atoi \"\")      => 0
(safe-atoi \"\") => nil"
  (if (wcmatch str "#*")
      (atoi str)
      nil))
(defun-r 'safe-atoi)

(defun-q dtr (rad)
  "Returns RAD in radians"
  (declare (vars (rad (int real))))
  (* pi (/ rad 180.0)))
(defun-r 'dtr)

(defun-q rtd (deg)
  "Return DEG in degrees"
  (/ (* deg 180.0) pi))
(defun-r 'rtd)

(defun-q vlax-collection->lst (collection / ret)
  "Convert COLLECTION to a list of names

COLLECTION should be a vlax collection of objects that support a Name property."
  (declare (vars (collection vla-object)))
  (vlax-for x collection (add-to-list 'ret (vla-get-Name x))))
(defun-r 'vlax-collection->lst)

(defun-q frac (n)
  "Alias for (rtos n 5 2)"
  ;; (declare (vars (n (int real)))
  ;;          (tests (= (frac 0.25) "1/4")))
  (rtos n 5 4))
(defun-r 'frac)

(defun-q ? (x)
  "Alias for (if X X)"
  (if x x))
(defun-r '?)

(defun-q ?? (x default)
  "Alias for (if X X DEFAULT)"
  (if x x default))
(defun-r '??)

(defun-q ?n (n)
  "Alias for (if N N 0)"
  (if n n 0))
(defun-r '?n)

(defun-q ?f (n)
  "Alias for (if N N 0.0)"
  (if n n 0.0))
(defun-r '?f)

(defun-q ?s (s)
  "Alias for (if S S \"\")"
  (if s s ""))
(defun-r '?s)

(defun-q unit-dir (n)
  "Returns 0 if N is 0, 1 if N is positive, or -1 if N is negative

Uses a precision fuzz of 0.00001 for comparison. See %>f."
  ;; (declare (vars (n nil (numberp n)))
  ;;          (tests (= (unit-dir 0) 0)
  ;;                 (= (unit-dir 10) 1)
  ;;                 (= (unit-dir -10) -1)))
  (cond ((>f n 0 0.00001) 1)
        ((<f n 0 0.00001) -1)
        (T 0)))
(defun-r 'unit-dir)

(defun-q num-val (x)
  "Alias for (if (numberp X) X 0)"
  (if (numberp x) x 0))
(defun-r 'num-val)

(defun-q vlist->safearray (vertices / arr)
  "Returns VERTICES as a safearray

VERTICES should be a list of lists of numbers."
  (declare (vars (vertices (list nil) (vl-every 'point-p vertices))))
  (setq vertices (mapcar 'float (apply 'append vertices))
        arr (vlax-make-safearray vlax-vbDouble
                                 (cons 0 (1- (length vertices)))))
  (vlax-safearray-fill arr vertices)
  arr)
(defun-r 'vlist->safearray)

(defun-q safearray->list (safearray / dim lst i item return)
  "Converts arbitrary-dimension SAFEARRAY to a list"
  (declare (vars (safearray safearray)))
  (setq dim (1+ (vlax-safearray-get-dim safearray))
        lst (vlax-safearray->list safearray))
  (while lst
    (repeat dim
            (add-to-list 'item (car lst))
            (setq lst (cdr lst)))
    (add-to-list 'return item)
    (setq item nil))
  return)
(defun-r 'safearray->list)

(defun-q collection->list (collection / ret)
  "Returns vlax COLLECTION as a list"
  (declare (vars (collection vla-object)))
  (vlax-map-collection collection '(lambda (x) (add-to-list 'ret x)))
  ret)
(defun-r 'collection->list)

