(set-file-docstring
 "Functions dealing with VLA Document objects")

(defun-q get-custom-dwgprop (key / si nprops ret si-key si-val)
  "Returns value of custom drawing property KEY in current acadDoc"
  (declare (vars (key str)))
  (setq si (vla-get-SummaryInfo acadDoc)
        nprops (vla-NumCustomInfo si))
  (foreach i (range 0 nprops 1)
           (vla-GetCustomByIndex si i 'si-key 'si-val)
           (if (= key si-key)
               (setq ret si-val)))
  ret)
(defun-r 'get-custom-dwgprop)

(defun-q set-custom-dwgprop (key val / si)
  "Sets value of custom drawing property KEY to VAL"
  (declare (vars (key str)
                 (val str)))
  (setq si (vla-get-SummaryInfo acadDoc))
  (if (get-custom-dwgprop key)
      (vla-SetCustomByKey si key val)
      (vla-AddCustomInfo si key val))
  val)
(defun-r 'set-custom-dwgprop)

(defun-q open-and (filename readonly func-sym args / oldAcadDoc ret)
  "Opens drawing FILENAME and executes function pointed to by FUNC-SYM with ARGS

If READONLY is non-nil, open the drawing in read-only mode. The drawing is automatically
closed and %reset-doc is called after."
  (declare (vars (filename str)
                 (func-sym sym (function-p (vl-symbol-value func-sym)))
                 (args nil (listp args))))
  (setq oldAcadDoc acadDoc
        acadDoc (vla-Open (vla-get-Documents acadObj)
                          filename
                          (if readonly
                              :vlax-true
                              :vlax-false))
        modelSpace (vla-get-modelSpace acadDoc))

  (setq ret (apply func-sym args))

  (vla-Close acadDoc :vlax-false)
  (reset-doc)
  ret)
(defun-r 'open-and)

;; these functions mask bits in osmode to simulate pressing F3
(defun-q osnap-off ()
  "Suppress object snaps in current acadDoc and set *error* to restore them"
  (defun *error* (msg)
    (osnap-on)
    (setq *error* nil))
  (vla-setvar 'osmode (logior (expt 2 14)
                              (vla-getvar 'osmode))))
(defun-r 'osnap-off)

(defun-q osnap-on ()
  "Restore object snaps in current acadDoc"
  (setq *error* nil)
  (vla-setvar 'osmode (logand (~ (expt 2 14))
                              (vla-getvar 'osmode))))
(defun-r 'osnap-on)

(defun-q setup-env (3d-p / textstyles textstyle-names linetypes linetype-names layers layer-names
                  color idx req-textstyles req-linestyles req-layers)
  "Ensure necessary layers, linetypes, and text styles are present

Includes different things if 3D-P is non-nil."
  (setq linetypes (vla-get-Linetypes acadDoc)
        linetype-names (vlax-collection->lst linetypes)
        req-linetypes (remove '("CONTINUOUS"
                                "HIDDEN2"
                                "CENTER2")
                              '(member x linetype-names))
        textstyles (vla-get-TextStyles acadDoc)
        textstyle-names (vlax-collection->lst textstyles)
        req-textstyles (remove '(("Calibri" "C:/windows/fonts/calibril.ttf"))
                               '(member (car x) textstyle-names))
        layers (vla-get-Layers acadDoc)
        layer-names (vlax-collection->lst layers)
        req-layers
          (remove
            (if 3d-p
             '(("No Plot" 78 "CONTINUOUS")
               ("3D Stair" AcWhite "CONTINUOUS")
               ("Stair Stringers" AcYellow "CONTINUOUS")
               ("Stair Supports" AcRed "CONTINUOUS")
               ("Stair Pans" AcGreen "CONTINUOUS")
               ("Stair Conc Fill" 9 "CONTINUOUS")
               ("Stair Quiet Tread" AcMagenta "CONTINUOUS")
               ("Stair Safe-T-Nose" AcBlue "CONTINUOUS")
               ("3D Land" AcWhite "CONTINUOUS")
               ("Land Channels" AcYellow "CONTINUOUS")
               ("Land Supports" AcRed "CONTINUOUS")
               ("Land Platform" AcGreen "CONTINUOUS")
               ("Land Conc Fill" 9 "CONTINUOUS")
               ("Land Quiet Tread" AcMagenta "CONTINUOUS")
               ("3D Rail" AcWhite "CONTINUOUS")
               ("Rail Handgrab" AcMagenta "CONTINUOUS")
               ("Rail Frame" AcGreen "CONTINUOUS")
               ("3D Embed" AcWhite "CONTINUOUS")
               ("Embed" AcCyan "CONTINUOUS")
               ("Head Clear" 9 "CONTINUOUS"))
             '(("NEAR-STAIR" AcYellow "CONTINUOUS")
               ("FAR-STAIR" AcGreen "HIDDEN2")
               ("LAND" AcRed "CONTINUOUS")
               ("TEXT" AcRed "CONTINUOUS")
               ("DIM" AcRed "CONTINUOUS")
               ("HATCH" AcYellow "CONTINUOUS")
               ("OBJ" AcGreen "CONTINUOUS")
               ("HIDDEN" AcGreen "HIDDEN2")
               ("CENTER" AcBlue "CENTER2")
               ("RAILS" AcMagenta "CONTINUOUS")
               ("NO PLOT" 78 "CONTINUOUS")
               ("CHECK" AcMagenta "CONTINUOUS")
               ("SOLID" AcWhite "CONTINUOUS")
               ("Stair Quiet Tread" AcMagenta "CONTINUOUS")
               ("STN" AcBlue "CONTINUOUS"))
            ) ;_ if
           '(member (car x) layer-names)
          ) ;_ remove
  ) ;_ setq
  (foreach lt req-linetypes
           (vla-Load linetypes lt "acad.lin"))
  (foreach ts req-textstyles
           (vla-Add textstyles (car ts))
           (vla-put-FontFile (vla-Item textstyles (car ts)) (cadr ts)))
  (foreach l req-layers
           (vla-Add layers (car l))
           (setq color (vlax-create-object (strcat "AutoCAD.AcCmColor."
                                                   (substr (getvar 'acadver) 1 2)))
                 idx (cadr l)
                 idx (if (= (type idx) 'SYM) (vl-symbol-value idx) idx))
           (vla-put-ColorIndex color idx)
           (vla-put-TrueColor (vla-Item layers (car l)) color)
           (vla-put-Linetype (vla-Item layers (car l)) (caddr l))))
(defun-r 'setup-env)

(defun-q reset-doc ()
  "Resets acadDoc and modelSpace global vars"
  (setq acadDoc (vla-get-ActiveDocument acadObj)
        modelSpace (vla-get-ModelSpace acadDoc)))
(defun-r 'reset-doc)

(defun-q set-active-dimstyle (name / )
  "Sets active dimension style in current acadDoc to NAME"
  (declare (vars (name str)))
  (vla-put-ActiveDimStyle acadDoc (vla-Item (vla-get-DimStyles acadDoc) name)))
(defun-r 'set-active-dimstyle)

(defun-q set-active-layer (name / )
  "Sets active layer in current acadDoc to NAME"
  (declare (vars (name str)))
  (vla-put-ActiveLayer acadDoc (vla-Item (vla-get-Layers acadDoc) name)))
(defun-r 'set-active-layer)

(defun-q set-active-textstyle (name / )
  "Sets active text style in current acadDoc to NAME"
  (declare (vars (name str)))
  (vla-put-ActiveTextStyle acadDoc (vla-Item (vla-get-TextStyles acadDoc) name)))
(defun-r 'set-active-textstyle)

(defun-q vla-getvar (sysvar / )
  "Returns the value of SYSVAR in current acadDoc"
  (declare (vars (sysvar (sym str))))
  (vlax-variant-value (vla-GetVariable acadDoc (to-string sysvar))))
(defun-r 'vla-getvar)

(defun-q vla-setvar (sysvar val / )
  "Sets the value of SYSVAR in current acadDoc to VAL"
  (declare (vars (sysvar (sym str))))
  (vla-SetVariable acadDoc (to-string sysvar) (vlax-make-variant val)))
(defun-r 'vla-setvar)
