(set-file-docstring
 "Functions for documentation build system")

(setq *docstrings-short* nil
      *docstrings-full* nil)

(defun-q doc--add-docstring (func-sym docstring)
  "Adds DOCSTRING for FUNCSYM to global docstring vars

Adds the whole DOCSTRING to *docstrings-full* and the first line only to
*docstrings-short*. Both are entered as a cons cell of the form (FUNC-SYM . DOCSTRING)."
  (add-to-list '*docstrings-short* (cons func-sym (first-line docstring)))
  (add-to-list '*docstrings-full* (cons func-sym docstring)))
(defun-r 'doc--add-docstring)

(defun-q defun-r--process-function ()
  "Puts current function into code manual"
  (add-to-alist '*file-docs*
                *file-name*
                (list
                 (list
                  ;; name
                  (lowercase (to-string func-sym))
                  ;; args
                  (lowercase (to-string func-sig))
                  ;; sig
                  (apply 'strcat
                         (append
                          (list "(<span class=\\\"funcname\\\">"
                                ;; HTML escape angle braces in sym
                                (string-subst-all "&lt;" "<"
                                  (string-subst-all "&gt;" ">"
                                    (lowercase (to-string func-sym))))
                                "</span>")
                          (mapcar '(lambda(a) (strcat " "
                                               (lowercase (to-string a))))
                                  func-sig)
                          '(")")))
                  ;; doc_short
                  (escape-quotes (first-line func-docstring))
                  ;; doc_full
                  (apply (compose '(escape-quotes escape-newlines))
                         (list func-docstring))))
   t))
(defun-r 'defun-r--process-function)

(defun-q defun-r--process-command ()
  "Puts current function in user manual commands list"
  (add-to-list '*command-docs*
               (list (substr (uppercase (to-string func-sym)) 3)
                     (first-line func-docstring)
                     (apply (compose '(escape-quotes escape-newlines))
                            (list func-docstring)))))
(defun-r 'defun-r--process-command)

(defun-q defun-r (func-sym / func-body func-args func-sig func-docstring func-declares no-func-name)
  "Registers the function FUNC-SYM points to

Enables documentation generation and lookups with %whatis and %docstring. Replaces declare
forms in the function body. See %defun-r--process-declare for declare forms.

Redefines function using defun unless *no-redefun* is non-nil.

If *dev-mode* is nil, skip most of the process: only extract docstrings and declare forms,
then redefine the function.

If *inhibit-func-name* is nil, advise the function to set a variable *func-name* with its
name as a string."
  (setq func-body (defun-q-list-ref func-sym)
        func-args (pop! 'func-body)
        func-docstring (pop! 'func-body)
        func-declares (if (and (listp (car func-body))
                               (= (caar func-body) 'declare))
                          (cdr (pop! 'func-body))))
  (if *dev-mode*
      (progn
        ;; declares
        (if (not func-declares)
            (setq func-declares (cdr (assoc func-sym *delayed-decl*))))
        (mapcar '(lambda (decl / docstr-advice body-advice)
                  (if (atom decl)
                      (setq decl (list decl)))
                  (setq docstr-advice ((lambda(/ str)
                                         (setq str (strcat "\n\n" (to-string (car decl))
                                                           (if (cdr decl) ":" "")))
                                         (foreach elt (cdr decl)
                                           (setq str (strcat str "\n" (to-string elt))))
                                         str))
                        func-docstring (strcat func-docstring docstr-advice))
                  (setq body-advice (defun-r--process-declare decl)
                        func-body (if body-advice (cons body-advice func-body) func-body)))
                func-declares)
        ;; *func-name*
        (if (not no-func-name)
            (progn
              (if (not (member '/ func-args))
                  (setq func-args (rcons func-args '/)))
              (setq func-args (rcons func-args '*func-name*)
                    func-body (cons (list 'setq '*func-name* (to-string func-sym)) func-body))))
        (doc--add-docstring func-sym func-docstring)
        ;; signature
        (setq func-sig (take-while '(lambda(a) (/= a '/)) func-args))
        (add-to-list '*signatures* (cons func-sym
                                         (lowercase
                                          (to-string
                                           (cons func-sym func-sig)))))
        (if *doc-build-p*
            (if (begins-with (to-string func-sym) "C:")
                (defun-r--process-command)
                (defun-r--process-function)))))
  (if *no-redefun*
      (defun-q-list-set func-sym (cons func-args func-body))
      (eval (append (list 'defun func-sym func-args) func-body))))
(defun-r 'defun-r)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; the following only needed in *dev-mode* ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(if *dev-mode*
  (progn
    (defun-q defun-r--process-declare (declare-form)
      "Returns a form for inclusion in the function being processed by %defun-r

Each possible declare form has an expansion into a proper expression form, which is
inserted in place of the declare form in the function. Multiple declare forms are inserted
in the order specified. Declare form processors may modify the args, body, docstring, or
even name of the defining function, but should expressly declare doing so.

To define a new form, define a function named defun-r--process-declare-NAME, where NAME is
the first element of the form, e.g. %defun-r--process-declare-vars is called when the form
is (vars ...). It will be called with a single arg: the cdr of the declare form. It should
return a form to be consed onto the front of the function's body. Processors may return
nil which will not be included in the function body. Put processor functions right after
this one in the file so they load as early as possible.

The entire declare form will be appended to the end of the function's docstring.

See the following for available declares:

- %defun-r--process-declare-with-data
- %defun-r--process-declare-vars
- %defun-r--process-declare-tests"
      ((symcat '("defun-r--process-declare-" #(to-string (car declare-form))))
       (cdr declare-form)))
    (defun-r 'defun-r--process-declare)

    (defun-q defun-r--process-declare-with-data (_)
      "Returns function advice that ensures this function was called inside a with-data call"
      '(assert 'from-with-data))
    (defun-r 'defun-r--process-declare-with-data)

    (defun-q defun-r--process-declare-no-func-name (_)
      "Tells %defun-r not to advise the function to advertise its name"
      (setq no-func-name t)
      nil)
    (defun-r 'defun-r--process-declare-no-func-name)

    (defun-q defun-r--process-declare-vars (vars-spec / ret)
      "Returns function advice that checks vars in the eval environment

VARS-SPEC should be a list of variable specifications for passing to
%defun-r--process-declare-var."
      (setq ret (apply 'append
                       (mapcar 'defun-r--process-declare-var
                               (subst '(from-with-data non-nil) 'with-data vars-spec))))
      (if ret (cons 'progn ret)))
    (defun-r 'defun-r--process-declare-vars)

    (defun-q defun-r--process-declare-var (var-spec / var types)
      "Processes a single form for %defun-r--process-declare-vars

The car of VAR-SPEC is the name of the variable.

The cadr is a list of types. The type of the var must be one of the types. Specify nil for
the types list to allow any type (but still include validators). To say a var must be nil,
specify types as (nil).

If there are still elements in VAR-SPEC, they are validator forms. Each is evaluated and
if any return nil, throw an error."
      (setq var (pop! 'var-spec)
            types (pop! 'var-spec))
      (if (not (listp types))
          (setq types (list types)))
      (append (if (member 'non-nil types)
                  (progn
                    (remove! 'types '(= x 'non-nil))
                    (#! '((assert (quote #var))))))
              (if types
                  (#! '((assert '(member (type #var) (quote #types))))))
              (mapcar '(lambda(v) (#! '(assert (quote #v)))) var-spec)))
    (defun-r 'defun-r--process-declare-var)

    (defun-q defun-r--process-declare-tests (tests / test-sym)
      "Adds tests for defining function. Nothing is added to function body

If *test-build-p* is nil, nothing happens. Otherwise, all forms are gathered together into
a test function (see %defun-t) that is automatically eval'd.

This is useful in that it adds expected behavior to docstrings automatically. Tests
specified this way should be short and easy to read -- ideally the function call wrapped
in an =/equal with its expected output. Tests that require setup/teardown should be
defined using %defun-t."
      (if *test-build-p*
          (progn
            (setq test-sym (read (strcat "autotest-" (to-string func-sym))))
            (eval (list 'defun-q
                        test-sym nil
                        "declare autotest"
                        (list 'assert-all (list 'quote tests))))
            (defun-t test-sym)))
      nil)
    (defun-r 'defun-r--process-declare-tests)

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; documentation system now up and running ;;
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;;;;;;;;;;;;;;;;;;;;;
    ;; query functions ;;
    ;;;;;;;;;;;;;;;;;;;;;

    (defun-q whatis (func-sym)
      "Prints the one-line description of the function FUNC-SYM points to

Returns nil if FUNC-SYM wasn't defined using the special documentation system."
      (declare (vars (func-sym (sym))))
      (doc--print-docstring func-sym "short"))
    (defun-r 'whatis)

    (defun-q docstring (func-sym)
      "Prints the full docstring of the function FUNC-SYM points to

Returns nil if FUNC-SYM wasn't defined using the special documentation system."
      (declare (vars (func-sym (sym))))
      (doc--print-docstring func-sym "full"))
    (defun-r 'docstring)

    (defun-q doc--print-docstring (func-sym which / dstr)
      "Print the docstring for FUNC-SYM from WHICH, which is either \"short\" or \"long\"

Will retrieve docstring from *docstrings-short* or *docstrings-long*, respectively. If
FUNC-SYM isn't valid, return nil."
      (setq dstr (cdr (assoc func-sym (symcat '("*docstrings-" #which "*")))))
      (if dstr
          (progn
            (princ (strcat (cdr (assoc func-sym *signatures*)) "\n" dstr))
            (princ))))
    (defun-r 'doc--print-docstring)

    ;;;;;;;;;;;;;;;;;;;;;
    ;; build functions ;;
    ;;;;;;;;;;;;;;;;;;;;;

    (defun-q doc-build ()
      "Bootstraps doc build process and opens a new blank drawing to carry it out"
      (propagate '*doc-build-p* t)
      (propagate '*test-build-p* t)
      (vla-Activate (vla-Add (vla-get-Documents acadObj)))
      (princ))
    (defun-r 'doc-build)

    (defun-q doc-build-html ()
      "Builds HTML docs from files.json (generated by %doc--write-json)"
      (doc--write-json)
      (princ "\nBuilding HTML files")
      (startapp (findfile "docs/code-manual/build/dist/build.exe")
                (strcat psc-src-dir "docs/code-manual/"))
      (startapp (findfile "docs/user-manual/build-commands/dist/build-commands.exe")
                (strcat psc-src-dir "docs/user-manual/")))
    (defun-r 'doc-build-html)

    (defun-q doc--write-json (/ f)
      "Writes file, function, and command docs to file

All currently generated file and function documentation lives in the global variable
*file-docs*. Its contents are formatted in JSON and written to
docs/code-manual/files.json

Command documentation lives in *command-docs*. Contents are written to
docs/user-manual/commands.json"
      (princ "\nWriting files.json")
      (setq f (open (findfile "docs/code-manual/files.json") "w"))
      (princ (doc--json-backend-docs) f)
      (close f)
      
      (princ "\nWriting commands.json")
      (setq f (open (findfile "docs/user-manual/commands.json") "w"))
      (princ (doc--json-frontend-docs) f)
      (close f))
    (defun-r 'doc--write-json)

    (defun-q doc--json-backend-docs ()
      "Returns the full documentation as a JSON string"
      (strcat "["
              (strjoin (mapcar 'doc--json-file-docs
                               (vl-sort *file-docs* (sort-by 'car '<)))
                       ",")
              "]"))
    (defun-r 'doc--json-backend-docs)

    (defun-q doc--json-file-docs (file-entry / file-name)
      "Returns file documentation as a JSON string"
      (setq file-name (pop! 'file-entry))
      (json-object
       (list (json-string-prop "name" file-name)
             (json-string-prop "doc_short" (pop! 'file-entry))
             (json-string-prop "doc_full" (apply (compose '(escape-newlines escape-quotes))
                                                 (list (pop! 'file-entry))))
             (json-array-prop "functions" (mapcar 'doc--json-func-docs
                                                  (vl-sort file-entry (sort-by 'car '<)))))))
    (defun-r 'doc--json-file-docs)

    (defun-q doc--json-func-docs (func-entry / name priv-p)
      "Returns function documentation as a JSON string"
      (declare (vars (func-entry list)
                     (file-name str)))
      (setq name (pop! 'func-entry)
            priv-p (wcmatch name "*--*"))
      (json-object
       (list (json-string-prop "file" file-name)
             (json-string-prop "name" name)
             (json-string-prop "access" (if priv-p "priv" "pub"))
             (json-string-prop "args" (pop! 'func-entry))
             (json-string-prop "sig" (pop! 'func-entry))
             (json-string-prop "doc_short" (pop! 'func-entry))
             (json-string-prop "doc_full" (pop! 'func-entry)))))
    (defun-r 'doc--json-func-docs)

    (defun-q doc--json-frontend-docs ()
      "Returns documentation for all commands as a JSON string"
      (strcat "["
              (strjoin (mapcar 'doc--json-command-docs
                               (vl-sort *command-docs* (sort-by 'car '<)))
                       ",")
              "]"))
    (defun-r 'doc--json-frontend-docs)

    (defun-q doc--json-command-docs (cmd-entry / name)
      "Returns documentation for a single command as a JSON string"
      (setq name (pop! 'cmd-entry))
      (json-object
       (list (json-string-prop "name" name)
             (json-string-prop "doc_short" (pop! 'cmd-entry))
             (json-string-prop "doc_full" (pop! 'cmd-entry)))))
    (defun-r 'doc--json-command-docs)))
;; end of *dev-mode* stuff

;; process functions defined earlier
;; see beginning of acaddoc.lsp for early defun-r definition
(map-apply '(lambda (*file-name* f-sym) (defun-r f-sym))
           *delayed-doc*)
(setq *delayed-doc* nil)
