(set-file-docstring
 "Error-handling functions")

(defun-q debug-print (msgs)
  "Prints debug messages when *dev-mode* is non-nil

MSGS is a list of things that will be %macro-expand'd, turned into strings, concatenated,
and printed to console. A newline will be added to the front of the resulting string
automatically."
  (if *dev-mode*
      (princ (apply 'strcat (cons "\n" (mapcar 'to-string (macro-expand msgs))))))
  (princ))
(defun-r 'debug-print)

(defun-q debug-print-vars (vars)
  "Uses %debug-print to print symbol values

VARS should be a list of symbols. Example use:

(setq *dev-mode* t a 1 b 2)
(debug-print-vars '(a b))

prints

A: 1
B: 2"
  (declare no-func-name)
  (debug-print (list (strjoin
                      (mapcar '(lambda(x)
                                (strcat (vl-symbol-name x) ": " (to-string (vl-symbol-value x))))
                              vars)
                      "\n")
                     "\n")))
(defun-r 'debug-print-vars)

(defun-q error (msg / )
  "Calls *error* if defined, prints MSG, and exits

If *error-prefix* is defined, it is used as the message prefix. Otherwise, the default
[Error] is used."
  (setq *last-error* msg)
  (if *error* (*error* ""))
  (*push-error-using-stack*)
  (defun *error* (x / prefix)
    (setq prefix (?? *error-prefix* "[Error]"))
    (if (not *no-print-errors*)
        (vla-Prompt (vla-get-Utility acadDoc)
                    (strcat "\n" prefix " " msg "\n")))
    (setq *error* nil))
  (exit))
(defun-r 'error)

(defun-q catch-all-error (attempt)
  "Wrapper for vl-catch-all-error-message

This function returns *last-error* as the error message if the error was a quit/exit (which should indicate we threw it using %error)."
  (if (= (vl-catch-all-error-message attempt)
         "quit / exit abort")
      *last-error*
      (vl-catch-all-error-message attempt)))
(defun-r 'catch-all-error)

(defun-q assert (expr / val)
  "Throws an %error if EXPR evaluates to nil

If *assert-return-error* is non-nil, return the error string rather than
throwing. Otherwise, returns the value of EXPR."
  (if (setq val (eval expr))
      (if *assert-return-error* nil val)
      (if *assert-return-error*
          (strcat "[Assert failed] " (to-string expr))
          ((lambda(*error-prefix*)
             (error (to-string expr)))
           (if *func-name*
               (strcat "[Assert failed (" *func-name* ")]")
               "[Assert failed]")))))
(declare-late 'assert '(no-func-name))
(defun-r 'assert)

(defun-q assert-all (expr-lst)
  "%assert all expressions in EXPR-LST. If no errors, return a list of results

If *ASSERT-RETURN-ERROR* is non-nil, catch the errors and return them as a list. If there
were no errors in this case, return nil."
  (strip-nil
   (mapcar 'assert expr-lst)))
(declare-late 'assert-all '(no-func-name))
(defun-r 'assert-all)
