(set-file-docstring
 "Function-handling functions")

(defun-q function-p (x)
  "Returns nil unless X is a function"
  (declare (tests (function-p +)
                  (not (function-p '(lambda() 1)))
                  (not (function-p 123))))
  (member (type x) '(SUBR USUBR)))
(defun-r 'function-p)

(defun-q function-or-lambda-p (x / xt)
  "Returns nil unless X is a function or a lambda"
  (declare (tests (function-or-lambda-p +)
                  (function-or-lambda-p '(lambda() 1))
                  (not (function-or-lambda-p 123))))
  (setq xt (type x))
  (cond ((function-p x) t)
        ((= xt 'list)
         (and (= (car x) 'lambda)
              (listp (cadr x))
              (> (length x) 2)))
        ((= xt 'sym) (function-p (vl-symbol-value x)))))
(defun-r 'function-or-lambda-p)

(defun-q attempt-apply (func args)
  "If FUNC is a function, returns the result of applying it to ARGS. Else, return nil"
  (declare (vars (func) (args (list nil))))
  (if (= (type func) 'SYM)
      (setq func (vl-symbol-value func)))
  (if (and (function-p func)
           (member (type args) '(LIST nil)))
      (apply 'func args)))
(defun-r 'attempt-apply)

(defun-q compose (funcs-lst)
  "Returns a lambda that is the result of composing functions in FUNCS-LST

FUNCS-LST should be a list of symbols pointing to single-argument functions. The resulting
function ill take a single argument and return the result of applying all the functions in
FUNCS-LST in the reverse of the order specified.

Example:
(compose '(lowercase to-string)) returns a function that take an argument and
returns the result of applying %to-string to it, followed by %lowercase."
  (if (null funcs-lst)
      '(lambda(itself) itself)
      (macro-expand '(lambda(compose-x)
                      (#(car funcs-lst) (#(compose (cdr funcs-lst)) compose-x))))))
(defun-r 'compose)

(defun-q apply-by-twos (func lst / arg1 arg2 ret)
  "Applies FUNC to adjacent elements of LST, accumulates and returns the results.

Example:
(apply-by-twos '+ '(1 2 3)) => '(3 5)

which is equivalently shown as

(list (+ 1 2) (+ 2 3))"
  (declare (vars (func nil (function-or-lambda-p func))
                 (lst nil (listp lst)))
           (tests (equal (apply-by-twos '+ '(1 2 3)) '(3 5))))
  (setq arg1 (car lst)
        arg2 (cadr lst)
        lst (cddr lst))
  (while arg2
    (add-to-list 'ret (apply func (list arg1 arg2)))
    (setq arg1 arg2
          arg2 (car lst)
          lst (cdr lst)))
  ret)
(defun-r 'apply-by-twos)

(defun-q map-apply (f map)
  "Returns the results of applying F to each element of MAP

MAP should be a list of lists of arguments to F."
  (mapcar '(lambda(f args) (apply f args))
          (lst* f (length map))
          map))
(declare-late 'map-apply
 '((vars (f nil (function-or-lambda-p f))
         (map nil (listp map)))
   (tests (equal (map-apply '+ '((1 2) (3 4))) '(3 7)))))
(defun-r 'map-apply)

(defun-q map-apply! (f map-sym)
  "Updates the list at MAP-SYM in place using %map-apply"
  (declare (vars (f nil (function-or-lambda-p f))
                 (map-sym sym (listp (vl-symbol-value map-sym)))))
  (set map-sym (map-apply f (vl-symbol-value map-sym))))
(defun-r 'map-apply!)
