(set-file-docstring
 "Functions related to points, vectors, and planes")

(defun-q calc-segs (pts / ret)
  "Returns a list of segment lengths between points in PTS

Holds the minimum starting segment at 6\" and the ending segment at 18\" for pipe
bending."
  (declare (vars (pts nil (listp pts) (vl-every 'point-p pts))))
  (add-to-list 'ret (max 6 (distance (nth 0 pts) (nth 1 pts))))
  (foreach i (range 1 (- (length pts) 2) 1)
           (add-to-list 'ret (distance (nth i pts) (nth (1+ i) pts))))
  (add-to-list 'ret (max 18 (distance (circular-nth -2 pts)
                                      (circular-nth -1 pts)))))
(defun-r 'calc-segs)

(defun-q calc-angles (pts / a b c ang normal ret)
  "Returns a list of internal angles between PTS"
  (while (> (length pts) 2)
    (setq ang (take 3 pts)
          normal (apply 'normal-vec ang)
          ang (abs (rtd (apply 'int-ang (mapcar '(lambda(pt) (trans pt 0 normal)) ang)))))
    (pop! 'pts)
    (add-to-list 'ret (- 180 ang)))
  ret)
(defun-r 'calc-angles)

(defun-q dihedral (v1 v2 / ax ay az bx by bz)
  "Returns the dihedral angle between two planes defined as normal vectors"
  (declare (vars (v1 list
                     (vl-every 'numberp v1)
                     (= (length v1) 3))
                 (v2 list
                     (vl-every 'numberp v2)
                     (= (length v2) 3))))
  (setq ax (nth 0 v1)
        ay (nth 1 v1)
        az (nth 2 v1)
        bx (nth 0 v2)
        by (nth 1 v2)
        bz (nth 2 v2))
  (acos (/ (abs (+ (* ax bx) (* ay by) (* az bz)))
           (* (sqrt (apply '+ (mapcar 'square (list ax ay az))))
              (sqrt (apply '+ (mapcar 'square (list bx by bz))))))))
(defun-r 'dihedral)

(defun-q normal-vec (pt1 pt2 pt3)
  "Returns the 3D vector normal to 3 points"
  (cross-product (vec- pt2 pt1) (vec- pt2 pt3)))
(defun-r 'normal-vec)

(defun-q vec- (v1 v2)
  "Returns the result of subtracting vector V1 from V2"
  (declare (vars (v1 list
                     (vl-every 'numberp v1)
                     (= (length v1) 3))
                 (v2 list
                     (vl-every 'numberp v2)
                     (= (length v2) 3))))
  (list (- (nth 0 v1) (nth 0 v2))
        (- (nth 1 v1) (nth 1 v2))
        (- (nth 2 v1) (nth 2 v2))))
(defun-r 'vec-)

(defun-q cross-product (v1 v2 / ax ay az bx by bz)
  "Returns the cross product of vectors V1 and V2"
  (declare (vars (v1 list
                     (vl-every 'numberp v1)
                     (= (length v1) 3))
                 (v2 list
                     (vl-every 'numberp v2)
                     (= (length v2) 3))))
  (setq ax (nth 0 v1)
        ay (nth 1 v1)
        az (nth 2 v1)
        bx (nth 0 v2)
        by (nth 1 v2)
        bz (nth 2 v2))
  (list (- (* ay bz) (* az by))
        (- (* az bx) (* ax bz))
        (- (* ax by) (* ay bx))))
(defun-r 'cross-product)

(defun-q pipe-bend-lines (segments-lst / segment ret radius k)
  "Returns the bend lines for a PSC pipe with segments SEGMENTS-LST

Each element to SEGMENTS-LST should be a real number. Assumes 1.66OD pipe, centerline
radius of 2.83\" and k value of 0.18."
  (setq radius 2.83
        k 0.18)
  (if (/= (rem (length segments-lst) 2) 0)
      (append! 'segments-lst '(0)))
  
  (defun segment (len ang prev / rang arclen htan flatten straight mark)
    (setq rang (dtr ang)
          arclen (* radius rang)
          htan (tan (/ rang 2.0))
          flatten (* k (/ ang 90.0))
          straight (- len (+ (* radius htan) flatten)))
    (if prev
        (setq straight (- straight
                          (+ (* radius
                                (xd-value "htan" prev))
                             (xd-value "flatten" prev)))
              mark (+ straight
                      (xd-value "mark" prev)
                      (xd-value "arclen" prev)))
        (setq mark straight))
    (#! '(("mark" #mark)
          ("ang" #ang)
          ("arclen" #arclen)
          ("htan" #htan)
          ("flatten" #flatten))))

  (setq ret (list (segment (car segments-lst)
                           (cadr segments-lst)
                           nil))
        segments-lst (cddr segments-lst)
        i -1)
  (while segments-lst
    (add-to-list 'ret (segment (car segments-lst)
                               (cadr segments-lst)
                               (nth (inc! 'i) ret)))
    (setq segments-lst (cddr segments-lst)))

  (reverse
   (cdr
    (reverse
     (apply 'append
            (mapcar '(lambda(x) (list (frac (xd-value "mark" x))
                                      (rtos (xd-value "ang" x) 2 1)))
                    ret))))))
(defun-r 'pipe-bend-lines)

(defun-q pipe-bend-lines-padded (segments-lst / cut-len)
  "Wrapper for %pipe-bend-lines that always returns a 9-item list to match fab drawing tables"
  (setq segments-lst (reverse (pipe-bend-lines segments-lst))
        cut-len (pop! 'segments-lst))
  (reverse
   (cons cut-len
         (?? (repeat (- 8 (length segments-lst))
                     (setq segments-lst (cons "" segments-lst)))
             segments-lst))))
(defun-r 'pipe-bend-lines-padded)

(defun-q within-box-p (pt boxpt1 boxpt2 / ptx ptx b1x b1y b2x b2y lx rx ty by)
  "Returns T if PT is within the box defined by BOXPT1 and BOXPT2"
  (declare (vars (pt list (point-p pt))
                 (boxpt1 list (point-p boxpt1))
                 (boxpt2 list (point-p boxpt2)))
           (tests (within-box-p '(0 0) '(-1 -1) '(1 1))
                  (not (within-box-p '(10 10) '(0 0) '(1 1)))))
  (setq ptx (car pt)
        pty (cadr pt)
        b1x (car boxpt1)
        b1y (cadr boxpt1)
        b2x (car boxpt2)
        b2y (cadr boxpt2)
        lx (min b1x b2x)
        rx (max b1x b2x)
        by (min b1y b2y)
        ty (max b1y b2y))
  (and (>= ptx lx)
       (<= ptx rx)
       (>= pty by)
       (<= pty ty)))
(defun-r 'within-box-p)

(defun-q point-p (pt)
  "Returns T if PT is a 2- or 3-item list of numbers"
  (and (= (type pt) 'list)
       (member (length pt) '(2 3))
       (vl-every 'numberp pt)))
(defun-r 'point-p)

(defun-q pts-list-p (pts-list)
  "Returns T if PTS-LIST is a list of points"
  (and (listp pts-list)
       (vl-every 'point-p pts-list)))
(defun-r 'pts-list-p)

(defun-q mid-2-pts (p1 p2)
  "Returns a 2D point which is midway between P1 and P2"
  (mapcar '+ p1 (mapcar '* '(0.5 0.5 1.0) (mapcar '- p2 p1))))
(defun-r 'mid-2-pts)

(defun-q int-ang (p1 p2 p3 / ret)
  "Return the internal angle of 3 points in radians"
  (abs (angle-between (angle p2 p1) (angle p2 p3))))
(defun-r 'int-ang)

(defun-q bend-deduct (thick ang rad k)
  "Calculates a bend deduction

THICK is the material thickness. ANG is the excluded bend angle in degrees. RAD is the
bend radius. K is the K-value."
  (- (* 2 (tan (/ (dtr ang) 2.0)) (+ rad thick))
     (dtr (* ang (+ rad (* thick k))))))
(defun-r 'bend-deduct)

(defun-q rect-pts (width height ins)
  "Returns the points list for a rectangle located at point INS"
  (declare (vars (width nil (numberp width))
                 (height nil (numberp height))
                 (ins list (point-p ins)))
           (tests (equal (rect-pts 1 1 '(0 0))
                         '((0 0) (1 0) (1 1) (0 1)))
                  (equal (rect-pts 1 2 '(3 4))
                         '((3 4) (4 4) (4 6) (3 6)))))
  (displace-pts (list '(0 0 0)
                      (list width 0 0)
                      (list width height 0)
                      (list 0 height 0))
                ins))
(defun-r 'rect-pts)

(defun-q dirmirror (pts)
  "Return PTS mirrored according to DIR"
  (declare (vars (pts list)))
  (if (= dir -1)
      (mirror-pts-v pts 0)
      pts))
(defun-r 'dirmirror)

(defun-q mirror-pts-v (pts x)
  "Returns points list PTS mirrored over vertical line at X"
  (declare (vars (pts list (vl-every 'point-p pts))
                 (x nil (numberp x)))
           (tests (equal (mirror-pts-v '((0 0) (1 1)) 0)
                         '((0 0) (-1 1)))))
  (mapcar '(lambda(pt) (cons (+ x (- x (car pt)))
                             (cdr pt)))
          pts))
(defun-r 'mirror-pts-v)

(defun-q mirror-pts-h (pts y)
  "Returns points list PTS mirrored over horizontal line at Y"
  (declare (vars (pts list (vl-every 'point-p pts))
                 (y nil (numberp y)))
           (tests (equal (mirror-pts-h '((0 0) (1 1)) 0)
                         '((0 0) (1 -1)))))
  (mapcar '(lambda(pt) (append (list (car pt)
                                     (+ y (- y (cadr pt))))
                               (cddr pt)))
          pts))
(defun-r 'mirror-pts-h)

(defun-q add-bulges (pts)
  "Returns point list PTS with all points padded to 3D"
  (mapcar '(lambda(pt)
            (if (= (length pt) 2)
                (append pt (list 0))
                pt))
          pts))
(defun-r 'add-bulges)

(defun-q calc-bulge (ang)
  "Returns the bulge value corresponding to a fillet of ANG radians"
  (setq ang (/ ang 4.0))
  (/ (sin ang)
     (cos ang)))
(defun-r 'calc-bulge)

(defun-q angle-between (a1 a2 / diff)
  "Returns the angle between angles A1 and A2 (radians)"
  (setq a1 (rtd a1)
        a2 (rtd a2)
        diff (- a1 a2))
  (dtr (cond
         ((> diff 180)
          (- (- 360) (- a2 a1)))

         ((< diff -180)
          (- 360 (- a2 a1)))

         (T diff))))
(defun-r 'angle-between)

(defun-q calc-fillet (pt1
               pt2
               pt3
               radius
               /
               l1-angle
               l1-len
               l2-angle
               l2-len
               pull-back
               int-angle
               arc
               chord
               pull
               )
  "Calculates a fillet for three 2D points at RADIUS

All points should be 2D. The return value will be a list of four 3D points, with the Z
coordinate representing the bulge."
  (declare (vars (pt1 list (point-p pt1))
                 (pt2 list (point-p pt2))
                 (pt3 list (point-p pt3))
                 (radius nil (numberp radius))))

  (defun pull-back (pt1 pt2 dist / a)
    (setq a (angle pt1 pt2))
    (list (- (car pt2)
             ;; rounding
             (read (rtos (* dist (cos a)) 2 5)))
          (- (cadr pt2)
             (read (rtos (* dist (sin a)) 2 5)))))

  (foreach pt '(pt1 pt2 pt3)
           (if (> (length (vl-symbol-value pt)) 2)
               (set pt (reverse (cdr (reverse (vl-symbol-value pt)))))))
  (setq l1-angle (angle pt2 pt1)
        l1-len (distance pt1 pt2)
        l2-angle (angle pt2 pt3)
        l2-len (distance pt2 pt3)
        int-angle (angle-between l1-angle l2-angle)
        int-angle (cond ((> int-angle 0) (- pi int-angle))
                        ((< int-angle 0) (- (- pi) int-angle)))
        arc (* radius (abs int-angle))
        chord (/ (* 2
                    radius
                    (sin (/ arc (* 2 radius))))
                 2.0)
        pull (/ chord
                (cos (/ (abs int-angle)
                        2.0))))

  (if (>= pull (min l1-len l2-len))
      (error "Fillet radius too large"))

  (list (append pt1 '(0))
        (append (pull-back pt1 pt2 pull)
                (list (calc-bulge int-angle)))
        (append (pull-back pt3 pt2 pull)
               '(0))
        (append pt3 '(0))))
(defun-r 'calc-fillet)

(defun-q fillet-all-pts (pts radius / i ncar ncdr f-pts)
  "Returns points list PTS with all corners filleted to RADIUS

See %calc-fillet for individual fillet calculations."
  (defun ncar (n lst / i return)
    (setq i 0)
    (while (< i n)
      (add-to-list 'return (nth i lst))
      (inc! 'i))
    return)

  (defun ncdr (n lst / i return)
    (setq i n)
    (while (< i (length lst))
      (add-to-list 'return (nth i lst))
      (inc! 'i))
    return)
  
  (setq i 0)
  (while (< i (- (length pts) 2))
    (setq f-pts (calc-fillet (nth i pts)
                             (nth (+ i 1) pts)
                             (nth (+ i 2) pts)
                             radius)
          pts (append (ncar i pts)
                      f-pts
                      (ncdr (+ i 3) pts))
          i (+ i 2)))
  pts)
(defun-r 'fillet-all-pts)

(defun-q norm_3pts (org xdir ydir / norm)
  "Returns a normal vector based on an origin x vector and y vector"
  (declare (vars (org list point-p) (xdir list point-p) (ydir list point-p)))
  (foreach v '(xdir ydir) (set v (mapcar '- (eval v) org)))
  (if (inters org xdir org ydir)
      (mapcar '(lambda (x) (/ x (distance '(0 0 0) norm)))
              (setq norm
		    (list
		     (- (* (cadr xdir) (caddr ydir)) (* (caddr xdir) (cadr ydir)))
		     (- (* (caddr xdir) (car ydir)) (* (car xdir) (caddr ydir)))
		     (- (* (car xdir) (cadr ydir)) (* (cadr xdir) (car ydir))))))))
(defun-r 'norm_3pts)

(defun-q safe-calc-fillet-3d (pts-lst rad / rad ret *no-print-errors*)
  "Repeats %calc-fillet-3d with reduced radius until no error is returned"
  (declare
    (vars
      (pts-lst list (vl-every 'point-p pts-lst) (vl-every 'cddr pts-lst))
      (rad nil (numberp rad))
    ) ;_ vars
  ) ;_ declare
  (setq *no-print-errors* t)
  (while
      (vl-catch-all-error-p
       (setq ret (vl-catch-all-apply 'calc-fillet-3d (list pts-lst rad)))
       ) ;_ vl-catch-all-error-p
    (setq rad (- rad 0.0625))
    ) ;_ while
  ret)
(defun-r 'safe-calc-fillet-3d)

(defun-q calc-fillet-3d (pts-lst rad / xyb-add-z xdir org ydir nor f-ret pts-ret)
  "Returns a definition list formatted for use with %draw-fillet-3d access with %safe-calc-fillet-3d"
  (defun xyb-add-z (xyb)
    (mapcar
     '(lambda (xyb)
       (add-to-list-at-pos 'xyb (caddr (trans (car pts-lst) 0 nor)) 2)
       ) ;_ lambda
     xyb
     ) ;_ mapcar
    ) ;_ defun xyb-add-z
  (if (and (pts-list-p pts-lst) (= (length pts-lst) 2))
      (setq pts-ret
            (list
             (list
              (mapcar '(lambda (p) (append p '(0.0))) pts-lst)
              '(0 0 1)
              ) ;_ list
             ) ;_ list
	    ) ;_ setq
      ) ;_ if
  (while (cddr pts-lst)
    (setq
     xdir (car pts-lst)
     org (cadr pts-lst)
     ydir (caddr pts-lst)
     nor (norm_3pts org xdir ydir)
     ) ;_ setq
    (setq f-ret
          (apply 'calc-fillet
		 (append
		  (mapcar '(lambda (p) (trans p 0 nor)) (list xdir org ydir))
		  (list rad)
		  ) ;_ append
		 ) ;_ apply
	  ) ;_ setq
    (add-to-list 'pts-ret
		 (list
		  (xyb-add-z
		   (if (= 3 (length pts-lst)) f-ret (remove-last f-ret))
		   ) ;_ xyb-add-z
		  nor
		  ) ;_ list
		 ) ;_ add-to-list
    (setq pts-lst
          (cons
           (mapcar '+
		   (car pts-lst)
		   (trans
		    (mapcar '- (caddr f-ret) (car f-ret))
		    (cadar (reverse pts-ret))
		    0
		    ) ;_ trans
		   ) ;_ mapcar
           (cddr pts-lst)
           ) ;_ cons
	  ) ;_ setq
    ) ;_ while
  pts-ret)
(defun-r 'calc-fillet-3d)

(defun-q rezero-pts (pts v / delta dx dy)
  "Returns points list PTS recentered on the Vth vertex

Rotates the list around until V is the 0th item and displaces the points so they are
correct. Adds Z coordinates if 2D."
  (declare (vars (pts nil (pts-list-p pts))
                 (v int (>= v 0) (<= v (length pts))))
           (tests (equal (rezero-pts '((0 0) (1 1)) 1) '((0 0 0) (-1 -1 0)))))
  (setq pts (if (= (length (car pts)) 2) (add-bulges pts) pts)
        delta (nth v pts)
        dx (car delta)
        dy (cadr delta))
  (circular-shift
   (mapcar '(lambda(pt) (list
                         (- (car pt) dx)
                         (- (cadr pt) dy)
                         (caddr pt)))
           pts)
   v))
(defun-r 'rezero-pts)

(defun-q displace-pt (pt d)
  "Returns point PT displaced by point D"
  (declare (vars (pt (list nil) (if pt (point-p pt) t))
                 (d (list nil) (if d (point-p d) t)))
           (tests (equal (displace-pt '(0 0) '(2 2)) '(2 2))
                  (equal (displace-pt '(1 2) '(3 4)) '(4 6))))
  (mapcar '+ pt d))
(defun-r 'displace-pt)

(defun-q displace-pts (pts d)
  "Returns points list PTS with all points displaced by point D"
  (declare (vars (pts nil (pts-list-p pts))
                 (d nil (point-p d)))
           (tests (equal (displace-pts '((0 0) (1 1)) '(2 2)) '((2 2) (3 3)))))
  (mapcar '(lambda(x) (mapcar '+ x d)) pts))
(defun-r 'displace-pts)

(defun-q rotate-pt (pt ang / x y)
  "Returns point PT rotated about the origin by ANG (radians)

Adds Z coordinate if 2D."
  (declare (vars (pt nil (point-p pt))
                 (ang nil (numberp ang))))
  (setq x (car pt)
        y (cadr pt))
  (list (- (* x (cos ang)) (* y (sin ang)))
        (+ (* y (cos ang)) (* x (sin ang)))
        (?f (caddr pt))))
(defun-r 'rotate-pt)

(defun-q rotate-pts (pts ang)
  "Return points list PTS with all points rotated about the orgin by ANG (radians)

Adds Z coordinates if 2D."
  (mapcar '(lambda(x) (rotate-pt x ang)) pts))
(defun-r 'rotate-pts)

(defun-q join-vertices (lines / tmp ret)
  "Joins a list of lines

LINES should be a list of lines (lists of points). If the last point of one line equals the
first point of the next, they are joined into one. Otherwise, they are left separate."
  (declare (vars (lines nil (listp lines)
                            (vl-every '(lambda(line) (vl-every 'point-p line))
                                      lines)))
           (tests (equal (join-vertices '(((0 0) (1 0))
                                          ((1 0) (1 1))
                                          ((1 2) (0 2))))
                         '(((0 0) (1 0) (1 1))
                           ((1 2) (0 2))))))
  (while lines
         (setq tmp (pop! 'lines))
         (if (equal (car tmp) (last (last ret)))
             (subst! (append (last ret) (list (cadr tmp))) (last ret) 'ret)
             (append! 'ret (list tmp))))
  (if (and (equal (caar ret) (last (last ret))) (> (length ret) 1))
      (progn
        (setq tmp (cdar ret) ret (cdr ret))
        (subst! (append (last ret) tmp) (last ret) 'ret)))
  ret)
(defun-r 'join-vertices)
