(set-file-docstring
 "Functions for creating JSON lists

See the following for usage:
- %doc--json-backend-docs
- %doc--json-file-docs
- %doc--json-func-docs
- %doc--json-frontend-docs
- %doc--json-command-docs")

(defun-q json-prop (key value)
  "Return KEY and VALUE as a JSON property"
  (declare (vars (key str)
                 (value str)))
  (strcat "\"" key "\"" ":" value))
(defun-r 'json-prop)

(defun-q json-string-prop (key value)
  "Return KEY and VALUE as a JSON string property"
  (declare (vars (key str)
                 (value str)))
  (json-prop key (strcat "\"" value "\"")))
(defun-r 'json-string-prop)

(defun-q json-array (items)
  "Returns a JSON array made up of ITEMS"
  (declare (vars (items list (vl-every 'stringp items))))
  (strcat "[" (strjoin items ",") "]"))
(defun-r 'json-array)

(defun-q json-array-prop (key value)
  "Return KEY and VALUE as a JSON array property"
  (declare (vars (key str)
                 (value list (vl-every 'stringp value))))
  (json-prop key (json-array value)))
(defun-r 'json-array-prop)

(defun-q json-object (props)
  "Combine the JSON properties in PROPS into an object and return it"
  (declare (vars (props list (vl-every 'stringp props))))
  (strcat "{" (strjoin props ",") "}"))
(defun-r 'json-object)

(defun-q json-block (ename / ent data block-prop)
  "Returns the JSON representation of a stair or landing block"
  (defun block-prop (kv / k v)
    (setq k (car kv) v (cadr kv))
    (if (= (type v) 'safearray)
        (setq v "SAFEARRAY"))
    (if (stringp v)
        (json-string-prop k (escape-quotes v))
        (json-prop k (to-string v))))
  (setq data (read-xdata-migrate ename "")
        ent (entget ename))
  (json-object
   (append
    (list
     (json-string-prop "layer" (cdr (assoc 8 ent)))
     (json-string-prop "ins" (strjoin (mapcar 'to-string (first-two (cdr (assoc 10 ent)))) ",")))
    (mapcar 'block-prop data))))
(defun-r 'json-block)

(defun-q json-dump-file (/ filename f blocks)
  "Writes a JSON dump file for a drawing"
  (setq filename (string-subst-all "/" "\\"
                   (strcat (vla-get-Path acadDoc) "/"
                           (vl-string-subst ".json" ".dwg" (vla-get-Name acadDoc))))
        blocks (ss->lst (ssget "A"
                               (dxf-filter
                                '((appid "Stair,Land,Landing")))))
        blocks (if blocks
                   (json-array-prop
                    "blocks"
                    (mapcar 'json-block (vl-sort blocks 'sort-stairland)))))
  (if (or blocks)
      (progn
        (setq f (open filename "w"))
        (defun *error* (msg)
          (princ (strcat "\nerror: " msg))
          (close f)
          (setq *error* nil))
        (write-line (json-object
                     (list
                      (json-string-prop "filename" filename)
                      (?s blocks)))
                    f)
        (close f))))
(defun-r 'json-dump-file)
