(set-file-docstring
 "List-handling functions")

(defun-q pop! (pop-lst-sym / pop-lst ret)
  "Sets LST-SYM to (cdr LST) and returns (car LST)"
  (setq pop-lst (vl-symbol-value pop-lst-sym)
        ret (car pop-lst))
  (set pop-lst-sym (cdr pop-lst))
  ret)
(declare-late 'pop!
  '((vars (pop-lst-sym sym))
    (tests
     (equal ((lambda(lst) (pop! 'lst)) nil) nil)
     (equal ((lambda(lst) (pop! 'lst)) '(1)) 1)
     (equal ((lambda(lst) (pop! 'lst) lst) '(1 2)) '(2)))))
(defun-r 'pop!)

(defun-q rcons (lst x)
  "Returns LST with X added to the end"
  (append lst (list x)))
(declare-late 'rcons
 '((vars
    (lst nil (listp lst)))
   (tests
    (equal (rcons '(1) 2) '(1 2))
    (equal (rcons nil 1) '(1)))))
(defun-r 'rcons)

(defun-q take (n lst / ret)
  "Returns the first N elements of LST"
  (setq n (min n (length lst)))
  (while (> n 0)
    (add-to-list 'ret (car lst))
    (setq lst (cdr lst))
    (dec! 'n))
  ret)
(declare-late 'take
 '((vars (n int) (lst nil (listp lst)))
   (tests
    (equal (take 1 '(1 2)) '(1))
    (equal (take 2 '(1)) '(1))
    (= (take 1 nil) nil))))
(defun-r 'take)

(defun-q take-while (predicate lst / x ret)
  "Returns the first elements of LST for which PREDICATE is true

PREDICATE should be a function that takes one element."
  (while lst
         (setq x (car lst)
               lst (cdr lst))
         (if (apply predicate (list x))
             (add-to-list 'ret x)
             (setq lst nil)))
  ret)
(declare-late 'take-while
 '((vars (predicate (list sym) (if (listp predicate)
                                   (= (car predicate) 'lambda)
                                   (function-p (vl-symbol-value predicate))))
         (lst (list nil) (listp lst)))
   (tests (equal (take-while 'numberp '(1 2 nil 3)) '(1 2)))))
(defun-r 'take-while)

(defun-q take! (n take-lst-sym / take-lst ret)
  "Returns the first N elements from TAKE-LST-SYM, removing them"
  (declare (vars (n (int))
                 (take-lst-sym sym (listp (vl-symbol-value take-lst-sym))))
           (tests
            (equal ((lambda(lst) (take! 1 'lst)) '(1 2)) '(1))
            (equal ((lambda(lst) (take! 1 'lst) lst) '(1 2)) '(2))))
  (setq take-lst (vl-symbol-value take-lst-sym)
        n (min n (length take-lst)))
  (while (> n 0)
    (add-to-list 'ret (car take-lst))
    (setq take-lst (cdr take-lst))
    (dec! 'n))
  (set take-lst-sym take-lst)
  ret)
(defun-r 'take!)

(defun-q zip (l1 l2 / ret)
  "Returns lists L1 and L2 interleaved"
  (declare (vars (l1 list (>= (length l1) (length l2)))
                 (l2 (list nil)))
           (tests (equal (zip '(1 2) '(a b)) '(1 a 2 b))
                  (equal (zip '(1 2 3) '(a)) '(1 a 2 3))))
  (while l1
    (add-to-list 'ret (pop! 'l1))
    (if l2 (add-to-list 'ret (pop! 'l2))))
  ret)
(defun-r 'zip)

(defun-q add-to-alist (alist-sym key value append-p / alist old oldval oldtype newtype)
  "Adds an entry for KEY to ALIST pointed to by ALIST-SYM

If ALIST already contains an entry for KEY, replace it. If so, and the following
conditions are met, append to the value of the entry rather than replacing:

1. APPEND-P must be non-nil
2. Original entry value and new value types must match
3. Types must be STR or LIST"
  (setq alist (vl-symbol-value alist-sym)
        old (assoc key alist)
        oldval (if old (cdr old) nil)
        oldtype (type oldval)
        newtype (type value))
  (if old
      (set alist-sym (subst (cons key (if (and append-p
                                               (= oldtype newtype)
                                               (member oldtype '(STR LIST)))
                                          ((if (= oldtype 'STR) strcat append) oldval value)
                                        value))
                            old
                            alist))
      (add-to-list alist-sym (cons key value))))
(declare-late 'add-to-alist
 '((vars (alist-sym sym (sym-lst-p alist-sym)))
   (tests (equal ((lambda(/ lst) (add-to-alist 'lst 'a 1 nil)))
                 '((a . 1)))
          (equal ((lambda(lst) (add-to-alist 'lst 'a 1 nil))
                  '((a . 2)))
                 '((a . 1)))
          (equal ((lambda(lst) (add-to-alist 'lst 'a ", world" t))
                  '((a . "hello")))
                 '((a . "hello, world"))))))
(defun-r 'add-to-alist)

(defun-q add-to-list (list-sym elt)
  "Adds ELT to list referred to by LIST-SYM"
  (set list-sym (append (vl-symbol-value list-sym) (list elt))))
(declare-late 'add-to-list
 '((vars (list-sym sym (sym-lst-p list-sym)))
   (tests (equal ((lambda(/ lst) (add-to-list 'lst 1))) '(1))
          (equal ((lambda(lst) (add-to-list 'lst 2)) '(1)) '(1 2)))))
(defun-r 'add-to-list)

(defun-q add-to-list-at-pos (lst-sym elt pos / beg end)
  "Adds ELT to list referred to by LST-SYM at POS"
  (declare (vars (lst-sym sym (sym-lst-p lst-sym))
                 (pos int))
           (tests (equal ((lambda(lst) (add-to-list-at-pos 'lst nil 1)) '(1 2)) '(1 nil 2))))
  (split-list (eval lst-sym) '(< (vl-position x (vl-symbol-value lst-sym)) pos) 'beg 'end)
  (append beg (list elt) end))
(defun-r 'add-to-list-at-pos)

(defun-q remove-nth (n lst / i)
  "Returns LST with Nth item removed"
  (setq i -1)
  (vl-remove-if '(lambda (x) (= (inc! 'i) n)) lst))
(declare-late 'remove-nth
 '((vars (n int)
         (lst nil (listp lst)))
   (tests (equal (remove-nth 1 '(1 2 3)) '(1 3)))))
(defun-r 'remove-nth)

(defun-q subst! (new old lst-sym)
  "Updates list at LST-SYM in place using subst"
  (set lst-sym (subst new old (vl-symbol-value lst-sym))))
(declare-late 'subst!
 '((vars (lst-sym sym (sym-lst-p lst-sym)))
   (tests (equal ((lambda(lst) (subst! 'a 1 'lst) lst) '(1)) '(a))
          (equal ((lambda(lst) (subst! 'a 1 'lst) lst) '(2)) '(2)))))
(defun-r 'subst!)

(defun-q remove (lst f)
  "Returns LST with only elements for which F returns nil

F should be a quoted list like '(= X 1), where X is the element in question. F will be
processed by %macro-expand."
  (declare (vars (lst nil (listp lst))
                 (f nil (listp f)))
           (tests (equal (remove '(1 2 3) '(> x 1)) '(1))
                  (equal (remove '(1 2 3) '(< x 1)) '(1 2 3))))
  (vl-remove-if (append '(lambda (x))
                        (list (macro-expand f)))
                lst))
(defun-r 'remove)

(defun-q remove! (lst-sym f)
  "Updates list at LST-STM in place using %remove"
  (declare (vars (lst-sym sym (sym-lst-p lst-sym))
                 (f nil (listp f)))
           (tests (equal ((lambda(lst) (remove! 'lst '(> x 1)) lst) '(1 2 3)) '(1))
                  (equal ((lambda(lst) (remove! 'lst '(< x 1)) lst) '(1 2 3)) '(1 2 3))))
  (set lst-sym (remove (vl-symbol-value lst-sym) f)))
(defun-r 'remove!)

(defun-q filter (lst f)
  "Returns LST without elements for which F returns nil

F should be a quoted list like '(= X 1), where X is the element in question. F will be
processed by %macro-expand."
  (declare (vars (lst nil (listp lst))
                 (f nil (listp f)))
           (tests (equal (filter '(1 2 3) '(< x 2)) '(1))
                  (equal (filter '(1 2 3) '(> x 0)) '(1 2 3))))
  (vl-remove-if-not (append '(lambda (x))
                            (list f))
                    lst))
(defun-r 'filter)

(defun-q filter! (lst-sym f)
  "Updates list at LST-STM in place using %filter"
  (declare (vars (lst-sym sym (sym-lst-p lst-sym))
                 (f nil (listp f))))
  (set lst-sym (vl-remove-if-not (append '(lambda (x))
                                         (list f))
                                 (vl-symbol-value lst-sym))))
(defun-r 'filter!)

(defun-q split-list (lst f true-lst-sym false-lst-sym / )
  "Puts items for which F returns nil in FALSE-LST and the rest in TRUE-LST

See %remove and %filter for how to format F."
  (declare (vars (lst nil (listp lst))
                 (true-lst-sym sym (sym-lst-p true-lst-sym))
                 (false-lst-sym sym (sym-lst-p false-lst-sym))))
  (set true-lst-sym (filter lst f))
  (set false-lst-sym (remove lst f)))
(defun-r 'split-list)

(defun-q split-list-n (lst n first-lst-sym second-lst-sym / tmp)
  "Splits LST on index N, putting first half in FIRST-LST and the rest in SECOND-LST"
  (declare (vars (lst nil (listp lst))
                 (n int)
                 (first-lst-sym sym (sym-lst-p first-lst-sym))
                 (second-lst-sym sym (sym-lst-p second-lst-sym))))
  (repeat n
          (add-to-list 'tmp (car lst))
          (setq lst (cdr lst)))
  (set first-lst-sym tmp)
  (set second-lst-sym lst))
(defun-r 'split-list-n)

(defun-q sort! (lst-sym f)
  "Sorts LST in place using vl-sort"
  (declare (vars (lst-sym sym (sym-lst-p lst-sym))
                 (f nil (function-or-lambda-p f)))
           (tests (equal ((lambda(lst) (sort! 'lst '<) lst) '(2 1 3)) '(1 2 3))))
  (set lst-sym (vl-sort (vl-symbol-value lst-sym) f)))
(defun-r 'sort!)

(defun-q safe-nth (n lst / )
  "Wrapper for nth that returns nil when LST is nil

Built-in nth errors in that case."
  (declare (vars (n int)
                 (lst nil (listp lst)))
           (tests (= (safe-nth 1 '(1 2)) 2)
                  (= (safe-nth 1 nil) nil)))
  (if lst
      (nth n lst)))
(defun-r 'safe-nth)

(defun-q uniquify (lst / ret)
  "Returns LST without duplicate items"
  (declare (vars (lst nil (listp lst)))
           (tests '(equal (uniquify '(1 2 3 1 2)) '(1 2 3))))
  (foreach elt lst
           (if (not (member elt ret))
               (add-to-list 'ret elt)))
  ret)
(defun-r 'uniquify)

(defun-q mapcar! (f lst-sym)
  "Updates list at LST-SYM in place using mapcar"
  (declare (vars (f nil (function-or-lambda-p f))
                 (lst-sym sym (sym-lst-p lst-sym)))
           (tests (equal ((lambda(lst) (mapcar! '1+ 'lst) lst) '(1 2 3)) '(2 3 4))))
  (set lst-sym (mapcar f (vl-symbol-value lst-sym))))
(defun-r 'mapcar!)

(defun-q map-append (f lst)
  "Alias for (apply 'append (mapcar f lst))"
  (declare (vars (f nil (function-or-lambda-p f))
                 (lst nil (listp lst)))
           (tests (equal (map-append 'reverse '((1 2) (3 4))) '(2 1 4 3))))
  (apply 'append (mapcar f lst)))
(defun-r 'map-append)

(defun-q test-for (test-type lst)
  "Returns T if all elements of LST are of type TEST-TYPE"
  (declare (vars (test-type sym)
                 (lst nil (listp lst)))
           (tests (test-for 'int '(1 2 3))
                  (test-for 'sym '(a b c))
                  (not (test-for 'int '(1 2 3.0)))))
  (apply 'and (mapcar '(lambda (x) (= test-type (type x))) lst)))
(defun-r 'test-for)

(defun-q append! (lst-sym app-lst / )
  "Updates LST-SYM in place using append"
  (declare (vars (lst-sym sym (sym-lst-p lst-sym))
                 (app-lst nil (listp app-lst)))
           (tests (equal ((lambda(lst) (append! 'lst '(3)) lst) '(1 2)) '(1 2 3))))
  (set lst-sym (append (eval lst-sym) app-lst)))
(defun-r 'append!)

(defun-q rpt-lst (n ele / lst)
  "Returns a list consisting of ELE repeated N times

Duplicate (args reversed) of %lst*."
  (declare (vars (n int))
           (tests (equal (rpt-lst 3 'q) '(q q q))))
  (repeat n (append! 'lst (list ele))))
(defun-r 'rpt-lst)

(defun-q lst* (lstelt n / ret)
  "Returns a list consisting of ELE repeated N times

Duplicate (args reversed) of %rpt-lst."
  (repeat n (add-to-list 'ret lstelt)))
(defun-r 'lst*)
(declare-late 'lst*
 '((vars (n int))
   (tests (equal (lst* 'q 3) '(q q q)))))

(defun-q remove-last (lst)
  "Returns LST without its last element"
  (declare (vars (lst nil (listp lst)))
           (tests (equal (remove-last '(1 2 3)) '(1 2))))
  (reverse (cdr (reverse lst))))
(defun-r 'remove-last)

(defun-q circular-nth (n lst / len)
  "Wrapper for nth that treats LST as periodic and infinite

If a negative is supplied for N, index from the end of LST. If N is greater than the
length of LST, wrap around to the beginning again as many times as necessary."
  (declare (vars (n int)
                 (lst nil (listp lst)))
           (tests (= (circular-nth 1 '(1 2 3)) 2)
                  (= (circular-nth -1 '(1 2 3)) 3)
                  (= (circular-nth 3 '(1 2 3)) 1)
                  (= (circular-nth 6 '(1 2 3)) 1)))
  (setq len (length lst))
  (while (< n 0) (setq n (+ n len)))
  (while (>= n len) (setq n (- n len)))
  (nth n lst))
(defun-r 'circular-nth)

(defun-q circular-shift (lst n / tmplst)
  "Returns LST left-shifted N times"
  (declare (vars (lst nil (listp lst))
                 (n int))
           (tests (equal (circular-shift '(1 2 3) 1) '(2 3 1))
                  (equal (circular-shift '(1 2 3) 2) '(3 1 2))))
  (while (> n 0)
    (add-to-list 'tmplst (car lst))
    (setq lst (cdr lst)
          n (1- n)))
  (append lst tmplst))
(defun-r 'circular-shift)

(defun-q find-first (pred lst / i)
  "Returns the first element in LST for which PRED returns non-nil

PRED should be a function of one argument."
  (declare (vars (pred nil (function-or-lambda-p pred))
                 (lst nil (listp lst)))
           (tests (= (find-first '(lambda(a) (> a 1)) '(1 2 3)) 2)))
  (setq i -1)
  (while (not (apply pred (list (nth (inc! 'i) lst)))))
  (nth i lst))
(defun-r 'find-first)

(defun-q count-items (ci-lst / keys ret)
  "Returns a list of counts of items in CI-LST

Each itme in the return list is a 2-item list whose car is the element and whose cadr is
the count of that element in CI-LST."
  (declare (vars (ci-lst nil (listp ci-lst)))
           (tests (equal (count-items '(1)) '((1 1)))
                  (equal (count-items '(1 1 2 2 2)) '((1 2) (2 3)))))
  (foreach x ci-lst
           (if (member x keys)
               (subst! (list x (1+ (xd-value x ret))) (assoc x ret) 'ret)
               (progn
                 (add-to-list 'keys x)
                 (add-to-list 'ret (list x 1)))))
  ret)
(defun-r 'count-items)

(defun-q range (from to by / ret)
  "Returns a list of numbers that range from FROM to TO by step BY

TO is exclusive"
  (declare (vars (from nil (numberp from))
                 (to nil (numberp to))
                 (by nil (numberp by)))
           (tests (equal (range 1 4 1) '(1 2 3))
                  (equal (range 0.75 1.25 0.125) '(0.75 0.875 1.0 1.125))))
  (setq from (- from by))
  (while (< (setq from (+ from by)) to)
         (add-to-list 'ret from)))
(defun-r 'range)

(defun-q cons-ind-lst (len i / ret)
  "Returns a range of integers starting at I (optional) of length LEN, by 1"
  (declare (vars (len int)
                 (i (int nil)))
           (tests (equal (cons-ind-lst 3 nil) '(0 1 2))
                  (equal (cons-ind-lst 3 5) '(5 6 7))))
  (setq i (?n i))
  (repeat len (add-to-list 'ret i) (setq i (1+ i)))
  ret)
(defun-r 'cons-ind-lst)

(defun-q 2-item-list (lst / )
  "Returns the elements of LST grouped by twos"
  (declare (vars (lst nil (listp lst)))
           (tests (equal (2-item-list '(1 2 a b)) '((1 2) (a b)))
                  (equal (2-item-list '(1 2 a)) '((1 2) (a nil)))))
  (if lst
      (cons (list (car lst) (cadr lst))
            (2-item-list (cddr lst)))))
(defun-r '2-item-list)

(defun-q 3-item-list (lst / )
  "Returns the elements of LST grouped by threes"
  (declare (vars (lst nil (listp lst)))
           (tests (equal (3-item-list '(1 2 3 a b c)) '((1 2 3) (a b c)))
                  (equal (3-item-list '(1 2 3 a b)) '((1 2 3) (a b nil)))))
  (if lst
      (cons (list (car lst) (cadr lst) (caddr lst))
            (3-item-list (cdddr lst)))))
(defun-r '3-item-list)

(defun-q first-two (lst)
  "Returns the first two elements of LST"
  (declare (vars (lst nil (listp lst)))
           (tests (equal (first-two '(1 2 3)) '(1 2))))
  (list (car lst)
        (cadr lst)))
(defun-r 'first-two)

(defun-q strip-nil (lst)
  "Returns LST with nil elements removed"
  (declare (vars (lst nil (listp lst)))
           (tests (equal (strip-nil '(1 nil 2)) '(1 2))
                  (equal (strip-nil '(1 2 3)) '(1 2 3))))
  (remove lst '(null x)))
(defun-r 'strip-nil)

(defun-q slice (lst start len)
  "Return a slice from LST of length LEN starting at START"
  (declare (vars (lst nil (listp lst))
                 (start int)
                 (len int))
           (tests (equal (slice '(1 2 3 4) 0 4) '(1 2 3 4))
                  (equal (slice '(1 2 3 4) 1 2) '(2 3))))
  (take! start 'lst)
  (take len lst))
(defun-r 'slice)

(defun-q member* (target list2)
  "Returns T if any members of TARGET appear in LST

TARGET can be a list or atom."
  (declare (vars (list2 nil (listp list2)))
           (tests (member 'a '(a b))
                  (not (member 'c '(a b)))
                  (member* '(a 1) '(a b))
                  (member* '(a 1) '(1 2))
                  (not (member* '(a b) '(1 2)))))
  (if (listp target)
      (vl-member-if '(lambda(x) (member x list2)) target)
      (member target list2)))
(defun-r 'member*)

(defun-q list-of-strings-p (x)
  "Returns T if X is a list of strings"
  (and (listp x) (vl-every 'stringp x)))
(defun-r 'list-of-strings-p)
