(set-file-docstring
 "VLA Object-handling functions")

(defun-q add-calcd-objs (container extrude-p defs / otype ret)
  "Processes calculated objects for 2D and 3D blocks

CONTAINER should be the target block definition or model space. If EXTRUDE-P is non-nil,
creates 3D solids. Otherwise, creates a 2D wireframe. Additionally, when EXTRUDE-P is
non-nil, the object will be translated into the perspective used in 3D drawings.

DEFS is a list of calculated objects. Each is a list whose car is a data list of object
properties and whose cdr is the particular object definition. The value of the type key in
the properties determines which handling function is called. The naming scheme for
handling functions is 'add-calcd-' followed by the type value which triggers it.

Available handling functions:
- %add-calcd-poly
- %add-calcd-pipe

Common properties:
- ins: insert point (all pts in definition displaced relative to ins)
- layer: layer to put obj on
- extrude: extrusion distance"
  (declare (vars (container VLA-Object)
                 (defs nil (listp defs))))
  (foreach def defs
           (setq otype (xd-value "type" (car def)))
           (cond ((= otype "poly") (add-to-list 'ret (add-calcd-poly container extrude-p def)))
                 ((= otype "pipe") (append! 'ret (add-calcd-pipe container extrude-p def)))))
  ret)
(defun-r 'add-calcd-objs)

(defun-q add-calcd-poly (container extrude-p def / data obj obj2 ins)
  "Adds a calculated polyline to CONTAINER

CONTAINER should be the target block definition or model space. DEF should be a list whose
car is a data list of properties and whose cdr is a list of points representing a closed
polyline. If EXTRUDE-P is non-nil, the closed polyline is extruded by the value of the
\"extrude\" property."
  (setq data (car def)
        ins (?? (xd-value "ins" data) '(0 0 0))
        obj (add-polyline container
                          (if extrude-p ins (first-two ins))
                          (cdr def)
                          T))
  (if extrude-p
      (progn
	(setq obj (plineObj->solidObj obj (xd-value "extrude" data) T))
	(vla-Rotate3D obj (3dpt 0 0) (3dpt 1 0) (dtr 90))))
  ((lambda(layer)
     (if layer (vla-put-Layer obj layer)))
   (xd-value "layer" data))
  obj)
(defun-r 'add-calcd-poly)

(defun-q add-calcd-pipe (container extrude-p def / data cope pts obj obj1 obj2)
  "Adds a calculated pipe to CONTAINER

Pipe objects represent bent pipe handrails. They are represented as open polylines. This
function fillets the points list and either offsets or extrudes the pipe.

Addition properties available to pipes:
- radius: fillet radius
- bulge: used as 4th argument to %offset&close"
  (setq
    data (car def)
    pts  (vl-remove-if '(lambda (a) (vl-every 'listp a)) (cdr def))
    cope (vl-remove-if '(lambda (a) (vl-every 'numberp a)) (cdr def))
  ) ;_ setq
  ( (lambda (ins)
      (if ins (setq pts (displace-pts pts ins)))
      (if (and ins cope)
        (map-apply! '(lambda (p v) (list (displace-pt p ins) v)) 'cope)
      ) ;_ if
    ) ;_ lambda
    (xd-value "ins" data)
  ) ;_ lambda-func
  (if extrude-p
    ( (lambda (pts)
        (setq obj (vla-Add3DPoly container (vlist->safearray pts)))
        (vla-put-Layer obj "NO PLOT")
        (setq obj2
          (draw-pipe-3d
            (if (> (length pts) 2)
              (draw-fillet-3d container (safe-calc-fillet-3d pts 2.83))
              (list (vla-Add3DPoly container (vlist->safearray pts)))
            ) ;_ if
            (xd-value "radius" data)
          ) ;_ draw-pipe-3d
        ) ;_ setq
        (if (assoc "layer" data)
          (foreach obj obj2 (vla-put-Layer obj (xd-value "layer" data)))
        ) ;_ if
        (if cope
          (progn
            (setq obj1 (car obj2) obj2 (cdr obj2))
            (append! 'obj2 (map-apply 'add-cope cope))
            (mapcar '(lambda (tmp) (vla-boolean obj1 acunion tmp)) obj2)
            (list obj obj1)
          ) ;_ progn
          (list obj obj2)
        ) ;_ if
      ) ;_ lambda
      (mapcar '(lambda (p) (trans p '(0 -1 0) 0)) pts)
    ) ;_ lambda-func
    (progn
      (setq
        pts  (fillet-all-pts (uniquify (mapcar 'first-two pts)) 2.83)	
        obj  (add-polyline container '(0 0 0) pts nil)
        obj2 (offset&close obj (xd-value "radius" data) T (xd-value "bulge" data))
      ) ;_ setq
        (if (assoc "layer" data) (vla-put-Layer obj2 (xd-value "layer" data)))
        (vla-Delete obj)
      ( (lambda (obj rad) (map-apply 'add-cope cope))
          obj2 (xd-value "radius" data)
      ) ;_ lambda-func
      (list obj2)
    ) ;_ progn
  ) ;_ if
) ;_ defun-q add-calcd-pipe
(defun-r 'add-calcd-pipe)

(defun-q add-cope (vtx vec / var->lst get-idx-int get-off-val)
  "Called from %add-calcd-pipe when a cope definition is included in the points list.
A cope definition is a two item list containing a point on the open polyline where
the centerlines of a coped connection would intersect and a vector representing the
direction the coped piece would extend from the intersection point.

When extrude-p is non-nil a short 3d cylinder is unioned to
the 3dsolid at each passed point representing a coped intersection

When extrude-p is nil two vertices and a bulge are added to
a polyline at each passed point representing a coped intersection"
  (defun var->lst (var)
    (vlax-safearray->list (vlax-variant-value var))
  ) ;_ defun var->lst
  (defun get-idx-int (pt1 pt2 / int)
    (if (setq int (inters pt1 pt2 vtx end T))
      (setq ret-lst (list idx int)) (inc! 'idx)
    ) ;_ if
  ) ;_ defun get-idx-int
  (defun get-off-val (id1 id2 / dir)
    (apply
     '(lambda (p1 p2)
        (setq dir (mapcar 'unit-dir (mapcar '- p2 p1)))
        (setq dir (mapcar '* dir (list rad rad)))
      ) ;_ lambda
      (mapcar
       '(lambda (i) (var->lst (vla-get-Coordinate obj i)))
        (list id1 id2)
      ) ;_ mapcar
    ) ;_ apply
  ) ;_ defun get-off-val
  (if extrude-p
    ( (lambda ( / ins1 nor1 obj1)
        (setq ins1 (trans vtx '(0 -1 0) 0))
        (setq nor1 (trans vec '(0 -1 0) 0))
        (setq obj1 (vla-addcircle container (3dpt 0 0 0) 0.83))
        (vla-put-normal obj1 (3dpt nor1))
        (vla-move obj1 (3dpt 0 0) (3dpt ins1))
        (setq obj1 (plineobj->solidobj obj1 0.83 T))
      ) ;_ lambda
    ) ;_ lambda-func
    ( (lambda ( / idx end ret-lst)
        (setq
          idx 0
          end (mapcar '+ vtx (mapcar '* vec (lst* (* 2.0 rad) 2)))
        ) ;_ setq
        (apply-by-twos 'get-idx-int
          (2-item-list (var->lst (vla-get-coordinates obj)))
        ) ;_ apply-by-twos
        (apply
         '(lambda (idx int / arr off)
            (setq off (get-off-val idx (1+ idx)))
            (setq arr (vlax-make-safearray vlax-vbDouble '(0 . 1)))
            (setq arr (vlax-safearray-fill arr (mapcar '- int off)))
            (vla-AddVertex obj (inc! 'idx) arr)
            (vla-SetBulge obj idx -0.542)
            (setq arr (vlax-make-safearray vlax-vbDouble '(0 . 1)))
            (setq arr (vlax-safearray-fill arr (mapcar '+ int off)))
            (vla-AddVertex obj (inc! 'idx) arr)
          ) ;_ lambda
          ret-lst
        ) ;_ apply
      ) ;_ lambda
    ) ;_ lambda-func
  ) ;_ if
) ;_ defun-q add-cope
(defun-r 'add-cope)

(defun-q vla-mirror-y (obj)
  "Mirrors VLA-object OBJ over Y-axis using a transformation matrix"
  (declare (vars (obj vla-object)))
  (vla-TransformBy obj
                   (vlax-tmatrix '((-1 0 0 0)
                                   (0  1 0 0)
                                   (0  0 1 0)
                                   (0  0 0 1)))))
(defun-r 'vla-mirror-y)

(defun-q vla-mirror-x (obj)
  "Mirrors VLA-object OBJ over x-axis using a transformation matrix"
  (declare (vars (obj vla-object)))
  (vla-TransformBy obj
                   (vlax-tmatrix '((1  0 0 0)
                                   (0 -1 0 0)
                                   (0  0 1 0)
                                   (0  0 0 1)))))
(defun-r 'vla-mirror-x)

(defun-q mirror&delete (obj pt axis / pt2 ret)
  "Mirrors OBJ over AXIS at PT, deletes the original, and returns the copy

OBJ should be a VLA-object. Axis should be one of the symbols 'x or 'y. PT should be a
point."
  (declare (vars (obj vla-object)
                 (pt list (point-p pt))
                 (axis sym (member axis '(x y)))))
  (setq pt2 (3dpt (displace-pt pt (if (= axis 'x) '(1 0) '(0 1))))
        ret (vla-Mirror obj (3dpt pt) pt2))
  (vla-delete obj)
  ret)
(defun-r 'mirror&delete)

(defun object-load (name)
  (psc-load (strcat "util/object/" name ".lsp")))

(mapcar 'object-load '(
		       "3D"		; solids, regions, extrusion and subtraction
		       "block"		; block definitions and references
		       "line"		; lines, polylines
		       "table"		; tables
		       "text"		; text, mtext, mleaders
		       ))
