  (set-file-docstring "Functions related to creation of 3D objects")

  (defun-q closed-pline-p (plineObj)
    "Returns T if passed PLINEOBJ is a closed polyline"
    (declare (vars (plineObj vla-object)))
    (or
      (= "AcDbCircle" (vla-get-objectname plineObj))
      (and
        (member
          (vla-get-objectname plineObj)
         '("AcDbPolyline" "AcDb3dPolyline")
        ) ;_ member
        (= :vlax-true (vla-get-closed plineObj))
      ) ;_ and
    ) ;_ or
  ) ;_ defun-q closed-pline-p
  (defun-r 'closed-pline-p)

  (defun-q path-p (obj)
    "Returns T if passed OBJ is a valid path"
    (declare (vars (obj vla-object)))
    (member
      (vla-get-objectname obj)
     '("AcDbLine" "AcDbPolyline" "AcDb3dPolyline" "AcDbSpline"
       "AcDbArc" "AcDbCircle" "AcDbEllipse")
    ) ;_ member
  ) ;_ defun-q path-p
  (defun-r 'path-p)

  (defun-q plineObj->solidObj (plineObj ext-param delete / regionObj solidObj)
    "Creates extruded 3D solid from passed polyline object

PLINEOBJ must be a VLA-OBJ which is closed
EXT-PARAM must be a number for depth or a VLA-OBJECT for path
PLINEOBJ will be deleted if DELETE is non-nil"
    (declare
      (vars
        (plineObj vla-object (closed-pline-p plineObj))
        (ext-param nil (or (numberp ext-param) (path-p ext-param)))
      ) ;_ vars
    ) ;_ declare
    (setq regionObj (plineObj->regionObj plineObj delete))
    (setq solidObj
    	(if (numberp ext-param)
        (vla-addextrudedsolid
          (block-containing regionObj)
          regionObj
          ext-param
          0
        ) ;_ vla-add-extrudedsolid
        (vla-addextrudedsolidalongpath
          (block-containing regionObj)
          regionObj
          ext-param
        ) ;_ vla-AddExtrudedSolidAlongPath
  	  ) ;_ if
  	) ;_ setq
    (vla-delete regionObj)
    (if (and delete (not (numberp ext-param))) (vla-delete ext-param))
    (if (and delete (not (vlax-erased-p plineObj))) (vla-delete plineObj))
    solidObj
  ) ;_ defun-q plineObj->solidObj
  (defun-r 'plineObj->solidObj)

  (defun-q plineObj->regionObj (plineObj delete / regionObj)
    "Creates a region from passed polyline object

PLINEOBJ must be a VLA-OBJ which is closed
PLINEOBJ will be deleted if DELETE is non-nil"
    (declare (vars (plineObj vla-object (closed-pline-p plineObj))))
    (setq regionObj
      (vlax-safearray-get-element
        (vlax-variant-value
          (vla-AddRegion
            (car
              (vl-remove-if-not
               '(lambda (obj)
                  (= (vla-get-ownerid plineObj) (vla-get-objectid obj))
                ) ;_ lambda
                (collection->list (vla-get-Blocks acadDoc))
              ) ;_ vl-remove-if-not
            ) ;_ car
            (vlax-safearray-fill
              (vlax-make-safearray vlax-vbObject '(0 . 0))
              (list PlineObj)
            ) ;_ vlax-safearray-fill
          ) ;_ vla-AddRegion
        ) ;_ vlax-variant-value
        0
      ) ;_vlax-safearray-get-element
    ) ;_ setq
    (if (and delete (not (vlax-erased-p plineObj))) (vla-delete plineObj))
    regionObj
  ) ;_ defun-q plineObj->regionObj
  (defun-r 'plineObj->regionObj)

  (defun-q plineObj->faceObj (plineObj dpt delete / vlist faceObj)
    "creates 3D face from an open pline input of exactly two vertices"
    (declare
      (vars
        (plineObj
          vla-object
          (= "AcDbPolyline" (vla-get-objectname plineObj))
          ( (lambda (lst) (= 2 (length (safearray->list lst))))
            (variant-value (vla-get-coordinates plineobj))
          ) ;_ lambda
        ) ;_ plineObj-vars
      ) ;_ vars
    ) ;_ declare
    (setq vlist
      (safearray->list (variant-value (vla-get-coordinates plineobj)))
      vlist
      (append
        (mapcar '(lambda (p) (append p '(0.0))) vlist)
        (mapcar '(lambda (p) (append p (list dpt))) (reverse vlist))
      ) ;_ append
    ) ;_ setq
    (setq faceObj
      (apply 'vla-add3dface
        (append (list (block-containing plineObj)) (mapcar '3dpt vlist))
      ) ;_ apply
    ) ;_ setq
    (if (and delete (not (vlax-erased-p plineObj))) (vla-delete plineObj))
    faceObj
  ) ;_ defun-q plineObj->faceObj
  (defun-r 'plineObj->faceObj)

  (defun-q put-circle-normal-to-path (obj path / start 3dpos xord yord zord pt)
    "Move a passed circle object normal to a passed path object"
    (declare
      (vars
        (obj vla-object (= "AcDbCircle" (vla-get-objectname obj)))
        (path vla-object (path-p path))
      ) ;_ vars
    ) ;_ declare
    (setq start (vlax-curve-getstartpoint path))
    (setq 3dpos
      (vlax-curve-getFirstDeriv
        path
        (vlax-curve-getParamAtPoint path start)
      ) ;_ vlax-curve-getFirstDeriv
    ) ;_ setq
    (setq
      xord (if (zerop (car 3dpos)) 0.0 (/ (car 3dpos) 1000))
      yord (if (zerop (cadr 3dpos)) 0.0 (/ (cadr 3dpos) 1000))
      zord (if (zerop (caddr 3dpos)) 0.0 (/ (caddr 3dpos) 1000))
      3dpos (list xord yord zord)
      pt (vla-get-center obj)
    ) ;_ setq
    (vla-move obj pt (3dpt start))
    (vlax-put-property-r
      obj
     'normal
      (vlax-3D-point 3dpos)
    ) ;_ vla-put-normal
  ) ;_ defun-q put-circle-normal-to-path
  (defun-r 'put-circle-normal-to-path)

  (defun-q draw-fillet-3d (container def-lst / xyzb-rem-z ret)
    "Creates a series of filleted plines from a definition list

DEF-LST is the return from %safe-calc-fillet-3d"
    (declare
      (vars
        (container vla-object)
        (def-lst
          list
          (vl-every ;vertices
           '(lambda (lst) (vl-every 'point-p (mapcar 'remove-last lst)))
            (mapcar 'car def-lst)
          ) ;_ mapcar
          (vl-every ;bulge
           '(lambda (lst) (vl-every 'numberp (mapcar 'last lst)))
            (mapcar 'car def-lst)
          ) ;_ mapcar
          (vl-every 'point-p (mapcar 'last def-lst)) ;norm
        ) ;_ def-lst-vars
      ) ;_ vars
    ) ;_ declare
    (defun xyzb-rem-z (xyzb)
      (mapcar '(lambda (xyzb) (remove-nth 2 xyzb)) xyzb)
    ) ;_ defun xyzb-rem-z
    (foreach def def-lst
      (apply
       '(lambda (lst nor / obj)
          (setq obj
            (add-polyline container '(0 0 0) (xyzb-rem-z lst) nil)
          ) ;_ setq
          (vla-put-normal obj (3dpt nor))
          (vla-move
            obj
            (3dpt (vlax-curve-getstartpoint obj))
            (3dpt (trans (remove-last (car lst)) nor 0))
          ) ;_ vla-move
          (add-to-list 'ret obj)
        ) ;_ lambda
        def
      ) ;_ apply
    ) ;_ foreach
  ) ;_ defun-q draw-fillet-3d
  (defun-r 'draw-fillet-3d)

  (defun-q draw-pipe-3d (path-lst rad / cen cir ret)
    "Create a series of extruded circles along list of path objects"
    (declare
      (vars
        (path-lst
          nil
          (vl-every '(lambda (obj) (= (type obj) 'vla-object)) path-lst)
        ) ;_ path-lst-vars
      ) ;_ vars
    ) ;_ declare
    (foreach path-obj path-lst
      (setq
        cen (vlax-curve-getStartPoint path-obj)
        cir (vla-addCircle (block-containing path-obj) (3dpt cen) rad)
      ) ;_ setq
      (put-circle-normal-to-path cir path-obj)
      (add-to-list 'ret (plineObj->solidObj cir path-obj T))
    ) ;_ foreach
    ret
  ) ;_ defun draw-pipe-3d
  (defun-r 'draw-pipe-3d)
