(set-file-docstring
 "Polyline handling functions")

(defun-q add-polyline (container ins vertices closed-p / points pline i)
  "Adds a LWPolyline to CONTAINER at INS

INS should be a 2D or 3D point. VERTICES should be a list of points. Z-coordinates in
VERTICES are treated as bulge values (all points are 2D). Points without Z-coordinates
have 0 added to the end.

If CLOSED-P is non-nil, set the Closed property to True."
  (declare (vars (container vla-object)
                 (ins list (point-p ins))
                 (vertices list (vl-every 'point-p vertices))))
  (setq vertices (add-bulges vertices)
        points (vlist->safearray (mapcar '(lambda(x) (reverse (cdr (reverse x)))) vertices))
        pline (vla-AddLightweightPolyline container points)
        i 0)
  (mapcar '(lambda (x)
             (vla-SetBulge pline i (caddr x))
             (setq i (1+ i)))
          vertices)
  (vla-put-Closed pline (if closed-p :vlax-true :vlax-false))
  (vla-Move pline (vlax-3d-point 0 0 0)
                  (apply 'vlax-3d-point ins))
  (vla-Update pline)
  pline)
(defun-r 'add-polyline)

(defun-q add-rectangle (container width height ins)
  "Adds a rectangular LWPolyline to CONTAINER using %add-polyline

INS is at the bottom-left corner of the rectangle. To put rectangles in other quadrants,
supply negative values for WIDTH, HEIGHT, or both."
  (declare (vars (container vla-object)
                 (width nil (numberp width))
                 (height nil (numberp width))
                 (ins list (point-p ins))))
  (add-polyline container
                ins
                (rect-pts width height '(0 0 0))
                t))
(defun-r 'add-rectangle)

(defun-q offset-polyline (pline dist / offsetObj)
  "Offsets PLINE by DIST and returns the resulting polyline

Essentially a wrapper for VLA-Offset that cracks the acorn and extracts the delicious
offset polyline hidden within a hard outer shell of safearray."
  (declare (vars (pline vla-object)
                 (dist nil (numberp dist))))
  (setq offsetObj (vlax-safearray-get-element (vlax-variant-value (vla-Offset pline dist))
                                              0)))
(defun-r 'offset-polyline)

(defun-q offset&close (pline dist dual bulge / offsetObj offset-pts n-pts offset-bulges)
  "Offsets PLINE by DIST and joins both lines into a closed polygon

DIST being negative means to offset in the negative direction. If DUAL is non-nil, offset
in both directions and deletes the original line.

If BULGE is non-nil, either or both of the closing lines will be bulged out to a
semicircle. If BULGE is a non-nil scalar, both ends will be bulged. BULGE may also be a
2-item list, with the boolean value of the car indicating whether to bulge the end at the
head of the original line, and the cadr meaning the tail end."
  (declare (vars (pline vla-object)
                 (dist nil (numberp dist))))
  (if dual (setq pline (offset-polyline pline dist)
                 dist (* dist -2)))
  (if (numberp bulge) (setq bulge (list bulge bulge)))
  (setq offsetObj (offset-polyline pline dist)
        offset-pts (reverse (safearray->list (vlax-variant-value (vla-get-Coordinates offsetObj))))
        n-pts (length (safearray->list (vlax-variant-value (vla-get-Coordinates pline))))
        last-vertex (1- n-pts))
  (if bulge
      (vla-SetBulge pline last-vertex (cadr bulge)))
  (setq v-offset 2)
  (foreach pt offset-pts
           (vla-AddVertex pline (inc! 'last-vertex) (vlist->safearray (list pt)))
           (if (not (= pt (last offset-pts)))
               (progn
                 (vla-SetBulge pline last-vertex (- (vla-GetBulge pline (- last-vertex v-offset))))
                 (setq v-offset (+ v-offset 2)))))
  (vla-put-Closed pline :vlax-true)
  (if bulge
      (vla-SetBulge pline last-vertex (car bulge)))
  (vla-Delete offsetObj)
  (vla-Update pline)
  pline)
(defun-r 'offset&close)

(defun-q dualoffset&delete (pline dist bulge / offsetObj)
  "Convenience wrapper for %offset&close

Supplies t for DUAL argument, deletes the original polyline, and returns the resulting
offset polyline."
  (declare (vars (pline vla-object)
                 (dist nil (numberp dist))))
  (setq offsetObj (offset&close pline dist T bulge))
  (vla-Delete pline)
  offsetObj)
(defun-r 'dualoffset&delete)

(defun-q add-slot (container ins len dia / obj ret)
  "Adds a slot polyline to CONTAINER

INS is in the middle of the centerline and the slot extends horizontally to the left and
right by half of LEN. LEN describes the extents of the hole, including radius, not just
the centerline."
  (declare (vars (container vla-object)
                 (ins list (point-p ins))
                 (len nil (numberp len))
                 (dia nil (numberp dia))))
  (setq pts (mapcar '(lambda (x) (list (* (- len dia) x) 0 0))
                    '(0.5 -0.5))
        obj (add-polyline container ins pts nil)
        ret (offset&close obj (/ dia 2.0) T 1))
  (vla-delete obj)
  ret)
(defun-r 'add-slot)

(defun-q get-vlist (pline / xypts i)
  "Returns the list of vertices comprising PLINE in the form (x y bulge)"
  (declare (vars (pline vla-object)))
  (setq xypts (safearray->list (vlax-variant-value (vla-get-Coordinates pline)))
        i -1)
  (while (< i (- (length xypts) 1))
    (add-to-list 'bulges (vla-getbulge pline (inc! 'i))))
  (mapcar '(lambda (a b) (append a (list b))) xypts bulges))
(defun-r 'get-vlist)

(defun-q line-obj->vert-lst (line / name ret add-z)
  "Returns the list of vertices comprising LINE

LINE can be a Line, Polyline or 3DPolyline"
  (declare (vars (line vla-object (member (vla-get-objectname line)
                                          '("AcDbLine" "AcDbPolyline" "AcDb3dPolyline")))))
  (defun add-z (p) (append p '(0.0)))
  (setq name (vla-get-objectname line))
  (cond
    ( (= name "AcDbLine")
      nil
      (mapcar
       '(lambda (f)
         (safearray-value (vlax-variant-value ((eval f) line)))
         ) ;_ lambda
       '(vla-get-startpoint vla-get-endpoint)
       ) ;_ mapcar
      ) ;_ condif
    ( (member name '("AcDbPolyline" "AcDb3dPolyline"))
      (setq ret
            (vlax-safearray->list
             (vlax-variant-value (vla-get-Coordinates line))
             ) ;_ vlax-safearray->list
            ) ;_ setq
      (setq ret
            (if (= name "AcDbPolyline")
                (mapcar 'add-z (2-item-list ret))
                (3-item-list ret))))))
(defun-r 'line-obj->vert-lst)

