(set-file-docstring
 "Table-handling functions

Some functions here have side effects such as modifying text in MLEADER callouts to match
the table. This is primarily fab-oriented but, with care, can be used for any table
operations.")

(defun-q table-p (obj)
  "Returns T if OBJ is a table"
  (declare (vars (obj vla-object)))
  (= "AcDbTable" (vla-get-ObjectName obj)))
(defun-r 'table-p)

(defun-q material-table-p (table)
  "Returns T if TABLE is a *Material table, i.e that its title %ends-with \"Material\""
  (declare (vars (table vla-object (table-p table))))
  (ends-with (table--get-title-maybe table) "Material"))
(defun-r 'material-table-p)

(defun-q get-all-tables (/ ret)
  "Returns a list of all tables"
  (vlax-for ent modelSpace
    (if (= (vla-get-objectname ent) "AcDbTable")
        (add-to-list 'ret (table--fix ent))))
  ret)
(defun-r 'get-all-tables)

(defun-q get-table-by-title (title / ret)
  "Returns the table with matching TITLE

TITLE is matched against cell A1 of all tables in the database using wcmatch, so any
wildcards wcmatch accepts will work here.

Automatically calls %table--fix"
  (declare (vars (title str)))
  (vlax-for ent modelSpace
    (if (and (= (vla-get-objectname ent) "AcDbTable")
             (wcmatch (vla-GetText ent 0 0) title))
        (setq ret ent)))
  (if ret
      (table--fix ret)))
(defun-r 'get-table-by-title)

(defun-q table-get-title (table)
  "Returns the contents of cell 0,0"
  (declare (vars (table vla-object (table-p table))))
  (vla-GetText table 0 0))
(defun-r 'table-get-title)

(defun-q table--get-title-maybe (table)
  "If the symbol TITLE is defined, returns its value. Otherwise, calls %table-get-title"
  (declare (vars (table vla-object (table-p table))))
  (if title title (table-get-title table)))
(defun-r 'table--get-title-maybe)

(defun-q table-get-nrows (table)
  "Returns the number of effective rows in TABLE

Effective rows includes the title, header, and body rows except for the weight summary row
in *Material tables."
  (declare (vars (table vla-object (table-p table))))
  ((if (material-table-p table) 1- +) (vla-get-Rows table)))
(defun-r 'table-get-nrows)

(defun-q table--body-rows (table)
  "Returns a list of body (not title or header) row indices in TABLE"
  (declare (vars (table vla-object (table-p table))))
  (range (1+ (if header-row header-row (table--get-header-row table)))
         (- (vla-get-Rows table)
            (if (ends-with (table--get-title-maybe table) "Material") 1 0))
         1))
(defun-r 'table--body-rows)

(defun-q table--cells (table / rows cols ret)
  "Returns a list of cells in TABLE as 2-item lists of the form (row col)"
  (declare (vars (table vla-object (table-p table))))
  (foreach row (range 0 (vla-get-Rows table) 1)
    (foreach col (range 0 (vla-get-Columns table) 1)
      (add-to-list 'ret (list row col)))))
(defun-r 'table--cells)

(defun-q get-table-cell (/ pt table)
  "Prompts user for a point and checks if it is inside in a table

If so, it returns a list of the form (table row column)."
  (osnap-off)
  (setq pt (getpoint "\nSelect table cell: ")
        table (car
               (strip-nil
                (mapcar '(lambda(table) (if (table--pt-in-table-p pt table) table))
                        (get-all-tables)))))
  (osnap-on)
  (if table
      (car
       (strip-nil
        (mapcar
         '(lambda(cell / row col)
            (setq row (car cell)
                  col (cadr cell))
            (if (apply 'within-box-p (cons pt (get-table-cell--bounding-box table row col)))
                (cons table cell)))
         (table--cells table))))))
(defun-r 'get-table-cell)

(defun-q table--pt-in-table-p (pt table / minpt maxpt min-x max-x min-y max-y x y)
  "Returns T if PT is within TABLE's bounding box"
  (declare (vars (pt list (point-p pt))
                 (table vla-object (table-p table))))
  (vla-GetBoundingBox table 'minpt 'maxpt)
  (within-box-p pt (vlax-safearray->list minpt) (vlax-safearray->list maxpt)))
(defun-r 'table--pt-in-table-p)

(defun-q get-table-cell--bounding-box (table row col / extents)
  "Returns the bounding box for the cell as a list of the form (minpt maxpt)"
  (declare (vars (table vla-object (table-p table))
                 (row int)
                 (col int)))
  (setq extents (3-item-list
                 (vlax-safearray->list
                  (vlax-variant-value
                   (vla-GetCellExtents table row col :vlax-false)))))
  (list (nth 2 extents) (nth 1 extents)))
(defun-r 'get-table-cell--bounding-box)

(defun-q table--format-column-name (name)
  "Returns NAME as lowercase with spaces replaced by underscores"
  (declare (vars (name str)))
  (string-subst-all "_" " " (lowercase name)))
(defun-r 'table--format-column-name)

(defun-q table--empty-first-column-p (table row)
  "Returns T if column 0 of ROW is empty"
  (declare (vars (table vla-object (table-p table))
                 (row int)))
  (= (vla-GetText table row 0) ""))
(defun-r 'table--empty-first-column-p)

(defun-q table--get-header-row (table)
  "Returns the index of the row containing the column headers"
  (declare (vars (table vla-object (table-p table))))
  (if (table--empty-first-column-p table 1) 2 1))
(defun-r 'table--get-header-row)

(defun-q table-get-column-names (table / header-row)
  "Returns the column names as a list of formatted strings

Strings are formatted by %table--format-column-name"
  (declare (vars (table vla-object (table-p table))))
  (setq header-row (table--get-header-row table))
  (mapcar '(lambda(col) (table--format-column-name (vla-GetText table header-row col)))
          (range 0 (vla-get-Columns table) 1)))
(defun-r 'table-get-column-names)

(defun-q table--get-column-names-maybe (table)
  "If the symbol 'col-names is defined, returns its value. Otherwise, calls %table-get-column-names"
  (declare (vars (table vla-object (table-p table))))
  (if col-names col-names (table-get-column-names table)))
(defun-r 'table--get-column-names-maybe)

(defun-q table-insert-rows (table nrows after / title rowht)
  "Inserts NROWS rows in a fab drawing TABLE after row AFTER

If this is a generic fab table there are other generic fab tables below this one, move
them down to accomodate. If this is the Assembly* table, update the quantities in Material
and Hardware tables, if found."
  (declare (vars (table vla-object (table-p table))
                 (nrows int)
                 (after int)))
  (setq rowht (vla-GetRowHeight table 0)
        title (table-get-title table))
  (--shift-tables-maybe title (- (* rowht nrows)))
  (vla-InsertRows table (1+ after) rowht nrows)
  (if (begins-with title "Assembly")
      (progn
        (setq mt (get-table-by-title "Material")
              ht (get-table-by-title "Hardware")
              table-nrows (length (table--body-rows table)))
        (if mt
            (progn
              (table-divide-qty mt (- table-nrows nrows))
              (table-multiply-qty mt table-nrows)))
        (if ht
            (progn
              (table-divide-qty ht (- table-nrows nrows))
              (table-multiply-qty ht table-nrows)))))
  (if (material-table-p table)
      (progn
        (mat-table--update-marks table + nrows (1+ after))
        (mat-table--update-total-weight table))))
(defun-r 'table-insert-rows)

(defun-q table-trim-empty-rows (table row / title nrows mt ht table-nrows)
  "Trims empty rows from TABLE, starting at ROW

Determines if row is empty using %table--empty-first-column-p"
  (declare (vars (table vla-object (table-p table))
                 (row int)))
  (setq title (table-get-title table)
        table-nrows (table-get-nrows table)
        nrows 0)
  (while (and (< row table-nrows)
              (table--empty-first-column-p table row))
         (inc! 'row)
         (inc! 'nrows))
  (if (= title "Material")
      (mat-table--update-marks table - nrows (- row nrows)))
  (if (> nrows 0)
      (vla-DeleteRows table (- row nrows) nrows))
  (if (begins-with title "Assembly")
      (progn
        (setq mt (get-table-by-title "Material")
              ht (get-table-by-title "Hardware"))
        (if mt
            (progn
              (table-divide-qty mt (- table-nrows 2))
              (table-multiply-qty mt (- table-nrows nrows 2))))
        (if ht
            (progn
              (table-divide-qty ht (- table-nrows 2))
              (table-multiply-qty ht (- table-nrows nrows 2))))))
  (--shift-tables-maybe title (* (vla-GetRowHeight table 0) nrows)))
(defun-r 'table-trim-empty-rows)

(defun-q --shift-tables-maybe (title delta-y)
  "If TITLE is one of the generic fab tables, call %--shift-tables"
  (cond ((= title "Material")
         (--shift-tables '("Hardware" "Assembly*" "Finish") delta-y))
        ((= title "Hardware")
         (--shift-tables '("Assembly*" "Finish") delta-y))
        ((wcmatch title "Assembly*")
         (--shift-tables '("Finish") delta-y))))
(defun-r '--shift-tables-maybe)

(defun-q --shift-tables (names delta-y)
  "Move the tables named in NAMES vertically by DELTA-Y"
  (declare (vars (names (list nil) (vl-every 'stringp names))
                 (delta-y nil (numberp delta-y))))
  (foreach tbl (mapcar 'get-table-by-title names)
           (vla-Move tbl (3dpt 0 0) (3dpt 0 delta-y))))
(defun-r '--shift-tables)

(defun-q table--column-name->letter (name / pos)
  "Returns the letter designation of the column named NAME

Either 'col-names or 'table must be defined."
  (declare (vars (name str)))
  (setq pos (vl-position name (table--get-column-names-maybe table)))
  (if (< pos 27)
      (chr (vl-string-elt "ABCDEFGHIJKLMNOPQRSTUVWXYZ" pos))
      ((lambda(/ mark)
         (setq mark "AA")
         (repeat (- pos 27) (setq mark (increment-mark mark)))
         mark))))
(defun-r 'table--column-name->letter)

(defun-q table-multiply-qty (table n / i lastrow qty row-ht)
  "Multiplies every quantity in TABLE by N"
  (declare (vars (table vla-object (table-p table))
                 (n int)))
  (setq qty-col (vl-position "qty" (table--get-column-names-maybe table))
        i 1
        lastrow (table-get-nrows table))
  (vla-put-RegenerateTableSuppressed table :vlax-true)
  (while (< (inc! 'i) lastrow)
         (setq qty (atoi (vla-GetText table i qty-col)))
         (if (> qty 0) (vla-SetText table i qty-col (to-string (* qty n)))))
  (vla-put-RegenerateTableSuppressed table :vlax-false))
(defun-r 'table-multiply-qty)

(defun-q table-divide-qty (table n / i lastrow qty)
  "Divides every quantity in TABLE by N"
  (declare (vars (table vla-object (table-p table))
                 (n int)))
  (setq qty-col (vl-position "qty" (table--get-column-names-maybe table))
        i 1
        lastrow (table-get-nrows table))
  (vla-put-RegenerateTableSuppressed table :vlax-true)
  (while (< (inc! 'i) lastrow)
         (setq qty (atoi (vla-GetText table i qty-col)))
         (if (> qty 0) (vla-SetText table i qty-col (to-string (/ qty n)))))
  (vla-put-RegenerateTableSuppressed table :vlax-false))
(defun-r 'table-divide-qty)

(defun-q table-set-row (table row data / title cols idx orig-ht)
  "Sets a ROW in TABLE

ROW is a row index and DATA is a data list whose keys correspond to the table columns. If TABLE is a material table, attempts to set the weight formula.

To add data in bulk with formula regeneration disabled to improve performance, see
%table-add-data"
  (declare (vars (table vla-object (table-p table))
                 (row int (>= row 0))
                 (data nil (data-list-p data))))
  (setq cols (table--get-column-names-maybe table)
        idx 0
        orig-ht (vla-GetRowHeight table row)
        title (table-get-title table))
  (foreach col cols
    (if (setq col (xd-value col data))
        (vla-SetText table row idx (to-string col)))
    (inc! 'idx))
  (if (ends-with title "Material")
      (mat-table--set-weight-column table row))
  (--shift-tables-maybe title (- (- (vla-GetRowHeight table row) orig-ht))))
(defun-r 'table-set-row)

(defun-q table-add-data (table datas / row title orig-ht)
  "Adds bulk data to the end of TABLE with regeneration suppressed

DATAS is a list of data lists representing row data."
  (declare (vars (table vla-object (table-p table))
                 (datas list (vl-every 'data-list-p datas))))
  (setq row (1- (table-get-nrows table)))
  (vla-put-RegenerateTableSuppressed table :vlax-true)
  (table-insert-rows table (length datas) row)
  (foreach data datas
           (inc! 'row)
           ((lambda(--shift-tables-maybe)
              (table-set-row table row data))
            list))
  (vla-put-RegenerateTableSuppressed table :vlax-false)
  (vla-Update table)
  (setq title (table-get-title table)
        orig-ht (vla-GetRowHeight table 0))
  (foreach _ datas
           (--shift-tables-maybe title (- (- (vla-GetRowHeight table row) orig-ht)))
           (dec! 'row)))
(defun-r 'table-add-data)

(defun-q parse-table (table / gettext empty-first-col format-col-name header-row rows ncols cols)
  "Returns the data in TABLE as a list of data lists

Each list item corresponse to a row. The keys of the datalists correspond to the column
names. The names are converted to lowercase and spaces are replaced with underscores to
form the key names.

Rows whose first column is blank are not considered. The first row is the title row, with
the table title at 0,0. The second row is considered the header row and column names come
from this row."
  (declare (vars (table vla-object (table-p table))))
  (defun gettext (nrow ncol) (vla-GetText table nrow ncol))
  (defun empty-first-col (nrow) (= (gettext nrow 0) ""))
  (defun format-col-name (name) (string-subst-all "_" " " (lowercase name)))
  (setq header-row (table--get-header-row table)
        rows (range (1+ header-row) (vla-get-Rows table) 1)
        ncols (range 0 (vla-get-Columns table) 1)
        cols (mapcar '(lambda(ncol) (format-col-name (gettext header-row ncol)))
                     ncols))
  (strip-nil
   (mapcar
    '(lambda(nrow)
      (if (not (empty-first-col nrow))
          (mapcar '(lambda(col-name ncol) (list col-name (gettext nrow ncol)))
                  cols ncols)))
    rows)))
(defun-r 'parse-table)

(defun-q table--fix (table)
  "Performs all necessary fixes on TABLE

Currently:
%table--abbr-desc-column
%mat-table--fix-summary-row (material tables only)
"
  (declare (vars (table vla-object (table-p table))))
  (table--abbr-desc-column table)
  (if (material-table-p table) (mat-table--fix-summary-row table))
  table)
(defun-r 'table--fix)

(defun-q table--abbr-desc-column (table / col-names header-row desc-col)
  "If there is a \"Description\" column in TABLE, rewrites it as \"Desc\""
  (declare (vars (table vla-object (table-p table))))
  (setq col-names (table-get-column-names table))
  (if (member "description" col-names)
      (vla-SetText table
                   (table--get-header-row table)
                   (vl-position "description" col-names)
                   "Desc"))
  table)
(defun-r 'table--abbr-desc-column)

(defun-q mat-table--fix-summary-row (table / row)
  "Change the last row to be a simple sum of weight column if necessary"
  (declare (vars (table vla-object (table-p table))))
  (setq row (1- (vla-get-Rows table)))
  (if (= (vla-GetText table row 3) "Tot Wt")
      (progn
        (mapcar '(lambda(col)
                  (vla-SetText table row col ""))
                '(3 4 5 6))
        (mat-table--update-total-weight table))))
(defun-r 'mat-table--fix-summary-row)

(defun-q table--decimal-format-cell (table row col)
  "Wrapper for vla-SetCellFormat"
  (declare (vars (table vla-object (table-p table))
                 (row int)
                 (col int)))
  (vla-setvar 'luprec 1)
  (vla-SetCellDataType table row col acDouble acUnitDistance)
  (vla-SetCellFormat table row col "%lu2%pr1"))
(defun-r 'table--decimal-format-cell)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; material table specific functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun-q mat-table--update-marks (table +- n row / prefix row rows mark marks)
  "Update fab mark numbers in TABLE and callouts

+- should be one of the built-in functions + or -, indicating whether the mark numbers
should increase or decrease. N is the number of places the number should jump. ROW is the
first row in the material table containing mark numbers to change."
  (declare (vars (table vla-object (table-p table))
                 (+- subr (member +- (list + -)))
                 (n int)
                 (row int)))
  (defun new-mark (mark)
    (strcat prefix (to-string (+- (atoi (vl-string-left-trim prefix mark)) n))))
  (setq mark-col (vl-position "mark" (table--get-column-names-maybe table))
        prefix (vl-string-right-trim "0123456789" (vla-GetText table 2 mark-col))
	rows (1- (vla-get-Rows table)))
  (while (and (< row rows)
	      (or (begins-with (setq mark (vla-GetText table row mark-col)) prefix)
		  (= mark "")))
    (if (/= mark "")
        (progn
          (add-to-list 'marks (list mark (new-mark mark)))
          (vla-SetText table row mark-col (new-mark mark))))
    (inc! 'row))
  (mapcar '(lambda(x) (vla-replace-string modelSpace (car x) (cadr x)))
          ((if (= +- +) reverse identity) marks))
  (vla-Regen acadDoc acAllViewPorts))
(defun-r 'mat-table--update-marks)

(defun-q mat-table--update-total-weight (table / row col col-letter header-row need-weight-p)
  "Updates the formula for each weight in TABLE to include any added rows"
  (declare (vars (table vla-object (table-p table) (material-table-p table))))
  (setq col (vl-position "weight" (table--get-column-names-maybe table))
        col-letter (table--column-name->letter "weight")
        header-row (table--get-header-row table)
        row header-row
        last-row (1- (vla-get-Rows table))
        need-weight-p T)
  (while (< (inc! 'row) last-row)
         (if (/= (vla-GetText table row col) "")
             (setq need-weight-p nil
                   row last-row)))
  (if need-weight-p
      (vla-SetText table (1+ header-row) col "0.0"))
  (table--decimal-format-cell table row col)
  (vla-SetText table last-row col
               (strcat "=SUM("
                       col-letter
                       (to-string (+ 2 header-row))
                       ":"
                       col-letter
                       (to-string last-row)
                       ")")))
(defun-r 'mat-table--update-total-weight)

(defun-q mat-table--set-weight-column (table row / formula weight-col wid-col len-col qty-col)
  "Does necessary formatting and sets the weight column"
  (declare (vars (table vla-object (table-p table))
                 (row int)))
  (setq weight-formula (mat-table--calc-weight-formula table row))
  (if weight-formula
      (progn
        (setq weight (vl-position "weight" cols))
        (table--decimal-format-cell table row weight)
        (vla-SetText table row weight weight-formula)
        (table--decimal-format-cell table row weight))))
(defun-r 'mat-table--set-weight-column)

(defun-q mat-table--calc-weight-formula (table row / cols description line wt wt-type col)
  "Returns the weight formula for a material description

For qty, weight, and length columns, if they are required and a null value is present, set the column to a default value. For weight and length, the default is 0. For qty, the default is 1 (Misc Material) or the number of assemblies (Material)."
  (declare (vars (table vla-object (table-p table))
                 (row int)))
  (setq cols (table--get-column-names-maybe table)
        description (vla-GetText table row (vl-position "desc" cols))
        line (assoc description (read-csv "fab/material.csv")))
  (if line
    (progn
      (defun col (name)
        (strcat (table--column-name->letter name) (to-string (1+ row))))
      (defun coldefault (col / coln def)
        (setq coln (vl-position col cols)
              def (cond ((member col '("width" "length"))
                         "0")
                        ((= col "qty")
                         (if (= (table-get-title table) "Material")
                             (to-string (length (parse-table (get-table-by-title "Assembly*"))))
                             "1"))))
        (if (= (vla-GetText table row coln) "")
            (vla-SetText table row coln def)))
      (setq wt (nth 1 line)
            wt-type (nth 2 line))
      (cond ((= wt-type "sqft")
             (mapcar 'coldefault '("qty" "width" "length"))
             (strcat "=(" (col "qty") "*" (col "length") "*" (col "width") "/144)*" wt))
            ((= wt-type "foot")
             (mapcar 'coldefault '("qty" "length"))
             (strcat "=(" (col "qty") "*" (col "length") "/12)*" wt))
            ((= wt-type "each")
             (coldefault "qty")
             (strcat "=" (col "qty") "*" wt))))))
(defun-r 'mat-table--calc-weight-formula)
