(set-file-docstring
 "Functions related to user selection")

(defun-q get-stairland (selection sort-p / filter ss enames)
  "Gets a selection of stairs and landings from the user

Returns enames of selected objects. SELECTION should be one of the following symbols:

- 'stair
- 'land
- 'both

SORT-P being non-nil means to sort selection set using %sort-stairland."
  (declare (vars (selection sym (member selection '(stair land both)))))
  (setq filter (list
                (list -3
                      (list
                       (cond ((= selection 'stair) "Stair,3D_Stair")
                             ((= selection 'land) "Land,3D_Land,Landing")
                             ((= selection 'both) "Stair,3D_Stair,Land,3D_Land,Landing"))))))

  ;; pickfirst
  (if (setq ss (ssget "I" filter))
      (progn
        (setq enames (ss->lst ss))
        (sssetfirst nil nil))
    ;; normal selection
    (if (setq ss (ssget filter))
        (setq enames (ss->lst ss))))

  (if (and enames sort-p)
      (vl-sort enames 'sort-stairland)
      enames))
(defun-r 'get-stairland)

(defun-q get-sset (app-ids-str / filter ss enames)
  "Gets a selection from the user of objects with given XDATA app-ID"
  (setq filter (list (list -3 app-ids-str)))
  (if (setq ss (ssget "I" filter))
      (progn (setq enames (ss->lst ss)) (sssetfirst nil nil))
      (if (setq ss (ssget filter)) (setq enames (ss->lst ss))))
  enames)
(defun-r 'get-sset)

(defun-q ss->lst (ss / i ents)
  "Returns selection set SS as a list of enames"
  (setq i -1)
  (repeat (sslength ss)
          (add-to-list 'ents (ssname ss (setq i (1+ i)))))
  ents)
(defun-r 'ss->lst)

(defun-q get-ins-and-delete (ename / blkIns-p ins)
  "Deletes the entity named ENAME and returns its former insertion point"
  (declare (vars (ename ename)))
  (setq ins (cdr (assoc 10 (entget ename))))
  (entdel ename)
  ins)
(defun-r 'get-ins-and-delete)

(defun-q vlasel ()
  "Alias for (ename>vlobj (car (entsel)))"
  ((lambda (ent) (if ent (ename>vlobj ent))) (car (entsel))))
(defun-r 'vlasel)

(defun-q getdkword (ig-args def-sym prompt-str / def-val kwords result)
  "Wrapper for initget/getkword that uses default values and remembers the last choice

IG-ARGS is a 2-item list of args to initget. The first element should be an int and the
second should be the list of keywords as a string. DEF-SYM is a symbol pointing to the
variable that holds the default value. PROMPT-STR is the prompt string, to which a leading
newline and the bracketed prompt items are added automatically.

If the keyword list contains a variant of \"matchlast\" (case-insensitive, so the caller
can determine the matching key sequence), the user may enter that keyword to skip
remaining prompts. This works by checking the value of the MATCH-LAST-P variable, which
should be set local in the calling command if used."
  (declare (vars (ig-args (list)
                          (= (length ig-args) 2)
                          (= (type (car ig-args)) 'int)
                          (stringp (last ig-args)))
                 (def-sym (sym))
                 (prompt-str (str))))
  (apply 'initget ig-args)
  (setq def-val (eval def-sym)
        kwords (string-subst-all "/" " " (last ig-args))
        result (if match-last-p
                   def-val
                   (getkword (strcat "\n" prompt-str " [" kwords "] <" def-val ">: "))))
  (if (and result
           (= (lowercase result) "matchlast"))
      (setq match-last-p T))
  (if (and result
           (/= (lowercase result) "matchlast"))
      (set def-sym result)
      (setq result def-val))
  result)
(defun-r 'getdkword)

(defun-q getdint (def-sym prompt-str / def-val)
  "Wrapper for getint that uses default values and remembers the last choice

See %getdkword for usage of DEF-SYM, PROMPT-STR, and MATCH-LAST-P."
  (declare (vars (def-sym (sym))
                 (prompt-str (str))))
  (setq def-val (vl-symbol-value def-sym))
  (or* (list (if match-last-p
                 def-val
                 (getint (strcat "\n" prompt-str " <" (to-string def-val) ">: ")))
             def-val)))
(defun-r 'getdint)

(defun-q ss-or (group-codes)
  "Returns a list of DXF codes wrapped in a DXF OR

Mainly for use in %dxf-filter."
  (declare (vars (group-codes nil (listp group-codes))))
  (append '((-4 . "<OR")) group-codes '((-4 . "OR>"))))
(defun-r 'ss-or)

(defun-q ss-and (group-codes)
  "Returns a list of DXF codes wrapped in a DXF AND

Mainly for use in %dxf-filter."
  (declare (vars (group-codes nil (listp group-codes))))
  (append '((-4 . "<AND")) group-codes '((-4 . "AND>"))))
(defun-r 'ss-and)

(defun-q dxf-filter (defs / dxf-codes dxf-trans)
  "Translates S-Expressions into DXF codes for passing to ssget

DEFS should be a list of group code S-Expressions. Note: %macro-expand is called on DEFS
before use. See source for which codes are available for use."
  (declare (vars (defs nil (listp defs) (vl-every 'listp defs)))
           (tests (equal (dxf-filter '((appid "Land"))) '((-3 ("Land"))))
                  (equal (dxf-filter '((and (type "CIRCLE") (layer "DIM"))))
                         '((-4 . "<AND") (0 . "CIRCLE") (8 . "DIM") (-4 . "AND>")))))
  (setq dxf-codes '((appid -3)
                    (type 0)
                    (text 1)
                    (name 2)
                    (linetype 6)
                    (text-style 7)
                    (layer 8)
                    (pt1 10)
                    (pt2 11)
                    (angle 50)
                    (color 62)))

  (defun dxf-trans (def / code)
    (setq code (car def))
    (if (member code '(or and))
        ((eval (read (strcat "ss-" (to-string code)))) (mapcar 'dxf-trans (cdr def)))
        (list (cons (xd-value code dxf-codes)
                    (if (= code 'appid)
                        (list (cdr def))
                        (cadr def))))))

  (defun flatten-to-cons-cells (lst / ret)
    (foreach elt lst
      (if (or (not (listp elt))
              ;; special-case appid as above
              (= (car elt) -3)
              (not (vl-list-length elt)))
          (add-to-list 'ret elt)
          (mapcar '(lambda(elt) (add-to-list 'ret elt))
                  (flatten-to-cons-cells elt))))
    ret)

  (flatten-to-cons-cells (mapcar 'dxf-trans (macro-expand defs))))
(defun-r 'dxf-filter)

(defun select-all (objtype / ss gpcode dataval ret)
  (setq ss (vla-Add (vla-get-SelectionSets acadDoc) (p-blk-name nil))
        gpcode (vlax-make-safearray vlax-vbInteger '(0 . 0))
        dataval (vlax-make-safearray vlax-vbVariant '(0 . 0)))
  (vlax-safearray-put-element gpcode 0 0)
  (vlax-safearray-put-element dataval 0 objtype)
  (vla-Select ss acSelectionSetAll nil nil gpcode dataval)
  (setq ret (collection->list ss))
  (vla-Delete ss)
  ret)

