(set-file-docstring "String functions")

(defun-q stringp (x)
  "Returns nil if X is not string"
  (= (type x) 'STR))

(declare-late 'stringp
 '((tests (stringp "hello")
          (not (stringp 1)))))
(defun-r 'stringp)

(defun-q strjoin (str-list delimiter)
  "Concatenates strings in STR-LIST with DELIMITER

Example:

(strjoin '(\"hello\" \"world\") \", \")

returns the string \"hello, world\"."
  (vl-string-right-trim delimiter
                        (apply 'strcat
                               (mapcar '(lambda (str) (strcat str delimiter))
                                       str-list))))

(declare-late 'strjoin
 '((vars (str-list nil (listp str-list)
                       (vl-every 'stringp str-list))
         (delimiter str))
   (tests (= (strjoin '("key" "val") "=") "key=val")
          (= (strjoin '("hello" "world") ", ") "hello, world"))))
(defun-r 'strjoin)

(defun-q strsplit (str delimiter / dlen return loc i next-word)
  "Splits STR on a DELIMITER

Example:

(strsplit \"hello, world\" \", \")

returns the list (\"hello\" \"world\")."

  (defun next-word (loc i)
    (substr str (1+ i) (- loc i)))

  (setq i 0
        dlen (strlen delimiter))
  (while (setq loc (vl-string-search delimiter str i))
    (add-to-list 'return (next-word loc i))
    (setq i (+ loc dlen)))
  (append return (list (next-word (strlen str) i))))

(declare-late 'strsplit
 '((vars (str str)
         (delimiter str))
   (tests (equal (strsplit "key=val" "=") '("key" "val"))
          (equal (strsplit "hello, world" ", ") '("hello" "world")))))
(defun-r 'strsplit)

(defun-q strbreak (str / return)
  "Returns a list of single character strings comprising STR"
  (foreach char (vl-string->list str)
           (add-to-list 'return (chr char))))

(declare-late 'strbreak
 '((vars (str str))
   (tests (equal (strbreak "hi") '("h" "i"))
          (= (strbreak "") nil))))
(defun-r 'strbreak)

(defun-q wcmatch-escape (str / spec)
  "Escapes special characters in STR for use in wcmatch"
  (setq spec "`#@.*?~[]-")
  (foreach char (strbreak spec)
           (setq str (vl-string-subst (strcat "`" char) char str))))

(declare-late 'wcmatch-escape
 '((vars (str str))
   (tests (= (wcmatch-escape "what # is 1*1?") "what `# is 1`*1`?")
          (= (wcmatch-escape "backtick[`]@my-mail.com~")
           "backtick`[```]`@my`-mail`.com`~"))))
(defun-r 'wcmatch-escape)

(defun-q begins-with (str pattern / spec)
  "Returns nil unless STR begins with PATTERN"
  (wcmatch str (strcat (wcmatch-escape pattern) "*")))

(declare-late 'begins-with
 '((vars (str str)
         (pattern str))
   (tests (begins-with "hello" "h")
          (begins-with "hello" "")
          (not (begins-with "hello" "a")))))
(defun-r 'begins-with)

(defun-q ends-with (str pattern / spec)
  "Returns nil unless STR ends with PATTERN"
  (wcmatch str (strcat "*" (wcmatch-escape pattern))))

(declare-late 'ends-with
 '((vars (str str)
         (pattern str))
   (tests (ends-with "hello" "o")
          (ends-with "hello" "")
          (not (ends-with "hello" "a")))))
(defun-r 'ends-with)

(defun-q lowercase (str)
  "Returns STR in lowercase"
  (strcase (to-string str) T))

(declare-late 'lowercase
 '((vars (str str))
   (tests (= (lowercase "hi") "hi")
          (= (lowercase "HI") "hi"))))
(defun-r 'lowercase)

(defun-q uppercase (str)
  "Returns STR in uppercase"
  (strcase (to-string str) nil))

(declare-late 'uppercase
 '((tests (= (uppercase "HI") "HI")
          (= (uppercase "hi") "HI"))))
(defun-r 'uppercase)

(defun-q string-subst-all (new old str / loc)
  "Uses vl-string-subst repeatedly to substitute all occurrences of OLD for NEW in STR"
  (if (= str "")
      ""
      (progn
        (setq loc (- (strlen new)))
        (while (setq loc (vl-string-search old str (+ loc (strlen new))))
               (setq str (vl-string-subst new old str loc)))
        str)))

(declare-late 'string-subst-all
 '((vars (new str)
         (old str)
         (str str))
   (tests (= (string-subst-all "hello" "hi" "hi hi") "hello hello"))))
(defun-r 'string-subst-all)

(defun-q first-line (str)
  "Returns the first line of STR"
  (car (strsplit str "\n")))

(declare-late 'first-line
 '((vars (str str))
   (tests (= (first-line "hi") "hi")
          (= (first-line "item1\nitem2") "item1"))))
(defun-r 'first-line)

(defun-q escape-quotes (str)
  "Returns STR with double-quotes escaped"
  (string-subst-all "\\\"" "\"" str))

;; testing this function does weird things to docstrings in HTML build
;; best to leave it out
(declare-late 'escape-quotes
 '((vars (str str))))
(defun-r 'escape-quotes)

(defun-q escape-newlines (str)
  "Returns STR with newlines escaped"
  (string-subst-all "\\n" "\n" str))

(declare-late 'escape-newlines
 '((vars (str str))
   (tests (= (escape-newlines "hi") "hi")
          (= (escape-newlines "item1\nitem2") "item1\\nitem2"))))
(defun-r 'escape-newlines)

(defun-q escape-url (url / subs)
  "Substitutes HTML escapes for problematic characters"
  (declare (vars (url str)))
  (setq subs '(("%20" .  " ")
               ("%0D%0A" . "\n")
               ("%22" . "\"")
               ("%3A" . ":")
               ("%3F" . "?")))
  (foreach sub subs
           (setq url (string-subst-all (car sub) (cdr sub) url)))
  url)
(defun-r 'escape-url)

(defun-q increment-mark (mark / glyphs i carry ret)
  "Return the next string in spreadsheet column order"
  (declare (vars (mark str))
           (tests (= (increment-mark "A") "B")
                  (= (increment-mark "Z") "AA")
                  (= (increment-mark "ZZ") "AAA")))
  (setq glyphs (vl-string->list "ABCDEFGHJKLMNQRSTWXYZ")
        mark (reverse (vl-string->list mark))
        carry T)
  (while (and carry mark)
         (setq i (vl-position (car mark) glyphs))
         (add-to-list 'ret
                      (if (= i (1- (length glyphs)))
                          (progn
                            (setq carry T)
                            (car glyphs))
                          (progn
                            (setq carry nil)
                            (nth (1+ i) glyphs))))
         (setq mark (cdr mark)))
  (cond (mark
         (setq ret (append ret mark)))
        (carry
         (setq ret (cons (ascii "A") ret))))
  (vl-list->string (reverse ret)))
(defun-r 'increment-mark)
