(set-file-docstring "Symbol-handling functions")

(defun-q defined (x)
  "Returns nil if X is null."
  (not (null x)))
(defun-r 'defined)

(defun-q swap! (v1 v2 / tmp)
  "Swaps two variables."
  (set 'tmp (vl-symbol-value v1))
  (set v1 (vl-symbol-value v2))
  (set v2 tmp))
(defun-r 'swap!)

(defun-q symcat (str-lst)
  "Concatenates strings in STR-LST and returns the value of the resulting symbol."
  (vl-symbol-value (read (apply 'strcat (macro-expand str-lst)))))
(defun-r 'symcat)

(defun-q macro-expand (expr-lst / eval-next)
  "Return quoted list EXPR-LST with certain elements evaluated.

Macro-expand will evaluate elements of EXPR-LST which are prefaced by a #. Example:

(setq foo 4
      bar 7)
(macro-expand '(+ 1 #foo #(+ bar foo 3)))

will return the list '(+ 1 4 14)."
  (foreach expr expr-lst
           (if (= (type expr) 'LIST)
               (if eval-next
                   (setq eval-next (subst! (eval (macro-expand expr)) expr 'expr-lst)
                         eval-next nil)
                 (subst! (macro-expand expr) expr 'expr-lst)))
           (if (= (type expr) 'SYM)
               (if (= expr '#)
                   (setq eval-next T
                         expr-lst (remove-nth (vl-position '# expr-lst) expr-lst))
                   ((lambda(/ name)
                      (setq name (vl-symbol-name expr))
                      (if (begins-with name "#")
                          (subst! (eval (read (substr name 2)))
                                  expr
                                  'expr-lst)))))))
  expr-lst)
(defun-r 'macro-expand)

(defun-q #! (expr-lst)
  "Alias for %macro-expand"
  (macro-expand expr-lst))
(defun-r '#!)

(defun-q sym-lst-p (sym)
  "Returns nil if SYM is not a symbol or if its value is not a list"
  (and (= (type sym) 'SYM)
       (listp (vl-symbol-value sym))))
(defun-r 'sym-lst-p)

(defun-q sidekey (key / )
  "Returns the value of the symbol named [SIDE]-[KEY]."
  (declare (vars (side str)
                 (key str)))
  (vl-symbol-value (read (strcat side "-" key))))
(defun-r 'sidekey)

(defun-q sidekey-rail (key / )
  "Returns the value of the symbol named [SIDE]-rail_[KEY]."
  (declare (vars (side str)
                 (key str)))
  (vl-symbol-value (read (strcat side "-rail_" key))))
(defun-r 'sidekey-rail)

(defun-q tbkey (key)
  "Returns the value of the symbol named [TB]_[KEY]."
  (declare (vars (tb str)
                 (key str)))
  (symcat '(#tb "_" #key)))
(defun-r 'tbkey)

(defun-q fbkey (key)
  "Returns the value of the symbol named [FB]_[KEY]."
  (declare (vars (fb str)
                 (key str)))
  (symcat '(#fb "_" #key)))
(defun-r 'fbkey)

(defun-q lrkey (key)
  "Returns the value of the symbol named [LR]_[KEY]."
  (declare (vars (lr str)
                 (key str)))
  (symcat '(#lr "_" #key)))
(defun-r 'lrkey)

(defun-q cornerkey (key)
  "Returns the value of the symbol named [CORNER]_[KEY]."
  (declare (vars (corner str)
                 (key str)))
  (symcat '(#corner "_" #key)))
(defun-r 'cornerkey)

(defun-q chkargs (chkargs-func-name
           chkargs-args
           /
           chkargs-validators
           chkargs-type
           chkargs-name
           chkargs-val
           chkargs-idx
           chkargs-call-lst
           chkargs-header-string
           chkargs-errors
           chkargs-arg-errors
           chkargs-error-idxs
           chkargs-add-error
           )
  "Validates arguments in the current function.

See %defun-r--process-declare-vars for a simplified way to invoke this function.

CHKARGS-FUNC-NAME should be the current function's name as a string. It is used for
reporting errors.

CHKARGS-ARGS should be an argument spec. Its length should be equal to the number of
arguments this function takes. At minimum, CHKARGS-ARGS should be a list of argument names
as symbols.

For each item in CHKARGS-ARGS:
[car] is the argument name as a symbol.
[cadr] is the expected type, which may be a symbol or a list of symbols
[caddr] is an optional list. Each item is a validator to be run against [car]. This may be
a symbol pointing to a function which takes one argument or a list to eval. If a validator
returns nil, throw an error."
  (defun chkargs-call-lst (error-idxs / lst)
    (setq lst (mapcar 'to-string (mapcar 'car chkargs-args)))
    (foreach i (uniquify error-idxs)
             (subst! (strcat "->" (nth i lst) "<-")
                     (nth i lst)
                     'lst))
    (strcat "        (" chkargs-func-name " " (strjoin lst " ") ")\n"))

  (defun chkargs-header-string (i)
    (strcat "Bad argument: " (to-string (nth i (mapcar 'car chkargs-args))) "\n"))

  (defun chkargs-add-error (msg)
    (add-to-list 'chkargs-arg-errors msg)
    (add-to-list 'chkargs-error-idxs chkargs-idx))

  (setq chkargs-idx -1)
  (foreach arg chkargs-args
           (setq chkargs-idx (1+ chkargs-idx)
                 chkargs-type (cadr arg)
                 chkargs-type (if (= chkargs-type 'LIST) '(LIST nil) chkargs-type)
                 chkargs-type (if (and chkargs-type (not (= (type chkargs-type) 'LIST)))
                                  (list chkargs-type)
                                chkargs-type)
                 chkargs-name (car arg)
                 chkargs-val (vl-symbol-value chkargs-name)
                 chkargs-name (to-string chkargs-name)
                 chkargs-validators (if (> (length arg) 2) (caddr arg))
                 chkargs-arg-errors '())
           ;; check type
           (if (and chkargs-type
                    (not (member (type chkargs-val) chkargs-type)))
               (chkargs-add-error (strcat "        Expected "
                                          (to-string chkargs-type)
                                          ", got "
                                          (to-string (type chkargs-val))
                                          "\n"))
             ;; check validators
             (foreach v chkargs-validators
                      (if (not (if (= (type v) 'SYM)
                                   (apply v (list chkargs-val))
                                 (eval v)))
                          (chkargs-add-error (strcat "        Failed validator: "
                                                     (to-string v)
                                                     "\n")))))
           (if chkargs-arg-errors
               (add-to-list 'chkargs-errors (strcat (chkargs-header-string chkargs-idx)
                                                    (apply 'strcat chkargs-arg-errors)))))
  (if chkargs-errors (error (strcat "Argument error(s):\n"
                                    (chkargs-call-lst chkargs-error-idxs)
                                    (apply 'strcat chkargs-errors)))))
(defun-r 'chkargs)

(defun-q propagate (var-sym value)
  "Sets VAR-SYM to VALUE and calls vl-propagate"
  (set var-sym value)
  (vl-propagate var-sym))
(defun-r 'propagate)
