(set-file-docstring
 "Functions related to unit testing")

(defun-q defun-t (test-sym / test-body test-args test-docstring)
  "Registers function at TEST-SYM as a unit test.

Test functions should:

- take no arguments
- have short (1-line) docstrings describing what they test
- return a list of assertion errors using %assert-all
- be self contained
- NOT set any symbols in the global environment

Each test function will be wrapped with advice that reports failures by printing its name,
docstring, and errors (otherwise it returns nil). A call to the resulting function will be
saved to *file-tests* so %test-build-lsp can write them to tests.lsp. Then, TEST-SYM will
be set to nil to mark the original test function for garbage collection."
  (if *test-build-p*
    (progn
      (setq test-body (defun-q-list-ref test-sym)
            test-args (pop! 'test-body)
            test-docstring (pop! 'test-body))
      (if (not (member '/ test-args))
          (setq test-args (rcons test-args '/)))
      (setq test-args (rcons test-args '*assert-return-error*)
            test-body (cons '(setq *assert-return-error* t) test-body))
      
      (add-to-alist
        '*file-tests*
        *file-name*
        (list
         (to-string
          (#!
           '((lambda (/ errors)
               (setq errors (#(cons 'lambda (cons test-args test-body))))
               (if errors
                   (progn
                     (princ #(strcat "\nTEST "
                              (lowercase (to-string test-sym))
                              ": "
                              test-docstring "\n"))
                     (princ (strjoin errors "\n"))
                     (princ "\n")
                     (setq n-errors (+ n-errors (length errors))))))))))
        t)))
  (set test-sym nil))
(defun-r 'defun-t)

;; pause to process tests defined earlier
;; see beginning of acaddoc.lsp for early defun-t definition
(map-apply '(lambda(*file-name* f-sym) (defun-t f-sym))
           *delayed-test*)
(setq *delayed-test* nil)

(defun-q test-build ()
  "Bootstraps test build process and opens a new blank drawing to carry it out."
  (propagate '*test-build-p* t)
  (vla-Activate (vla-Add (vla-get-Documents acadObj)))
  (princ))
(defun-r 'test-build)

(defun-q test-build-lsp (/ f)
  "Writes tests.lsp with all the test forms in *file-tests*"
  (setq f (open (strcat psc-src-dir "tests.lsp") "w"))
  (foreach file *file-tests*
    (princ (strcat "(psc-include (list \"" (car file) "\"))") f)
    (foreach test (cdr file)
      (princ (strcat "\n;;\n" (to-string test) "\n") f))
    (princ "\n\n" f))
  (princ "(princ)" f)
  (close f))
(defun-r 'test-build-lsp)

(defun-q test-run (/ n-errors)
  "Alias for (load \"tests.lsp\")."
  (setq n-errors 0)
  (load "tests.lsp")
  (if (= n-errors 0)
      (princ "All tests passed!")
      (princ (strcat (to-string n-errors) " errors")))
  (princ))
(defun-r 'test-run)
