(set-file-docstring
 "XDATA-handling functions")

(defun-q format-xdata (xdata_list / xdata_temp xd_type)
  "Formats a list to be applied to an entity as xdata"
  (setq xdata_list
	(apply 'append
               (vl-remove-if '(lambda (p) (= (car p) "app_id")) xdata_list)))
  (if (/= (rem (length xdata_list) 2) 0)
      (error "[format-xdata] Bad data list: odd number of elements"))
  (foreach x xdata_list
	   (setq xd_type (type x))
	   (add-to-list 'xdata_temp
			(cons (cond ((= xd_type 'STR) 1000)
                                    ((= xd_type 'INT) 1070)
                                    ((= xd_type 'REAL) 1040))
			      x))))
(defun-r 'format-xdata)

(defun-q assign-xdata (source app-id xd-lst / bound xd-dxf xd-val)
  "Applies a formatted data list to an entity as xdata"
  (setq xd-lst
        (append (list '(1000 . "app_id") (cons 1000 app-id)) xd-lst))
  (setq bound (cons 0 (length xd-lst))
        xd-dxf (vlax-make-safearray vlax-vbInteger bound)
        xd-val (vlax-make-safearray vlax-vbVariant bound)
        source (if (= (type source) 'ename) (ename>vlobj source) source))
  (if (not (vla-appid-defined app-id))
      (vla-add (vla-get-registeredapplications acaddoc) app-id))
  (vlax-safearray-fill xd-dxf (cons 1001 (mapcar 'car xd-lst)))
  (vlax-safearray-fill xd-val (cons app-id (mapcar 'cdr xd-lst)))
  (vla-setxdata source xd-dxf xd-val)
  source)
(defun-r 'assign-xdata)

(defun-q read-xdata (source app-id / xd-dxf xd-val xd-lst xd-tmp xd-ret i)
  "Extracts xdata from an entity and returns a data list"
  (setq source (if (= (type source) 'ename) (ename>vlobj source) source)
        app-id (if (= "*" app-id) "" app-id))
  (vla-getxdata source app-id 'xd-dxf 'xd-val)
  (if xd-val
      (progn
        (setq xd-val (remove (mapcar 'vlax-variant-value
                                     (cdr (vlax-safearray->list xd-val)))
                             '(member x '("{" "}")))
              i 0)
        (repeat (/ (length xd-val) 2)
                (add-to-list 'xd-ret
                             (list (nth i xd-val) (nth (1+ i) xd-val)))
                (setq i (+ i 2)))
        xd-ret)))
(defun-r 'read-xdata)

(defun-q read-xdata-migrate (ename-var
                             appid
                             /
                             appeq
                             data
                             version
                             latest-version
                             derived-keys
                             add-derived-keys
                             tmp
                             str
                             )
  "Reads xdata from an entity, performing migrations if it is an old version"
  (if (setq data (read-xdata ename-var appid))
      (progn
        (ver-ctrl appid)
        (if (< version latest-version)
            (progn
              (foreach var '(derived-keys add-derived-keys)
                       (setq str (strcat appeq "-" (vl-symbol-name var)))
                       (set var (vl-symbol-value (read str))))
              (remove! 'data '(member (car x) derived-keys))
              ;; pending updates
              (while (< version latest-version)
                     (setq tmp (to-string (1+ version))
                           tmp (read (strcat appeq "-migration-" tmp))
                           tmp (attempt-apply tmp (list data))
                           data (if tmp tmp data) updatep T)
                     (subst! (list "version" (inc! 'version))
                             (assoc "version" data)
                             'data))
              (setq data (add-derived-keys data))))
        data)))
(defun-r 'read-xdata-migrate)

(defun draw-xdata-migrate (ename-sym
                           appid
                           /
                           appeq
                           data
                           version
                           latest-version
                           draw
                           )
  (if (setq data (read-xdata (vl-symbol-value ename-sym) appid))
      (progn
        (ver-ctrl appid)
        (if (< version latest-version)
            (progn
              (setq
               data (read-xdata-migrate (vl-symbol-value ename-sym) appid)
               draw (vl-symbol-value (read (strcat "draw-" appid)))
               ) ;_ setq
              (set ename-sym (draw data (get-ins-and-delete (vl-symbol-value ename-sym))))
              ) ;_ progn
            ) ;_ if
        data
        ) ;_ progn
      ) ;_ if
  ) ;_ defun draw-xdata-migrate

  (defun ver-ctrl (appid)
    (setq
      appeq (app-id-equiv appid)
      latest-version (vl-symbol-value (read (strcat appeq "-version")))
    ) ;_ setq
    (if (setq version (assoc "version" data))
      (setq version (cadr version))
      (setq data (append '(("version" 0)) data) version 0)
    ) ;_ if
  ) ;_ defun ver-ctrl

;; alias appids for 3D objects
(defun app-id-equiv (appid)
  (cond ((member (lowercase appid) '("3d_stair")) "Stair")
        ((member (lowercase appid) '("3d_land")) "Land")
        ((member (lowercase appid) '("3d_guard")) "Guard")
        ((member (lowercase appid) '("3d_embed")) "Embed")
        ((= T) appid)))

(defun-q with-data (datalst exprs / keys from-with-data data)
  "Evals EXPRS with DATALST as the environment"
  (setq keys '(/)
        from-with-data T
        data datalst)
  (mapcar! '(lambda(x / key val)
              (setq key (read (car x))
                    val (cadr x))
              (if (= (type val) 'LIST)
                  (setq val (cons 'list val)))
              (add-to-list 'keys key)
              (list 'setq key val))
           'datalst)
  (apply (append '(lambda) (list keys) datalst exprs)
         nil))
(defun-r 'with-data)

(defun-q expect-with-data (funcname)
  "Throws an error if not inside %with-data"
  (if (null from-with-data)
      (error (strcat funcname ": expected with-data call"))))
(defun-r 'expect-with-data)

(defun-q combine-like-data (datalsts keys primary-key
                                     / memb heap queue i h)
  "Combines similar data lists

Similarity is determined by equality on the values of keys listed in KEYS. PRIMARY-KEY, if
provided, is the unique ID key for a list. The values of this key are preserved in the
final grouping. If a key named 'qty' is present in DATALSTS, it will be preserved and
incremented to reflect grouping."
  (defun memb (elt / i h ret)
    (setq i 0
          elt (mapcar '(lambda(k) (xd-value k elt)) keys))
    (while (< i (length heap))
      (setq h (nth i heap))
      (setq h (mapcar '(lambda(k) (xd-value k h)) keys))
      (if (vl-every 'equal h elt)
          (setq ret i
                i 1234567890)
        (inc! 'i)))
    ret)

  (if primary-key
      (mapcar! '(lambda(x) (subst-key primary-key (list (xd-value primary-key x)) x)) 'datalsts))
  (setq heap (list (car datalsts))
        queue (cdr datalsts))
  (foreach q queue
           (if (and (setq i (memb q))
                    (< i 123456789))
               (progn
                 (setq h (nth i heap))
                 (subst-key! "qty"
                             (+ (?n (xd-value "qty" h))
                                (?n (xd-value "qty" q)))
                             'h)
                 (if primary-key
                     (subst-key! primary-key
                                 (append (xd-value primary-key h)
                                         (xd-value primary-key q))
                                 'h))
                 (subst! h (nth i heap) 'heap))
             (add-to-list 'heap q)))
  heap)
(defun-r 'combine-like-data)

(defun-q translate-keys (lst exp-lst / parse-list)
  "Translates key names in EXP-LST based on their value in data list LST

DEPRECATED: use %with-data to do this sort of thing"
  (defun parse-list (exp-lst / res)
    (if (/= (type exp-lst) 'LIST)
        (setq exp-lst (list 'eval exp-lst)))
    (foreach x exp-lst
             (if (= (type x) 'LIST)
                 (subst! (parse-list x) x 'exp-lst))
             (if (= (type x) 'SYM)
                 (if (setq res (assoc (lowercase (to-string x)) lst))
                     (subst! (cadr res) x 'exp-lst))))
    exp-lst)
  (eval (parse-list exp-lst)))
(defun-r 'translate-keys)

(defun-q tk (lst)
  "Alias for %translate-keys

DEPRECATED"
  (translate-keys data lst))
(defun-r 'tk)

(defun-q tkmx (lst)
  "Alias for composition of %translate-keys and %macro-expand

 DEPRECATED"
  (tk (macro-expand lst)))
(defun-r 'tkmx)

(defun-q translate-keys-block (lst-sym block)
  "DEPRECATED"
  (foreach x block
           (translate-keys (vl-symbol-value lst-sym) x)))
(defun-r 'translate-keys-block)

(defun-q get-ent-xdata (appids / ent)
  "Prompts for entity selection and gets XDATA from one of supplied APPIDS

Throws an error if selected entity has no data for any of the APPIDS."
  (declare (vars (appids list (vl-every 'stringp appids))))
  (setq ent (car (entsel "Select: ")))
  (if (and ent (assoc -3 (entget ent appids)))
      (read-xdata ent "")
      (error "XDATA not found")))
(defun-r 'get-ent-xdata)

(defun-q get-stair-xdata ()
  "Wrapper for %get-ent-xdata for stairs"
  (get-ent-xdata '("stair" "3d_stair" "stairtop")))
(defun-r 'get-stair-xdata)

(defun-q get-land-xdata ()
  "Wrapper for %get-ent-xdata for landings"
  (get-ent-xdata '("land" "3d_land" "landtop")))
(defun-r 'get-land-xdata)

;; convenience functions for testing
(defun-q with-data-get-stair (exprs)
  "Evals EXPRS in stair xdata context"
  (with-data (get-stair-xdata) exprs))
(defun-r 'with-data-get-stair)

(defun-q with-data-get-land (exprs)
  "Evals EXPRS in landing xdata context"
  (with-data (get-land-xdata) exprs))
(defun-r 'with-data-get-land)

(defun-q data-list-p (data)
  "Returns T if DATA is a data list"
  (and (listp data)
       (vl-every '(lambda(x) (and (listp x) (stringp (car x))))
                   data)))
(defun-r 'data-list-p)

(defun-q xd-value (key datalst / tmp)
  "Return the value of KEY in DATALIST"
  (if (setq tmp (assoc key datalst)) (cadr tmp)))
(defun-r 'xd-value)

(defun-q vla-appid-defined (appName)
  "Returns the RegisteredApplication object for APPNAME, if it exists"
  (car
   (vl-remove-if-not
    '(lambda (obj) (= (strcase appName) (strcase (vla-get-name obj))))
    (collection->list (vla-get-registeredapplications acaddoc)))))
(defun-r 'vla-appid-defined)

(defun-q get-xdata-find (obj app / assembly coll)
  "Returns the most recently edited OBJ with APP xdata in the current drawing"
  (setq assembly (xd-value "assembly" (read-xdata obj "")))
  (find-first
    '(lambda(x) (= (xd-value "assembly" x) assembly))
    (mapcar '(lambda(e) (read-xdata e app))
            (ss->lst (ssget "A" (#! '((-3 (#app)))))))))
(defun-r 'get-xdata-find)

(defun-q subst-key (key new-val lst)
  "Returns data list LST with value of KEY replaced with NEW-VAL"
  (declare (vars (key str)
                 (lst nil (data-list-p lst)))
           (tests (equal (subst-key 'a 2 '((a 1))) '((a 2)))
                  (equal (subst-key 'a 2 '((b 1))) '((b 1)))))
  (subst (list key new-val) (assoc key lst) lst))
(defun-r 'subst-key)

(defun-q subst-key! (key new-val lst-sym)
  "Updates data list at LST-SYM in place using %subst-key"
  (declare (vars (key str)
                 (lst-sym sym (data-list-p (vl-symbol-value lst-sym))))
           (tests (equal ((lambda(lst) (subst-key! 'a 2 'lst) lst) '((a 1))) '((a 2)))
                  (equal ((lambda(lst) (subst-key! 'a 2 'lst) lst) '((b 1))) '((b 1)))))
  (set lst-sym (subst-key key new-val (vl-symbol-value lst-sym))))
(defun-r 'subst-key!)

(defun-q has-duplicate-keys-p (data / last-key ret)
  "Returns T if any keys in data list DATA are repeated"
  (declare (vars (data nil (data-list-p data)))
           (tests (has-duplicate-keys-p '(("a" 1)("b" 2)("a" 3)))
                  (not (has-duplicate-keys-p '(("a" 1)("b" 2))))))
  (sort! 'data '(lambda(a b) (< (car a) (car b))))
  (foreach elt data
           (if (= (car elt) last-key) (setq ret T))
           (setq last-key (car elt)))
  ret)
(defun-r 'has-duplicate-keys-p)

